#==============================================================================
# package      DE_EPAGES::GoogleSmartShopping::UI::Shop
# state        private
#==============================================================================
package DE_EPAGES::GoogleSmartShopping::UI::Shop;
use base DE_EPAGES::Presentation::UI::Object;

use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::GoogleSmartShopping::API::GoogleApiService;
use DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService;
use DE_EPAGES::GoogleSmartShopping::API::GoogleSmartShoppingUrlBuilder;
use DE_EPAGES::GoogleSmartShopping::API::Tools qw ( LogGoogleSmartShopping GetGoogleApiErrorCode );
use DE_EPAGES::Shop::API::Url qw ( BuildShopUrl );
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError Error );
use HTTP::Status qw (HTTP_OK HTTP_BAD_REQUEST HTTP_FORBIDDEN HTTP_INTERNAL_SERVER_ERROR);
use DE_EPAGES::GoogleSmartShopping::API::Constants qw (
    ERROR_FORM_SUBMIT
    ERROR_ACCEPT_TERMS
    ERROR_TOKEN_EXPIRED_OR_REVOKED
);

use strict;

use constant OAUTH_ONBOARDING => 0;
use constant ACCOUNT_CONNECTION_GET_EMAIL => 1;
use constant MERCHANT_CENTER_ACCOUNT_CONNECTED => 2;
use constant GOOGLE_ADS_ACCOUNT_CONNECTED => 3;

#==============================================================================
# Event handlers
#==============================================================================
sub UnityViewGoogleSmartShopping {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $aSetupStages = $self->_reckonSetupStages($Shop);

    if (OAUTH_ONBOARDING ~~ $aSetupStages) {
        $Servlet->vars('GoogleOAuthConnect', $self->_buildConnectOAuthURL($Servlet, $Shop));
        return $self->SUPER::View($Servlet);
    }

    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    my $hGoogleSmartShoppingConfig = $GoogleSmartShoppingConfig->get([
        'ChecklistContactInfo',
        'ChecklistReturnPolicy',
        'ChecklistBillingTOC',
        'ChecklistGoogleAdsPolicies',
        'MerchantCenterAccountName',
        'MerchantCenterAccountID',
    ]);

    $Servlet->vars('GoogleSmartShoppingConfig', $hGoogleSmartShoppingConfig);
    $Servlet->vars('GoogleMerchantCenterRequirementsMet', $GoogleSmartShoppingConfig->areGoogleMerchantCenterRequirementsMet());

    if (ACCOUNT_CONNECTION_GET_EMAIL ~~ $aSetupStages) {
        my $hEmailResult = $self->_getConnectedAccountEmail($Shop);
        if (defined $hEmailResult && isValid($hEmailResult->{'email'})) {
            $Servlet->vars('ConnectedGoogleAccountEmail', $hEmailResult->{'email'});
            if (defined $hEmailResult->{'error'} && $hEmailResult->{'error'} eq ERROR_TOKEN_EXPIRED_OR_REVOKED) {
                $Servlet->vars('TokenExpiredOrRevoked', 1);
                $Servlet->vars('GoogleOAuthConnect', $self->_buildConnectOAuthURL($Servlet, $Shop));
                return $self->SUPER::View($Servlet);
            }
        } else {
            $Servlet->vars('ConnectGoogleAccountError', 1);
            return $self->SUPER::View($Servlet);
        }
    }

    if(MERCHANT_CENTER_ACCOUNT_CONNECTED  ~~ $aSetupStages) {
        $Servlet->vars('MerchantCenterAccountConnected', 1);
        my $hMerchantCenterAccountStatuses = $self->_getMerchantCenterAccountStatuses($Shop);

        if(defined($hMerchantCenterAccountStatuses)) {
            $Servlet->vars('MerchantCenterAccountStatuses', $hMerchantCenterAccountStatuses);
        }
    } else {
        my $MerchantCenterAccounts = $self->_getMerchantCenterAccounts($Shop, $GoogleSmartShoppingConfig);
        unless (defined($MerchantCenterAccounts)) {
            $Servlet->vars('MerchantCenterAccountsError', 1);
            return $self->SUPER::View($Servlet);
        }
        $Servlet->vars('MerchantCenterAccounts', $MerchantCenterAccounts);
    }

    unless (GOOGLE_ADS_ACCOUNT_CONNECTED  ~~ $aSetupStages) {
        my $GoogleAdsAccounts = $self->_getGoogleAdsAccounts($Shop, $GoogleSmartShoppingConfig);
        unless (defined($GoogleAdsAccounts)) {
            $Servlet->vars('GoogleAdsAccountsListError', 1);
            return $self->SUPER::View($Servlet);
        }
        $Servlet->vars('GoogleAdsAccounts', $GoogleAdsAccounts);
    }

    return $self->SUPER::View($Servlet);
}

#----------------------------------------------------------------------------------------
sub _reckonSetupStages {
    my $self = shift;
    my ($Shop) = @_;

    return [ OAUTH_ONBOARDING ] unless ($Shop->existsChild('GoogleSmartShoppingConfig'));
    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    return [ OAUTH_ONBOARDING ] unless ($GoogleSmartShoppingConfig->get('AccessToken'));
    return [ OAUTH_ONBOARDING ] unless ($GoogleSmartShoppingConfig->get('RefreshToken'));

    my $aSecondStageMultiStep = [ ACCOUNT_CONNECTION_GET_EMAIL ];

    if (isValid($GoogleSmartShoppingConfig->get('MerchantCenterAccountID'))) {
        push(@{$aSecondStageMultiStep}, MERCHANT_CENTER_ACCOUNT_CONNECTED);
    }

    if ($self->_isGoogleAdsAccountConnected($GoogleSmartShoppingConfig)) {
        push(@{$aSecondStageMultiStep}, GOOGLE_ADS_ACCOUNT_CONNECTED);
    }

    return $aSecondStageMultiStep;
}
#----------------------------------------------------------------------------------------


sub GoogleOAuthCallback {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $Form = $Servlet->form();
    my $hQueryParameters = $Form->queryValues();
    my $payload;
    my $Code;

    if ($hQueryParameters->{'code'}) {
        LogGoogleSmartShopping('Callback Success: ', $hQueryParameters);
        my $hResponse = $self->_exchangeAuthCodeForAccessTokens($Shop, $hQueryParameters->{'code'}->[0]);
        LogGoogleSmartShopping('OAuth code exchange response: ', $hResponse);
        if ($hResponse->{'access_token'}) {
            $self->_saveTokensToConfig($Shop, $hResponse->{'access_token'}, $hResponse->{'refresh_token'});
            $payload = "window.opener.ep.fn.callbackSuccess();";
            $Code = HTTP_OK;
        } else {
            $payload = "window.opener.ep.fn.showError('" . quotemeta($hResponse->{'error'}->[0]) . "');";
            $Code = HTTP_INTERNAL_SERVER_ERROR;
        }
    } elsif ($hQueryParameters->{'error'}) {
        LogGoogleSmartShopping('Callback Error: ', $hQueryParameters);
        $payload = "window.opener.ep.fn.showError('" . quotemeta($hQueryParameters->{'error'}->[0]) . "');";
        $Code = HTTP_FORBIDDEN;
    } else {
        LogGoogleSmartShopping('Unexpected parameters to OAuth Callback: ', $hQueryParameters);
        $payload = "window.opener.ep.fn.showError('unknown_error');";
        $Code = HTTP_BAD_REQUEST;
    }
    my $jsContent = <<"EOF";
<script type="text/javascript">
    $payload
    window.close();
</script>
EOF
    return $self->_writeHTTPResponse($Servlet, $Code, {'ContentType' => 'text/html'}, $jsContent);
}

#========================================================================================
# function     CreateMerchantCenterAccount
# state        public
#----------------------------------------------------------------------------------------
# syntax       CreateMerchantCenterAccount( $Servlet );
#----------------------------------------------------------------------------------------
# description  Create a new SubAccount on ePages MerchantCenterAccount
#----------------------------------------------------------------------------------------
# input        $Servlet | Presentation Object | object
# return       $Account | Account details for created MerchantCenterAccount | object
#========================================================================================
sub CreateMerchantCenterAccount {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $Form = $Servlet->form();
    my $hFormValues = $Form->formValues();

    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    Error('Undefined configuration for GoogleSmartShopping') unless defined $GoogleSmartShoppingConfig;

    my $FormErrorReason = $self->_validateCreateSubAccountRequest($hFormValues, $GoogleSmartShoppingConfig);

    if ($FormErrorReason) {
        $Form->executeFormError({
            'Name'   => 'CreationFailed',
            'Reason' => $FormErrorReason
        });
    }

    my $GoogleApiService = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $Shop});

    my $NewMerchantCenterSubAccount;
    eval {
        $NewMerchantCenterSubAccount = $GoogleApiService->createMerchantCenterSubAccount(
            $hFormValues->{'ConnectedGoogleAccountEmail'}->[0],
            $hFormValues->{'HasAdultContent'}->[0]
        );
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Account creation failed', {
            'Error' => $Error->message(),
        });

        $Form->addFormError({
            'Name'   => 'CreationFailed',
            'Reason' => GetGoogleApiErrorCode($Error->message())
        });
    }

    $Form->executeFormError();
    $Servlet->vars('ResponseContent', $NewMerchantCenterSubAccount);

    return;
}

#========================================================================================
# function     DisconnectPersonalAccount
# state        public
#----------------------------------------------------------------------------------------
# syntax       DisconnectPersonalAccount( $Servlet );
#----------------------------------------------------------------------------------------
# description  Disconnects the Google personal account and MerchantCenter account
#----------------------------------------------------------------------------------------
# input        $Servlet | presentation object | object
#========================================================================================
sub DisconnectPersonalAccount {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $Form = $Servlet->form();
    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    return unless defined($GoogleSmartShoppingConfig);

    my $GoogleApiService = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $Shop});

    eval {

        if ($GoogleSmartShoppingConfig->get('IsMerchantCenterSubAccount')) {
            $GoogleApiService->deleteMerchantCenterSubAccount($GoogleSmartShoppingConfig->get('MerchantCenterAccountID'));
        }

        my $AccessToken = $GoogleSmartShoppingConfig->get('AccessToken');
        my $hRevokeTokenResult = $GoogleApiService->revokeAccessToken($AccessToken);
        if (defined $hRevokeTokenResult->{'error'}) {
            Error("Google revoke token failed: ", $hRevokeTokenResult);
        }
    };
    if (ExistsError()) {
        my $Error = GetError();

        LogGoogleSmartShopping('DisconnectPersonalAccount failed', {
            'Error' => $Error->message(),
        });

        $Form->addFormError({
            'Name' => 'DisconnectPersonalAccount',
            'Reason' => GetGoogleApiErrorCode($Error->message())
        });
    }

    $self->_deleteGoogleSmartShoppingConfig($GoogleSmartShoppingConfig);
    LogGoogleSmartShopping("GoogleSmartShoppingConfig deleted for Shop ", $Shop->id() );

    $Form->executeFormError();
    return;
}

#========================================================================================
# function     DisconnectMerchantCenterAccount
# state        public
#----------------------------------------------------------------------------------------
# syntax       DisconnectMerchantCenterAccount( $Servlet );
#----------------------------------------------------------------------------------------
# description  Disconnects the MerchantCenterAccount
#----------------------------------------------------------------------------------------
# input        $Servlet | presentation object | object
#========================================================================================
sub DisconnectMerchantCenterAccount {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $Form = $Servlet->form();
    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    return unless defined($GoogleSmartShoppingConfig);

    my $GoogleApiService = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $Shop});

    eval {
        if ($GoogleSmartShoppingConfig->get('IsMerchantCenterSubAccount')) {
            $GoogleApiService->deleteMerchantCenterSubAccount($GoogleSmartShoppingConfig->get('MerchantCenterAccountID'));
        }
    };
    if (ExistsError()) {
        my $Error = GetError();

        LogGoogleSmartShopping('DisconnectMerchantCenterAccount failed', {
            'Error' => $Error->message(),
        });

        $Form->addFormError({
            'Name' => 'DisconnectMerchantCenterAccount',
            'Reason' => GetGoogleApiErrorCode($Error->message())
        });
    }

    $GoogleSmartShoppingConfig->deleteMerchantCenterAccount();
    LogGoogleSmartShopping("GoogleSmartShoppingConfig MerchantCenterAccount deleted", {});

    $Form->executeFormError();
    return;
}

#========================================================================================
# function     UnitySFViewGoogleSmartShoppingHTML
# state        public
#----------------------------------------------------------------------------------------
# syntax       UnitySFViewGoogleSmartShoppingHTML( $Servlet );
#----------------------------------------------------------------------------------------
# description  Hook function for site verification
#               Generates HTML response with google-site-verification token
#----------------------------------------------------------------------------------------
# input        $Servlet | presentation object | object
#========================================================================================
sub UnitySFViewGoogleSmartShoppingHTML {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object->get('Site');
    my $Response = $Servlet->response;
    $Response->header( 'Content-Type' => 'text/html' );

    return unless ($Shop->existsChild('GoogleSmartShoppingConfig'));
    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');

    my $GSSSiteVerificationToken= $GoogleSmartShoppingConfig->get('SiteVerificationToken');
    my $GoogleSearchSiteVerificationToken = $Shop->get('GoogleSiteVerificationToken');

    my $RequestURI;
    if ($Servlet->serverVariables('QUERY_STRING') =~ /^URI=([^&]+)/) {
        $RequestURI = $1;
    }
    if($GSSSiteVerificationToken && $GSSSiteVerificationToken eq $RequestURI) {
        # display html file for Google Smart shopping
        my $Content = 'google-site-verification: ' . $GSSSiteVerificationToken;
        $Response->content($Content);
    } elsif($GoogleSearchSiteVerificationToken && $GoogleSearchSiteVerificationToken eq $RequestURI) {
        # display html file for Google search
        my $Content= 'google-site-verification: ' . $GoogleSearchSiteVerificationToken;
        $Response->content( $Content);
    } else {
        # display 404 error
        $Response->content('No google*.html available.');
        $Response->code('404');
    }

    $Servlet->vars('ExitAfterEvent', 1);

    return;
}

#----------------------------------------------------------------------------------------
sub _validateCreateSubAccountRequest {
    my $self = shift;
    my ($hFormValues, $GoogleSmartShoppingConfig) = @_;

    return ERROR_FORM_SUBMIT unless (isValid($hFormValues->{'ConnectedGoogleAccountEmail'}) && isValid($hFormValues->{'ConnectedGoogleAccountEmail'}->[0]));
    return ERROR_FORM_SUBMIT unless (isValid($hFormValues->{'HasAdultContent'}) && isValid($hFormValues->{'HasAdultContent'}->[0]));
    return ERROR_ACCEPT_TERMS unless (isValid($hFormValues->{'newMerchantAccountAcceptTerms'}) && $hFormValues->{'newMerchantAccountAcceptTerms'}->[0]);
    return ERROR_ACCEPT_TERMS unless ($GoogleSmartShoppingConfig->areGoogleMerchantCenterRequirementsMet());

    return;
}

sub _buildConnectOAuthURL {
    my $self = shift;
    my ( $Servlet, $Shop ) = @_;

    my $hParams = {
        'ChangeAction' => 'GoogleOAuthCallback',
        'ObjectID'   => $Shop->id(),
    };
    my $hOptions = {
        'Type' => 'admin',
        'AddToken' => 1,
        'UseSSL' => 1,
    };
    my $CallbackUrl = BuildShopUrl($Shop, $hParams, $hOptions, $Servlet);
    LogGoogleSmartShopping('Callback URL is:', $CallbackUrl);
    my $GoogleUrlBuilder = DE_EPAGES::GoogleSmartShopping::API::GoogleSmartShoppingUrlBuilder->new();

    return $GoogleUrlBuilder->buildConnectOAuthUrl($CallbackUrl);
}

#----------------------------------------------------------------------------------------
sub _exchangeAuthCodeForAccessTokens {
    my $self = shift;
    my ($Shop, $OAuthCode) = @_;

    my $GoogleService = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $Shop});
    return $GoogleService->getOAuthTokens($OAuthCode);
}

#----------------------------------------------------------------------------------------
sub _saveTokensToConfig {
    my $self = shift;
    my ($Shop, $AccessToken, $RefreshToken) = @_;

    unless($Shop->existsChild('GoogleSmartShoppingConfig')) {
        $Shop->insertChild('GoogleSmartShoppingConfig', { 'Alias' => 'GoogleSmartShoppingConfig' });
    }
    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    $GoogleSmartShoppingConfig->set({ 'AccessToken' => $AccessToken });

    if (defined $RefreshToken) {
        $GoogleSmartShoppingConfig->set({ 'RefreshToken' => $RefreshToken });
    }

}

#----------------------------------------------------------------------------------------
sub _getConnectedAccountEmail {
    my $self = shift;
    my ($Shop) = @_;

    my $service = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $Shop});
    my $hUserInfo = $service->getAndSaveUserInfo();
    if (!defined $hUserInfo->{'email'}) {
        LogGoogleSmartShopping("Google OpenID userinfo lookup failed: ", $hUserInfo);
    }
    return $hUserInfo;
}

sub _deleteGoogleSmartShoppingConfig {
    my $self = shift;
    my ($GoogleSmartShoppingConfig) = @_;

    $GoogleSmartShoppingConfig->delete();

}

#----------------------------------------------------------------------------------------
sub _writeHTTPResponse {
    my $self = shift;
    my ($Servlet, $Code, $hParams, $Content) = @_;

    $Servlet->response->code($Code);
    foreach my $key (keys %{$hParams}) {
        $Servlet->params($key, $hParams->{$key});
    }
    return $Servlet->writeContentOutput($Content);
}

sub _getMerchantCenterAccounts {
    my $self = shift;
    my ($Shop) = @_;

    my $service = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $Shop});
    return $service->getMerchantCenterAccounts();
}

#----------------------------------------------------------------------------------------
sub _getMerchantCenterAccountStatuses {
    my $self = shift;
    my ($Shop) = @_;

    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    return unless defined $GoogleSmartShoppingConfig;

    my $MerchantCenterAccountID = $GoogleSmartShoppingConfig->get('MerchantCenterAccountID');
    return unless defined $MerchantCenterAccountID;

    my $MultiClientAccountId = $GoogleSmartShoppingConfig->{'MerchantCenterMultiClientAccountID'} // $MerchantCenterAccountID;

    my $service = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $Shop});
    LogGoogleSmartShopping("Google accountLevelIssues lookup : ", {$MultiClientAccountId,$MerchantCenterAccountID});
    my $hResult = $service->getMerchantCenterAccountStatuses($MultiClientAccountId, $MerchantCenterAccountID);

    unless (defined $hResult) {
        LogGoogleSmartShopping("Google accountLevelIssues lookup failed: ", $hResult);
        return;
    }

    my $hMerchantCenterAccountStatuses = {
        AccountLevelIssues => $hResult->{'accountLevelIssues'},
        IsWebsiteClaimed => 0,
        IsWebsiteClaimedByOthers => 0,
    };

    if ($hResult->{'websiteClaimed'}) {
        $hMerchantCenterAccountStatuses->{'IsWebsiteClaimed'} = 1;
        $hMerchantCenterAccountStatuses->{'IsWebsiteClaimedByOthers'} = 0;
        unless ($GoogleSmartShoppingConfig->get('MerchantCenterMultiClientAccountID') eq $self->_getSystemMerchantCenterMultiClientAccountID()) {
            $hMerchantCenterAccountStatuses->{'IsWebsiteClaimedByOthers'} = 1;
        }
    }

    return $hMerchantCenterAccountStatuses;
}

sub _getSystemMerchantCenterMultiClientAccountID {
    my $self = shift;

    return LoadRootObject()->get('GoogleSmartShopping_MerchantCenterMultiClientAccountID');

}

sub _getGoogleAdsAccounts {
    my $self = shift;
    my ($Shop) = @_;

    my $service = DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService->new({Shop => $Shop});
    my $AdsAccounts;
    eval {
        $AdsAccounts = $service->getGoogleAdsAccounts();
        LogGoogleSmartShopping('GoogleAdsAccounts:', $AdsAccounts);
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Error getting GoogleAdsAccounts: ', $Error->message());
    }
    return $AdsAccounts;
}

sub _isGoogleAdsAccountConnected {
    my $self = shift;
    my ($GoogleSmartShoppingConfig) = @_;

    return isValid($GoogleSmartShoppingConfig->get('GoogleAdsAccountID'));

}
1;
