#========================================================================================
# package      DE_EPAGES::GoogleSmartShopping::t::unit::GoogleApiService
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for UI::Provider
#========================================================================================
package DE_EPAGES::GoogleSmartShopping::t::unit::GoogleAdsApiService;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use Test::Mockify::Sut;
use Test::Mockify::Matcher qw ( String );
use Test::Mockify::Verify qw ( WasCalled GetParametersFromMockifyCall );
use Test::Exception;
use Test::MockObject::Extends;
use Test::More;

use strict;

use constant AUTH_CODE => 'SOME_AUTH_CODE';

use constant ACCOUNT_EMAIL => 'some_email@example.com';
use constant ADMIN_REFRESH_TOKEN => 'SOME_ADMIN_REFRESH_TOKEN';
use constant ADMIN_ACCESS_TOKEN => 'SOME_ADMIN_ACCESS_TOKEN';
use constant EPAGES_ADS_DEVELOPER_TOKEN => 'SOME_EPAGES_ADS_DEVELOPER_TOKEN';

use constant APP_CLIENT_ID => 'SomeClientId';
use constant APP_CLIENT_SECRET => 'SomeClientSecret';
use constant AUTH_TOKEN_URL => 'SomeAuthTokenUrl';
use constant APP_REDIRECT_URL => 'SomeAppRedirectUrl';
use constant ADMIN_MULTI_ACCOUNT_ID => 53255422348;

use constant GOOGLE_ADS_ACCOUNT_ID => 9871958046;
use constant EXPECTED_GOOGLE_ADS_ACCOUNTS_RESPONSE => [{
    'name' => 'someGoogleAdsAccountName',
    'id' => GOOGLE_ADS_ACCOUNT_ID }];
use constant EXPECTED_GOOGLE_ADS_ACCOUNTS_EPAGESJ_URL => 'misc/google-smart-shopping/accounts/ads/list';

sub test {
    my $self = shift;

    $self->test_get_merchant_center_accounts_success();
    $self->test_get_merchant_center_accounts_no_accounts();

    done_testing();
}

#----------------------------------------------------------------------------------------
# Tests
#----------------------------------------------------------------------------------------

sub test_get_merchant_center_accounts_success {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = undef;
    my $SUT;
    my $Response;

    lives_ok(
        sub { $SUT = $self->_createSUT({
            'UserAgent' => $MockedUserAgent,
            'DoGetResponse' => EXPECTED_GOOGLE_ADS_ACCOUNTS_RESPONSE}); },
        "$SubTestName - Service initialization works."
    );

    lives_ok(
        sub { $Response = $SUT->getGoogleAdsAccounts(); },
        "$SubTestName - Verify that getGoogleAdsAccounts can be called without errors."
    );

    ok(
        WasCalled($SUT, 'doGet'),
        $SubTestName . ' - Verify that doGet was called.'
    );

    my $aGetParams = GetParametersFromMockifyCall($SUT, 'doGet');
    is( @$aGetParams[0], EXPECTED_GOOGLE_ADS_ACCOUNTS_EPAGESJ_URL, "$SubTestName - Verify doGet call correctly the epagesj endpoint." );
    is( $Response, EXPECTED_GOOGLE_ADS_ACCOUNTS_RESPONSE, "$SubTestName - Verify that Google Ads Accounts list correctly returned." );
}

sub test_get_merchant_center_accounts_no_accounts {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = undef;
    my $SUT;
    my $Response;


    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent, 'DoGetResponse' => []}); },
        "$SubTestName - Service initialization works."
    );

    lives_ok(
        sub { $Response = $SUT->getGoogleAdsAccounts(); },
        "$SubTestName - getGoogleAdsAccounts can be called without errors."
    );
    is_deeply( $Response, [], "$SubTestName - Google Ads Accounts empty list correctly returned." );
}

#----------------------------------------------------------------------------------------
# Helpers
#----------------------------------------------------------------------------------------

sub _createSUT {
    my $self = shift;
    my ($hSUTParams) = @_;

    my $hServiceParams = { 'oauth_code' => AUTH_CODE, 'Shop' => $hSUTParams->{'Shop'} // undef };

    my $Mockify = Test::Mockify::Sut->new('DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService', [$hServiceParams]);

    $Mockify->mockImported('DE_EPAGES::GoogleSmartShopping::API::Tools', 'LogGoogleSmartShopping')
        ->whenAny()
        ->thenReturnUndef();

    if (defined($hSUTParams->{'UserAgent'})) {
        $Mockify->mockStatic('LWP::UserAgent::new')->whenAny()->thenReturn($hSUTParams->{'UserAgent'});
    }

    if (defined($hSUTParams->{'DoGetResponse'})) {
        $Mockify->mock('doGet')
            ->whenAny()
            ->thenReturn($hSUTParams->{'DoGetResponse'});
    }

    if (defined($hSUTParams->{'DoPostResponse'})) {
        $Mockify->mock('doPost')
            ->whenAny()
            ->thenReturn($hSUTParams->{'DoPostResponse'});
    }

    if (defined($hSUTParams->{'DoDeleteResponse'})) {
        $Mockify->mock('doDelete')
            ->whenAny()
            ->thenReturn($hSUTParams->{'DoDeleteResponse'});
    }

    if (defined($hSUTParams->{'Error'})) {
        $Mockify->mockImported('DE_EPAGES::Core::API::Error', 'ExistsError')
            ->whenAny()
            ->thenReturn(1);
        $Mockify->mockImported('DE_EPAGES::Core::API::Error', 'GetError')
            ->whenAny()
            ->thenReturn($hSUTParams->{'Error'});
    }

    $Mockify->mock('GetAccessToken')
        ->when(String(ADMIN_REFRESH_TOKEN))
        ->thenReturn(ADMIN_ACCESS_TOKEN);

    $Mockify->mock('loadGoogleSettings')
        ->when()
        ->thenReturnUndef();

    return $Mockify->getMockObject();
}

__PACKAGE__->runScript();

1;