#========================================================================================
# package      DE_EPAGES::GoogleSmartShopping::t::unit::GoogleApiService
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for UI::Provider
#========================================================================================
package DE_EPAGES::GoogleSmartShopping::t::unit::GoogleApiService;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use Test::Mockify::Sut;
use Test::Mockify::Matcher qw ( Object String );
use Test::Mockify::Verify qw ( WasCalled GetParametersFromMockifyCall );
use Test::Exception;
use Test::MockObject::Extends;
use Test::More;

use JSON::MaybeXS qw ( encode_json );
use HTTP::Status qw( HTTP_OK HTTP_BAD_REQUEST HTTP_INTERNAL_SERVER_ERROR );
use DE_EPAGES::GoogleSmartShopping::API::Constants qw (
    EPAGESJ_BASE_URI
);
use strict;

use constant AUTH_CODE => 'SOME_AUTH_CODE';
use constant ACCESS_TOKEN => 'SOME_ACCESS_TOKEN';
use constant REFRESH_TOKEN => 'SOME_REFRESH_TOKEN';
use constant ACCOUNT_EMAIL => 'some_email@example.com';
use constant ADMIN_REFRESH_TOKEN => 'SOME_ADMIN_REFRESH_TOKEN';
use constant ADMIN_ACCESS_TOKEN => 'SOME_ADMIN_ACCESS_TOKEN';

use constant GOOGLE_SITE_VERIFICATION_TOKEN => 'google12345.html';

use constant APP_CLIENT_ID => 'SomeClientId';
use constant APP_CLIENT_SECRET => 'SomeClientSecret';
use constant AUTH_TOKEN_URL => 'SomeAuthTokenUrl';
use constant APP_REDIRECT_URL => 'SomeAppRedirectUrl';
use constant STOREFRONT_WEBSITE_URL => 'https://shop.example.com';
use constant MULTI_CLIENT_ACCOUNT_ID => 389123456;
use constant MERCHANT_CENTER_ID => 458599058;
use constant EXPECTED_ACCESS_TOKEN_RESPONSE => { 'access_token' => ACCESS_TOKEN };
use constant EXPECTED_USERINFO_RESPONSE => { 'email' => ACCOUNT_EMAIL };
use constant EXPECTED_MERCHANT_CENTER_RESPONSE => [{ 'email' => ACCOUNT_EMAIL, 'id' => MERCHANT_CENTER_ID }];
use constant EXPECTED_MERCHANT_CENTER_ACCOUNTS_RESPONSE => [{ 'name' => 'someMerchantCenterAccountName', 'id' => MERCHANT_CENTER_ID }];
use constant EXPECTED_MERCHANT_CENTER_ACCOUNTS_EPAGESJ_URL => 'misc/google-smart-shopping/accounts/merchant-center-accounts';
use constant EXPECTED_SITE_VERIFICATION_TOKEN_EPAGESJ_URL => 'misc/google-smart-shopping/accounts/site-verification-token';
use constant EXPECTED_SITE_VERIFICATION_EPAGESJ_URL => 'misc/google-smart-shopping/accounts/site-verification';
use constant EXPECTED_ERROR_RESPONSE => { 'error' => 'INTERNAL_SERVER_ERROR', 'errorMessage' => 'Internal Server Error' };
use constant REFRESH_ERROR_DESCRIPTION => 'RefreshToken expired or has been revoked';
use constant EXPECTED_INVALID_TOKEN_RESPONSE => { 'error' => 'invalid_token', 'errorMessage' => 'Token expired or revoked'};
use constant EXPECTED_GOOGLE_INVALID_TOKEN_RESPONSE => { 'error' => 'invalid_token', 'error_description' => 'Token expired or revoked'};
use constant INVALID_GRANT_ERROR_RESPONSE => {'error' => 'invalid_grant', 'error_description' => 'Bad Request'};
use constant ADMIN_MULTI_ACCOUNT_ID => 53255422348;
use constant REFRESH_ERROR_DEFUNCT_TOKEN => {'error' => 'token_expired_or_revoked'};

use constant EXPECTED_SITE_VERIFICATION_TOKEN_RESPONSE => {'token' => GOOGLE_SITE_VERIFICATION_TOKEN};
use constant EXPECTED_SITE_VERIFICATION_RESPONSE => {};

use constant EXPECTED_MERCHANT_CENTER_ACCOUNT_STATUSES_EPAGESJ_RESPONSE => {
    "accountId"=> "414922848",
    "accountLevelIssues"=> [
        {
            "documentation"=> "https://support.google.com/merchants/answer/9849758",
            "id"=> "merchant_quality_low",
            "severity"=> "error",
            "title"=> "Account isn't eligible for enhanced free listings"
        },
        {
            "documentation"=> "https://support.google.com/merchants/answer/6159060",
            "id"=> "missing_ad_words_link",
            "severity"=> "error",
            "title"=> "No Google Ads account linked"
        }
    ],
    "kind"=> "content#accountStatus",
    "websiteClaimed"=> "false",
    "accountManagement"=> "manual"
};
use constant EXPECTED_MERCHANT_CENTER_ACCOUNT_STATUSES_EPAGESJ_URL => 'misc/google-smart-shopping/accounts/'.MULTI_CLIENT_ACCOUNT_ID.'/account-statuses/'.MERCHANT_CENTER_ID;

sub test {
    my $self = shift;

    $self->test_get_oauth_token_success();
    $self->test_get_oauth_token_failure();
    $self->test_get_oauth_token_bad_request();
    $self->test_get_and_save_user_info_success();
    $self->test_get_and_save_user_info_failure();
    $self->test_get_and_save_user_info_defunct_token_no_email();
    $self->test_get_and_save_user_info_defunct_token_with_email();
    $self->test_revoke_token_success();
    $self->test_revoke_token_failure();
    $self->test_revoke_token_server_error();
    $self->test_post_create_merchantCenter_subAccount_success();
    $self->test_delete_merchantCenter_subAccount_success();
    $self->test_get_merchant_center_accounts_success();
    $self->test_get_merchant_center_accounts_no_accounts();
    $self->test_post_claim_website_success_no_multiClientId();
    $self->test_post_claim_website_success_epagesMultiClientId();

    $self->test_get_site_verification_token();
    $self->test_trigger_site_verification();

    $self->test_get_merchant_center_account_statuses_success();
    $self->test_get_merchant_center_account_statuses_error();

    done_testing();
}

#----------------------------------------------------------------------------------------
# Tests
#----------------------------------------------------------------------------------------

sub test_get_oauth_token_success {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = $self->_createUserAgent(HTTP_OK, EXPECTED_ACCESS_TOKEN_RESPONSE);
    my $SUT;
    my $hTokenResponse;

    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent}); },
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $hTokenResponse = $SUT->getOAuthTokens(); },
        "$SubTestName - getOAuthTokens can be called without errors."
    );
    is( $hTokenResponse->{'access_token'}, ACCESS_TOKEN, "$SubTestName - AccessToken correctly returned." );

}

sub test_get_oauth_token_failure {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = $self->_createUserAgent(HTTP_INTERNAL_SERVER_ERROR, EXPECTED_ERROR_RESPONSE);
    my $SUT;
    my $hTokenResponse;

    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent});},
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $hTokenResponse = $SUT->getOAuthTokens(); },
        "$SubTestName - getOAuthTokens can be called without errors."
    );
    is( $hTokenResponse->{'error'}, 'INTERNAL_SERVER_ERROR', "$SubTestName - AccessToken lookup failed." );
}

sub test_get_oauth_token_bad_request {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = $self->_createUserAgent(HTTP_BAD_REQUEST, INVALID_GRANT_ERROR_RESPONSE);
    my $SUT;
    my $hTokenResponse;

    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent}); },
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $hTokenResponse = $SUT->getOAuthTokens(); },
        "$SubTestName - getOAuthTokens can be called without errors.");

    is( $hTokenResponse->{'error'}, 'invalid_grant', "$SubTestName - AccessToken bad request result (invalid grant)." );

}

sub test_get_and_save_user_info_success {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedShoppingConfig = $self->_createGoogleSmartShoppingConfig();
    my $AlreadyExistingConfig = 1;
    my $MockedShop = $self->_createShop($MockedShoppingConfig, $AlreadyExistingConfig);
    my $MockedUserAgent = undef;
    my $DoGetResponse = EXPECTED_USERINFO_RESPONSE;
    my $SUT;
    my $hUserInfo;

    lives_ok(
        sub { $SUT = $self->_createSUT({
                'UserAgent' => $MockedUserAgent,
                'DoGetResponse' => $DoGetResponse,
                'Shop' => $MockedShop,
        }); },
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $hUserInfo = $SUT->getAndSaveUserInfo(); },
        "$SubTestName - getAndSaveUserInfo can be called without errors."
    );
    is( $hUserInfo->{'email'}, ACCOUNT_EMAIL, "$SubTestName - UserInfo email correctly returned." );
    is(
        $SUT->{'Shop'}->get('GoogleSmartShoppingConfig')->get('AccountEmail'),
        ACCOUNT_EMAIL,
        "$SubTestName - UserInfo email correctly stored in GoogleSmartShoppingConfig object."
    );
}


sub test_get_and_save_user_info_failure {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedShoppingConfig = $self->_createGoogleSmartShoppingConfig();
    my $AlreadyExistingConfig = 1;
    my $MockedShop = $self->_createShop($MockedShoppingConfig, $AlreadyExistingConfig);
    my $MockedUserAgent = undef;
    my $DoGetResponse = EXPECTED_ERROR_RESPONSE;
    my $SUT;
    my $hUserInfo;

    lives_ok(
        sub { $SUT = $self->_createSUT({
                'UserAgent' => $MockedUserAgent,
                'DoGetResponse' => $DoGetResponse,
                'Shop' => $MockedShop,
        }); },
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $hUserInfo = $SUT->getAndSaveUserInfo($MockedShoppingConfig);},
        "$SubTestName - getAndSaveUserInfo can be called without errors."
    );
    is( $hUserInfo->{'error'}, 'INTERNAL_SERVER_ERROR', "$SubTestName - UserInfo lookup failed." );
}

sub test_get_and_save_user_info_defunct_token_with_email {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedShoppingConfig = $self->_createGoogleSmartShoppingConfig({AccountEmail => ACCOUNT_EMAIL});
    my $AlreadyExistingConfig = 1;
    my $MockedShop = $self->_createShop($MockedShoppingConfig, $AlreadyExistingConfig);
    my $MockedUserAgent = undef;
    my $DoGetResponse = undef;
    my $Error = $self->_createError(REFRESH_ERROR_DESCRIPTION);
    my $SUT;
    my $hUserInfo;

    lives_ok(
        sub { $SUT = $self->_createSUT({
                'UserAgent' => $MockedUserAgent,
                'DoGetResponse' => $DoGetResponse,
                'Shop' => $MockedShop,
                'Error' => $Error,
        }); },
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $hUserInfo = $SUT->getAndSaveUserInfo($MockedShoppingConfig);},
        "$SubTestName - getAndSaveUserInfo can be called without errors."
    );
    my $aParams = GetParametersFromMockifyCall($SUT, 'LogGoogleSmartShopping');
    is_deeply(
        $aParams,
        [ 'Error getting account user information: ', REFRESH_ERROR_DESCRIPTION ],
        "$SubTestName - Invalid token error was thrown."
    );
    is_deeply(
        $hUserInfo,
        {
            'email' => ACCOUNT_EMAIL,
            'error' => REFRESH_ERROR_DEFUNCT_TOKEN->{'error'},
        },
        "$SubTestName - error code is `token_expired_or_revoked`, but we still return the saved AccountEmail",
    );
}

sub test_get_and_save_user_info_defunct_token_no_email {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedShoppingConfig = $self->_createGoogleSmartShoppingConfig();
    my $AlreadyExistingConfig = 1;
    my $MockedShop = $self->_createShop($MockedShoppingConfig, $AlreadyExistingConfig);
    my $MockedUserAgent = undef;
    my $DoGetResponse = undef;
    my $Error = $self->_createError(REFRESH_ERROR_DESCRIPTION);
    my $SUT;
    my $hUserInfo;

    lives_ok(
        sub { $SUT = $self->_createSUT({
                'UserAgent' => $MockedUserAgent,
                'DoGetResponse' => $DoGetResponse,
                'Shop' => $MockedShop,
                'Error' => $Error,
        }); },
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $hUserInfo = $SUT->getAndSaveUserInfo($MockedShoppingConfig);},
        "$SubTestName - getAndSaveUserInfo can be called without errors."
    );
    my $aParams = GetParametersFromMockifyCall($SUT, 'LogGoogleSmartShopping');
    is_deeply(
        $aParams,
        [ 'Error getting account user information: ', REFRESH_ERROR_DESCRIPTION ],
        "$SubTestName - Invalid token error was thrown."
    );
    is_deeply(
        $hUserInfo,
        {
            'email' => undef,
            'error' => REFRESH_ERROR_DEFUNCT_TOKEN->{'error'},
        },
        "$SubTestName - error code is `token_expired_or_revoked`",
    );
}

sub test_revoke_token_success {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = $self->_createUserAgent(HTTP_OK);
    my $SUT;
    my $Response;

    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent}); },
        "$SubTestName - Service initialization works."
    );

    lives_ok(sub { $Response = $SUT->revokeAccessToken(ACCESS_TOKEN);},
        "$SubTestName - Test revokeAccessToken lives ok.");

    is_deeply($Response, {} , "$SubTestName - Test revokeAccessToken success.");

}

sub test_revoke_token_failure {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = $self->_createUserAgent(HTTP_INTERNAL_SERVER_ERROR, EXPECTED_ERROR_RESPONSE);
    my $SUT;
    my $Response;

    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent}); },
        "$SubTestName - Service initialization works."
    );

    lives_ok(sub { $Response = $SUT->revokeAccessToken(ACCESS_TOKEN);},
        "$SubTestName - Test revokeAccessToken lives ok.");

    is_deeply($Response, EXPECTED_ERROR_RESPONSE, "$SubTestName - Test revokeAccessToken failure.");

}

sub test_revoke_token_server_error {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = $self->_createUserAgent(HTTP_BAD_REQUEST, EXPECTED_GOOGLE_INVALID_TOKEN_RESPONSE);
    my $SUT;
    my $Response;

    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent}); },
        "$SubTestName - Service initialization works."
    );

    lives_ok(sub { $Response = $SUT->revokeAccessToken(ACCESS_TOKEN);},
        "$SubTestName - Test revokeAccessToken lives ok.");

    is_deeply($Response, EXPECTED_INVALID_TOKEN_RESPONSE, "$SubTestName - Test revokeAccessToken server error.");

}

sub test_get_merchant_center_accounts_success {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = undef;
    my $SUT;
    my $Response;

    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent, 'DoGetResponse' => EXPECTED_MERCHANT_CENTER_ACCOUNTS_RESPONSE}); },
        "$SubTestName - Service initialization works."
    );

    lives_ok(
        sub { $Response = $SUT->getMerchantCenterAccounts(); },
        "$SubTestName - Varify that getMerchantCenterAccounts can be called without errors."
    );

    ok(
        WasCalled($SUT, 'doGet'),
        $SubTestName . ' - Verify that doGet was called.'
    );
    my $aGetParams = GetParametersFromMockifyCall($SUT, 'doGet');
    is( @$aGetParams[0], EXPECTED_MERCHANT_CENTER_ACCOUNTS_EPAGESJ_URL, "$SubTestName - Verify doGet call correctly the epagesj endpoint." );
    is( $Response, EXPECTED_MERCHANT_CENTER_ACCOUNTS_RESPONSE, "$SubTestName - Varify that MerchantCenterAccounts list correctly returned." );
}

sub test_get_merchant_center_accounts_no_accounts {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = undef;
    my $SUT;
    my $Response;


    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent, 'DoGetResponse' => []}); },
        "$SubTestName - Service initialization works."
    );

    lives_ok(
        sub { $Response = $SUT->getMerchantCenterAccounts(); },
        "$SubTestName - getMerchantCenterAccounts can be called without errors."
    );
    is_deeply( $Response, [], "$SubTestName - MerchantCenterAccounts empty list correctly returned." );
}

sub test_delete_merchantCenter_subAccount_success {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = undef;
    my $MockedShoppingConfig = $self->_createGoogleSmartShoppingConfig();
    my $MockedShop = $self->_createShop($MockedShoppingConfig, 1);
    my $DoDeleteResponse = {};
    my $SUT;

    $SUT = $self->_createSUT({
        'UserAgent' => $MockedUserAgent,
        'DoDeleteResponse' => $DoDeleteResponse,
        'Shop' => $MockedShop,
    });

    $SUT->{'GoogleSmartShopping_AdminRefreshToken'} = ADMIN_REFRESH_TOKEN;
    $SUT->{'GoogleSmartShopping_MerchantCenterMultiClientAccountID'} = ADMIN_MULTI_ACCOUNT_ID;
    $SUT->{'GoogleSmartShoppingConfig'} = $MockedShoppingConfig;

    lives_ok(
        sub { $SUT->deleteMerchantCenterSubAccount(MERCHANT_CENTER_ID); },
        "$SubTestName - deleteMerchantCenterSubAccount can be called without errors."
    );

    my $aParams = GetParametersFromMockifyCall($SUT, 'doDelete');
    is_deeply(
        $aParams,
        [
            EPAGESJ_BASE_URI.'accounts/'.ADMIN_MULTI_ACCOUNT_ID.'/sub-accounts/'.MERCHANT_CENTER_ID,
            {'AdminToken' => ADMIN_ACCESS_TOKEN}
        ],
        "$SubTestName - Invalid token error was thrown."
    );

}

sub test_post_create_merchantCenter_subAccount_success {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = undef;
    my $MockedShoppingConfig = $self->_createGoogleSmartShoppingConfig();
    my $MockedShop = $self->_createShop($MockedShoppingConfig, 1);
    my $DoPostResponse = {
        'adultContent' => bless( do{\(my $o = 0)}, 'JSON::XS::Boolean' ),
        'emails' => [
            'epagestesttwo@gmail.com'
        ],
        'id' => 472175609,
        'name' => 'ShopNameOrAlias',
        'websiteUrl' => 'https://unity87t8e67r87ertewqs.vm-hsilva.intern.epages.de'
    };
    my $SUT;
    my $result;

    $SUT = $self->_createSUT({
        'UserAgent' => $MockedUserAgent,
        'DoPostResponse' => $DoPostResponse,
        'Shop' => $MockedShop,
    });

    $SUT->{'GoogleSmartShopping_AdminRefreshToken'} = ADMIN_REFRESH_TOKEN;
    $SUT->{'GoogleSmartShopping_MerchantCenterMultiClientAccountID'} = ADMIN_MULTI_ACCOUNT_ID;
    $SUT->{'GoogleSmartShoppingConfig'} = $MockedShoppingConfig;

    lives_ok(
        sub { $result = $SUT->createMerchantCenterSubAccount('ConnectedGoogleAccount@Email', 1); },
        "$SubTestName - createMerchantCenterSubAccount can be called without errors."
    );

    is( $result, $DoPostResponse, "$SubTestName - account correctly returned." );

    is( $SUT->{'GoogleSmartShoppingConfig'}->get('MerchantCenterAccountID'), 472175609, "$SubTestName - AccountID was inserted in Config." );
    is( $SUT->{'GoogleSmartShoppingConfig'}->get('MerchantCenterAccountName'), 'ShopNameOrAlias', "$SubTestName - Account Name was inserted in Config." );
    is( $SUT->{'GoogleSmartShoppingConfig'}->get('IsMerchantCenterSubAccount'), 1, "$SubTestName - IsMerchantCenterSubAccount was inserted in Config." );
    is( $SUT->{'GoogleSmartShoppingConfig'}->get('MerchantCenterMultiClientAccountID'), ADMIN_MULTI_ACCOUNT_ID, "$SubTestName - MerchantCenterMultiClientAccountID was inserted in Config." );
    ok( defined($SUT->{'GoogleSmartShoppingConfig'}->get('MerchantCenterSubAccountCreationTime')), "$SubTestName - SubAccountCreationTime was inserted in Config." );
}

sub test_post_claim_website_success_no_multiClientId {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = undef;
    my $MockedShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        MerchantCenterAccountID => MERCHANT_CENTER_ID,
        IsMerchantCenterSubAccount => 0,
    });
    my $MockedShop = $self->_createShop($MockedShoppingConfig, 1);
    my $DoPostResponse = {};
    my $SUT;

    $SUT = $self->_createSUT({
        'UserAgent' => $MockedUserAgent,
        'DoPostResponse' => $DoPostResponse,
        'Shop' => $MockedShop,
    });

    $SUT->{'GoogleSmartShopping_AdminRefreshToken'} = ADMIN_REFRESH_TOKEN;
    $SUT->{'GoogleSmartShopping_MerchantCenterMultiClientAccountID'} = ADMIN_MULTI_ACCOUNT_ID;
    $SUT->{'GoogleSmartShoppingConfig'} = $MockedShoppingConfig;

    lives_ok(
        sub { $SUT->claimWebsite(); },
        "$SubTestName - claimWebsite can be called without errors."
    );

    
    ok(
        WasCalled($SUT, 'doPost'),
        $SubTestName . ' - Verify that doPost was called.'
    );
    my $aPostParams = GetParametersFromMockifyCall($SUT, 'doPost');
    is_deeply( $aPostParams, [
        EPAGESJ_BASE_URI.'accounts/claim-website',
        {},
        {
            merchantAccountId => MERCHANT_CENTER_ID,
            multiClientAccountId => MERCHANT_CENTER_ID,
        }
    ], "$SubTestName - Verify doPost claimWebsite call correctly the epagesj endpoint." );

}

sub test_post_claim_website_success_epagesMultiClientId {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = undef;
    my $MockedShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        MerchantCenterAccountID => MERCHANT_CENTER_ID,
        MerchantCenterMultiClientAccountID => ADMIN_MULTI_ACCOUNT_ID,
        IsMerchantCenterSubAccount => 1,
    });
    my $MockedShop = $self->_createShop($MockedShoppingConfig, 1);
    my $DoPostResponse = {};
    my $SUT;

    $SUT = $self->_createSUT({
        'UserAgent' => $MockedUserAgent,
        'DoPostResponse' => $DoPostResponse,
        'Shop' => $MockedShop,
    });

    $SUT->{'GoogleSmartShopping_AdminRefreshToken'} = ADMIN_REFRESH_TOKEN;
    $SUT->{'GoogleSmartShopping_MerchantCenterMultiClientAccountID'} = ADMIN_MULTI_ACCOUNT_ID;
    $SUT->{'GoogleSmartShoppingConfig'} = $MockedShoppingConfig;

    lives_ok(
        sub { $SUT->claimWebsite(); },
        "$SubTestName - claimWebsite can be called without errors."
    );

    
    ok(
        WasCalled($SUT, 'doPost'),
        $SubTestName . ' - Verify that doPost was called.'
    );
    my $aPostParams = GetParametersFromMockifyCall($SUT, 'doPost');
    is_deeply( $aPostParams, [
        EPAGESJ_BASE_URI.'accounts/claim-website',
        {},
        {
            merchantAccountId => MERCHANT_CENTER_ID,
            multiClientAccountId => ADMIN_MULTI_ACCOUNT_ID,
        }
    ], "$SubTestName - Verify doPost claimWebsite with epagesMultiaccountId call correctly the epagesj endpoint." );

}

sub test_get_site_verification_token {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = undef;
    my $DoPostResponse = EXPECTED_SITE_VERIFICATION_TOKEN_RESPONSE;
    my $SUT;
    my $hSiteVerificationToken;

    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent, 'DoPostResponse' => $DoPostResponse }); },
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $hSiteVerificationToken = $SUT->getSiteVerificationToken(STOREFRONT_WEBSITE_URL); },
        "$SubTestName - getSiteVerificationToken can be called without errors."
    );
    my $aGetParams = GetParametersFromMockifyCall($SUT, 'doPost');
    is( @$aGetParams[0], EXPECTED_SITE_VERIFICATION_TOKEN_EPAGESJ_URL,
        "$SubTestName - Verify doPost calls correctly the epagesj site verification endpoint." );


    is( $hSiteVerificationToken->{'token'}, GOOGLE_SITE_VERIFICATION_TOKEN, "$SubTestName -  site verification token correctly returned." );

}

sub test_trigger_site_verification {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = $self->_createUserAgent(HTTP_OK, {});
    my $DoPostResponse = EXPECTED_SITE_VERIFICATION_RESPONSE;
    my $SUT;
    my $response;

    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent, 'DoPostResponse' => $DoPostResponse }); },
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $response= $SUT->triggerSiteVerification(STOREFRONT_WEBSITE_URL); },
        "$SubTestName - triggerSiteVerification can be called without errors."
    );
    my $aGetParams = GetParametersFromMockifyCall($SUT, 'doPost');
    is( @$aGetParams[0], EXPECTED_SITE_VERIFICATION_EPAGESJ_URL,
        "$SubTestName - Verify doPost calls correctly the epagesj site verification endpoint." );


    is_deeply( $response, {}, "$SubTestName - Verify that site verification returns empty response." );

}

sub test_get_merchant_center_account_statuses_success {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedUserAgent = undef;
    my $SUT;
    my $Response;

    lives_ok(
        sub {
            $SUT = $self->_createSUT({
                'UserAgent' => $MockedUserAgent,
                'DoGetResponse' => EXPECTED_MERCHANT_CENTER_ACCOUNT_STATUSES_EPAGESJ_RESPONSE
            });
        },
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $Response = $SUT->getMerchantCenterAccountStatuses(MULTI_CLIENT_ACCOUNT_ID, MERCHANT_CENTER_ID); },
        "$SubTestName - Varify that getMerchantCenterAccounts can be called without errors."
    );

    ok(WasCalled($SUT, 'doGet'), $SubTestName . ' - Verify that doGet was called.');

    my $ExpectedAccountStatus = {
        "documentation"=> "https://support.google.com/merchants/answer/6159060",
        "id"=> "missing_ad_words_link",
        "severity"=> "error",
        "title"=> "No Google Ads account linked"
    };

    my $aGetParams = GetParametersFromMockifyCall($SUT, 'doGet');
    is( @$aGetParams[0], EXPECTED_MERCHANT_CENTER_ACCOUNT_STATUSES_EPAGESJ_URL, "$SubTestName - Verify doGet call correctly the epagesj mc account statuses endpoint." );
    is_deeply( $Response->{'accountLevelIssues'}[1], $ExpectedAccountStatus, "$SubTestName - Varify that MerchantCenter account statuses correctly returned." );
}

sub test_get_merchant_center_account_statuses_error {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];


    my $MockedUserAgent = $self->_createUserAgent(HTTP_INTERNAL_SERVER_ERROR, EXPECTED_ERROR_RESPONSE);
    my $SUT;
    my $hResponse;

    lives_ok(
        sub { $SUT = $self->_createSUT({'UserAgent' => $MockedUserAgent});},
        "$SubTestName - Service initialization works."
    );
    lives_ok(
        sub { $hResponse = $SUT->getMerchantCenterAccountStatuses(MERCHANT_CENTER_ID); },
        "$SubTestName - Verify that getMerchantCenterAccountStatuses can be called without errors."
    );
    is( $hResponse, undef, "$SubTestName - Verify that getMerchantCenterAccountStatuses lookup failed returns undef as response." );

}

#----------------------------------------------------------------------------------------
# Helpers
#----------------------------------------------------------------------------------------

sub _createSUT {
    my $self = shift;
    my ($hSUTParams) = @_;

    my $hServiceParams = { 'oauth_code' => AUTH_CODE, 'Shop' => $hSUTParams->{'Shop'} // undef };

    my $Mockify = Test::Mockify::Sut->new('DE_EPAGES::GoogleSmartShopping::API::GoogleApiService', [$hServiceParams]);

    $Mockify->mockImported('DE_EPAGES::Object::API::Factory', 'LoadRootObject')
        ->whenAny()
        ->thenReturn($self->_createSystem());

    $Mockify->mockImported('DE_EPAGES::GoogleSmartShopping::API::Tools', 'LogGoogleSmartShopping')
        ->whenAny()
        ->thenReturnUndef();

    if (defined($hSUTParams->{'UserAgent'})) {
        $Mockify->mockStatic('LWP::UserAgent::new')->whenAny()->thenReturn($hSUTParams->{'UserAgent'});
    }

    if (defined($hSUTParams->{'DoGetResponse'})) {
        $Mockify->mock('doGet')
            ->whenAny()
            ->thenReturn($hSUTParams->{'DoGetResponse'});
    }

    if (defined($hSUTParams->{'DoPostResponse'})) {
        $Mockify->mock('doPost')
            ->whenAny()
            ->thenReturn($hSUTParams->{'DoPostResponse'});
    }

    if (defined($hSUTParams->{'DoDeleteResponse'})) {
        $Mockify->mock('doDelete')
            ->whenAny()
            ->thenReturn($hSUTParams->{'DoDeleteResponse'});
    }

    if (defined($hSUTParams->{'Error'})) {
        $Mockify->mockImported('DE_EPAGES::Core::API::Error', 'ExistsError')
            ->whenAny()
            ->thenReturn(1);
        $Mockify->mockImported('DE_EPAGES::Core::API::Error', 'GetError')
            ->whenAny()
            ->thenReturn($hSUTParams->{'Error'});
    }

    $Mockify->mock('GetAccessToken')
        ->when(String(ADMIN_REFRESH_TOKEN))
        ->thenReturn(ADMIN_ACCESS_TOKEN);

    $Mockify->mock('loadGoogleSettings')
        ->when()
        ->thenReturnUndef();

    return $Mockify->getMockObject();
}

sub _createShop {
    my $self = shift;
    my ($GoogleSmartShoppingConfig, $AlreadyExistingConfig) = @_;

    my $hAttributes = {
        'ID' => undef,
        GoogleSmartShoppingConfig => $GoogleSmartShoppingConfig,
        'LanguageID' => 1,
        'NameOrAlias' => 'ShopNameOrAlias',
        'WebUrlUnitySSL' => 'WebUrlUnitySSL',
    };
    my $MockifyShop = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Shop::API::Object::Shop', $hAttributes);
    $MockifyShop->addMockWithReturnValueAndParameterCheck(
        'existsChild', $AlreadyExistingConfig, [ {'string' => 'GoogleSmartShoppingConfig'} ]
    );

    return $MockifyShop->getMockObject();
}

sub _createSystem {
    my $self = shift;

    my $hAttributes = {
        'GoogleSmartShopping_AppClientID' => APP_CLIENT_ID,
        'GoogleSmartShopping_AppClientSecret' => APP_CLIENT_SECRET,
        'GoogleSmartShopping_AuthTokenURL' => AUTH_TOKEN_URL,
        'GoogleSmartShopping_AppRedirectURL' => APP_REDIRECT_URL,
        'GoogleSmartShopping_AdminRefreshToken' => ADMIN_REFRESH_TOKEN,
        'GoogleSmartShopping_MerchantCenterMultiClientAccountID' => ADMIN_MULTI_ACCOUNT_ID,
    };

    my $MockifySystem = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Object::API::Object::Site', $hAttributes
    );

    return $MockifySystem->getMockObject();
}

sub _createUserAgent {
    my $self = shift;
    my ($Code, $hResponse) = @_;

    my $Response = $hResponse // {};
    my $MockifyUA = Test::Mockify->new('LWP::UserAgent');
    $MockifyUA->mock('request')->when(Object('HTTP::Request'))->thenReturn($self->_createResponse($Code, $Response));
    return $MockifyUA->getMockObject();
}

sub _createResponse {
    my $self = shift;
    my ($Code, $hResponse) = @_;

    my $MockifyResponse = Test::Mockify->new('HTTP::Response');
    $MockifyResponse->mock('code')->when()->thenReturn($Code);
    my $Response = encode_json($hResponse);
    $MockifyResponse->mock('decoded_content')->when()->thenReturn($Response);
    return $MockifyResponse->getMockObject();
}

sub _createError {
    my $self = shift;
    my ($Message) = @_;

    my $MockifyError = Test::Mockify->new('DE_EPAGES::Core::API::Error');
    $MockifyError->mock('message')->whenAny()->thenReturn($Message);
    return $MockifyError->getMockObject();
}

sub _createGoogleSmartShoppingConfig {
    my $self = shift;
    my ($hConfigAttributes) = @_;
 
    my $hAttributes = {
        AccountEmail => $hConfigAttributes->{'AccountEmail'} // undef,
        MerchantCenterSubAccountCreationTime  => undef,
        IsMerchantCenterSubAccount => $hConfigAttributes->{'IsMerchantCenterSubAccount'} // undef,
        MerchantCenterAccountID => $hConfigAttributes->{'MerchantCenterAccountID'} // undef,
        MerchantCenterMultiClientAccountID => $hConfigAttributes->{'MerchantCenterMultiClientAccountID'} // undef,
        MerchantCenterMultiClientAccountName => undef,
    };

    my $MockifyConfig = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::GoogleSmartShopping::API::Object::GoogleSmartShoppingConfig', $hAttributes
    );

    return $MockifyConfig->getMockObject();
}

__PACKAGE__->runScript();

1;