#========================================================================================
# package      DE_EPAGES::Installer::API::GitPatchBuilder
# state        public
# description  this module generates the whole patch for a new epages release
#               starting with version specific changes,
#               adding Jira Tasks
#               version control branching
#               cartridge patch building
#               language pack patch building
#               patch package building
#========================================================================================
package DE_EPAGES::Installer::API::GitPatchBuilder;

use strict;
use feature 'switch';
use DE_EPAGES::Core::API::File qw (
    GetSubDirectories
    ExistsDirectory
    ExistsFile
    CreateDirectory
    DeleteDirectory
    RunOnDirectory
);
use DE_EPAGES::Installer::API::PatchBuilder;
use DE_EPAGES::Core::API::Git;
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Installer::API::Cartridge;
use DE_EPAGES::Core::API::SystemCommand qw(SystemCommand);

use Git::Repository;

#ERRORDEFINITION
sub ErrorDefinition {
    return {
        'MISSING_TAG' => "Tag '#Tag' could not be found in the git repository:\n".
            "  #Repository\n"
            . "Please check whether the tag exists on the remote repository\n"
            . "and has been fetched into the local repository.\n"
    }
}
#ERRORDEFINITION END

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $GitPatchBuilder = DE_EPAGES::Installer::API::GitPatchBuilder->new(
#                   'verbose' => 0,
#               );
#----------------------------------------------------------------------------------------
# description  Creates a new GitPatchBuilder object, responsible for building patch files
#               for each cartridge and commit changes
#----------------------------------------------------------------------------------------
# input        $hOptions | keys
#                   <ul>
#                       <li>verbose - if 1 all messages where printed
#                   </ul>
#               | hash
# return       $GitPatchBuilder | GitPatchBuilder object | object
#========================================================================================
sub new {
    my $class = shift;
    my ($hOptions) = @_;
    my $self = bless $hOptions, $class;

    return $self;
}

sub validVersion {
    my $self = shift;
    my ($Version) = @_;

    return $Version =~ /[\d]+\.[\d]+\.[\d]+/;
}

sub buildCartridgePatches {
    my $self = shift;
    my ($OldVersion, $NewVersion, $hOptions) = @_;

    die "\nmissing parameter new version" unless (defined $OldVersion);
    die "\nmissing parameter new version" unless (defined $NewVersion);
    die "\nversion '$OldVersion' not valid" unless ($self->validVersion($OldVersion));
    die "\nversion '$NewVersion' not valid" unless ($self->validVersion($NewVersion));
    die "\nforbidden option combination 'build_only_cartridge' and 'build_all_cartridges'" if (
        (defined $hOptions->{'build_only_cartridge'}) and
        (defined $hOptions->{'build_all_cartridges'})
    );
    die "\nbuild all cartridges and build only cartridge '$hOptions->{cartridge}' is not valid"
        if ((defined $hOptions->{'build_all_cartridges'}) and ($hOptions->{'cartridge'}));

    my $ReleaseOld = $self->getOldReleaseTag($OldVersion);
    my $ReleaseNew = $self->getNewReleaseTag($NewVersion);

    my $CommitBranch = $hOptions->{'branch'};
    my $CompanyName = $hOptions->{'company'};
    $CompanyName = 'DE_EPAGES' unless( defined $CompanyName);
    my $CartridgeDir = $ENV{'EPAGES_CARTRIDGES'} . '/' . $CompanyName;
    my $useGit = 1;

    $self->checkoutTargetBranch($CommitBranch);
    my $Commit = $self->_getHeadCommitId();
    $self->printMessage("\tUsing branch '$CommitBranch' with commit '$Commit' to check in generated patch files");

    my $aEpagesCartridgeNames;
    $aEpagesCartridgeNames = GetSubDirectories($CartridgeDir) if ExistsDirectory($CartridgeDir);
    if (((scalar @$aEpagesCartridgeNames) == 0)) {
        die "found no cartridges to build\n";
    }
    foreach my $CartridgePath (@$aEpagesCartridgeNames) {
        my $CartridgeName = $CartridgePath;
        $CartridgeName =~ s/\Q$CartridgeDir\E\///;

        if ($hOptions->{'build_only_cartridge'}) {
            next if ($CartridgeName ne $hOptions->{'build_only_cartridge'});
        }
        next if ($CartridgeName eq 'TLEParser');

        $self->printMessage(
            "\n################################################################".
            "\n##### Starting Patch Cartridge: $CartridgeName #####".
            "\n################################################################\n");

        my $TempDir = $CartridgePath . '/Patches/PatchTemp';
        my $OldCartridgeRootDir = $TempDir .'/'. $ReleaseOld;
        my $NewCartridgeRootDir = $TempDir .'/'. $ReleaseNew;
        $self->printMessage("\t##### Step-01: export release folders #####");
        $self->exportGitReleaseFolders($OldCartridgeRootDir, $NewCartridgeRootDir, $ReleaseOld, $ReleaseNew, $CommitBranch);

        my $OldCartridgeDir = $OldCartridgeRootDir.'/'.$CompanyName.'/'.$CartridgeName;
        my $NewCartridgeDir = $NewCartridgeRootDir.'/'.$CompanyName.'/'.$CartridgeName;

        # check if Cartridge is new
        if (not ExistsDirectory($OldCartridgeDir.'/API')) {
            DeleteDirectory($CartridgePath . '/Patches/PatchTemp/', { 'recursive'=>1});

            # new Cartridge must have: 'Version' => $NewVersion
            my $CartridgeVersion = $CompanyName.'::'.$CartridgeName.
                '::API::Cartridge->new()->version()';
            my $PerlCode = "require \"$NewCartridgeDir/API/Cartridge.pm\";".
                ' die $@ if $@; print '.$CartridgeVersion.';'; ## no critic qw(RequireInterpolationOfMetachars)
            my $Version = SystemCommand( "$ENV{PERL} -e '$PerlCode'", { 'Capture' => 1 });
            die "\nNew cartridge must have new version: $NewVersion (has: $Version)"
                if $Version ne $NewVersion;

            next;
        }

        $self->printMessage("\t##### Step-02: generate patch files #####");
        my $PatchBuilder = DE_EPAGES::Installer::API::PatchBuilder->new(
            {
                OldReleaseTag => $ReleaseOld,
                NewReleaseTag => $ReleaseNew,
                Verbose       => $self->{'verbose'},
                AutoMode      => 0,
                CommitBranch  => $CommitBranch,
            }
        );

        my $PatchFolder = $self->getPatchFolder($CartridgePath, $OldVersion);

        DeleteDirectory($PatchFolder, { 'recursive'=>1});

        $PatchBuilder->generatePatchFiles($OldCartridgeDir, $NewCartridgeDir, $CartridgePath, $useGit);
        $PatchBuilder->createPatchXMLFromPatchFiles($CartridgePath, "${CompanyName}::${CartridgeName}");
        if (ExistsFile($CartridgePath . '/API/Cartridge.pm')){
            my $generateWarnings = 1;
            $PatchBuilder->patchCartridgePm($CartridgePath,$generateWarnings, $useGit);
        }
        DeleteDirectory($CartridgePath . '/Patches/PatchTemp/', { 'recursive'=>1});

        $self->printMessage("\t##### Step-03: add patch folder and Cartridge.pm to CommitBranch '$CommitBranch' #####");
        $self->addPatchFolderToGit($CommitBranch, $PatchFolder);
        $self->addCartridgePmToGit($CartridgePath, $CommitBranch);
    }

    $self->printMessage("\t##### Step-04: commit patch files after commit '$Commit' #####");
    $self->commitChangesToBranch($CommitBranch, $OldVersion, $Commit);
}

sub getNewReleaseTag{
    my $self=shift;
    my ($Version, $hOptions) = @_;

    #get tag from version
    $Version =~ s/\./\_/g;
    my $Tag =   'test_release_'   . $Version;

    return $Tag;
}

sub getOldReleaseTag{
    my $self=shift;
    my ($Version) = @_;

    #get tag from version
    $Version =~ s/\./\_/g;
    my $Tag =   'release_'   . $Version;

    return $Tag;
}

sub getPatchTag {
    my $self=shift;
    my ($Version, $hOptions) = @_;

    #get tag from version
    $Version =~ s/\./\_/g;
    my $Tag =   'patch_'   . $Version;

    return $Tag;
}

#========================================================================================
# function     getPatchFolder
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $PatchFolder = self->getPatchFolder($CartridgePath, $PatchVersion)
#----------------------------------------------------------------------------------------
# description  Helper method for the construction of the name of the folder with the
#               patch files.
#----------------------------------------------------------------------------------------
# input        $CartridgePath | The path for the directory of the Cardridge patch files. | string
# input        $PatchVersion | The version from which the patch folder
#               (e.g. Patch_6_17_36) is to be build. | string
# return       $PatchFolder | The directory of the folder which shall contain the patch
#               files. | string
#========================================================================================
sub getPatchFolder {
    my $self = shift;
    my ($CartridgePath, $PatchVersion) = @_;

    my $PatchVersionUnderline = $PatchVersion;
    $PatchVersionUnderline =~ s/\./\_/g;
    my $PatchFolder = $CartridgePath.'/Patches/Patch_'.$PatchVersionUnderline;

    return $PatchFolder;
}

#========================================================================================
# function     printMessage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->printMessage('message to test');
#----------------------------------------------------------------------------------------
# description  print message if verbose is true
#----------------------------------------------------------------------------------------
# input        $Message | message to print | string
#========================================================================================
sub printMessage {
    my $self = shift;
    my ($Message) = @_;

    print $Message, "\n" if $self->{'verbose'};
}

#========================================================================================
# function     checkoutTargetBranch
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->checkoutTargetBranch($TargetBranch);
#----------------------------------------------------------------------------------------
# description  create the given $TargetBranch from the current branch
#----------------------------------------------------------------------------------------
# input         $TargetBranch | target branch to build patch files | string
#========================================================================================
sub checkoutTargetBranch {
    my $self = shift;
    my ($TargetBranch) = @_;

    RunOnDirectory (
        $ENV{EPAGES_CARTRIDGES},
        sub {
            my $Git = DE_EPAGES::Core::API::Git->new();
            if($Git->ExistsLocalBranch($TargetBranch)) {
                die "Patch file target branch '$TargetBranch' already exists. This case is not supported yet.";
            }
            else {
                $self->printMessage("\tCreate new branch '$TargetBranch'");
                $Git->Checkout($TargetBranch, ['-b']);
            }
            return;
        }
    );
}

#========================================================================================
# function     exportGitReleaseFolders
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportGitReleaseFolders($OldDir, $NewDir, $ReleaseOld, $ReleaseNew, $CommitBranch);
#----------------------------------------------------------------------------------------
#description  export two versions of one cartridge, one tagged with old release and
#               one tagged with the new release tag to different folders
#----------------------------------------------------------------------------------------
# input        $OldDir | directory to export cartridge version tagged with old release tag | string
# input        $NewDir | directory to export cartridge version tagged with new release tag | string
# input        $ReleaseOld | the epages release code for the old version, e.g. release_6_17_36 | string
# input        $ReleaseNew | the epages release code for the new version, e.g. release_6_17_36_1 | string
# input        $CommitBranch | the branch on which the patch code is to be commited. | string
#========================================================================================
sub exportGitReleaseFolders {
    my $self = shift;
    my ($OldDir, $NewDir, $ReleaseOld, $ReleaseNew, $CommitBranch) = @_;

    DeleteDirectory($OldDir, { 'recursive'=>1});
    DeleteDirectory($NewDir, { 'recursive'=>1});
    CreateDirectory($OldDir);
    CreateDirectory($NewDir);

    #get git tags
    my $OldVersion = $self->_getGitTag($ReleaseOld);
    my $NewVersion = $self->_getGitTag($ReleaseNew);

    $self->printMessage("\tGet Git tags for old version '$OldVersion' and new version '$NewVersion'");

    #get cartridge path
    my $CartridgePath = $OldDir;
    $CartridgePath =~ s/\/Patches\/PatchTemp\/$ReleaseOld//;

    #export cartridge versions tagged with old and new release
    RunOnDirectory(
        $ENV{EPAGES_CARTRIDGES},
        sub {
            my $OldVersionTag = 'v'.$OldVersion;

            unless( Git::Repository->run( 'show-ref' => $OldVersionTag)) {
                Error( 'MISSING_TAG', { Tag => $OldVersionTag, Repository => $ENV{EPAGES_CARTRIDGES} } );
            }

            my $Git = DE_EPAGES::Core::API::Git->new();
            $Git->GetPatchCartridge($OldVersion, $OldVersionTag, $CartridgePath, $OldDir);
            $Git->GetPatchCartridge($NewVersion, $CommitBranch, $CartridgePath, $NewDir);
            return;
        }
    );
}

#========================================================================================
# function     addPatchFolderToGit
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->addPatchFolderToGit($CommitBranch, $PatchFolder)
#----------------------------------------------------------------------------------------
# description  add all files under ../Patch_Version to CommitBranch
#----------------------------------------------------------------------------------------
# input        $CommitBranch | current branch | string
# input        $PatchFolder | folder of generated patch files | string
#========================================================================================
sub addPatchFolderToGit {
    my $self = shift;
    my ($CommitBranch, $PatchFolder) = @_;

    if(ExistsDirectory($PatchFolder)) {
        RunOnDirectory(
            $ENV{EPAGES_CARTRIDGES},
            sub {
                my $Git = DE_EPAGES::Core::API::Git->new();
                my @Arguments = ('--all');
                $Git->Add($PatchFolder, \@Arguments);
                return;
            }
        );
    }
}

#========================================================================================
# function     addCartridgePmToGit
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->addCartridgePmToGit($CartridgePath, $CommitBranch)
#----------------------------------------------------------------------------------------
# description  add Cartridge.pm to CommitBranch
#----------------------------------------------------------------------------------------
# input        $CartridgePath | path to specified Cartridge | string
# input        $CommitBranch | current branch | string
#========================================================================================
sub addCartridgePmToGit {
    my $self = shift;
    my ($CartridgePath, $CommitBranch) = @_;

    my $CartridgePm = $CartridgePath.'/API/Cartridge.pm';

    RunOnDirectory(
        $ENV{EPAGES_CARTRIDGES},
        sub {
            my $Git = DE_EPAGES::Core::API::Git->new();
            $Git->Add($CartridgePm);
            return;
        }
    );
}

#========================================================================================
# function     commitChangesToBranch
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->commitChangesToBranch($CommitBranch, $OldVersion)
#----------------------------------------------------------------------------------------
# description  commit Cartridge.pm and patch files to CommitBranch
#----------------------------------------------------------------------------------------
# input        $CommitBranch | current branch | string
# input        $OldVersion | previous epages release version | string
# input        $CommitId | The SHA-1 hash value identifying a commit | string
#========================================================================================
sub commitChangesToBranch {
    my $self = shift;
    my ($CommitBranch, $OldVersion, $CommitId) = @_;

    die "\nmissing parameter commit branch" unless (defined $CommitBranch);
    die "\nmissing parameter old version" unless (defined $OldVersion);
    die "\nmissing parameter commit id" unless (defined $CommitId);

    my $CommitString = "Add the generated patch files: from '$OldVersion' to '$CommitId'";

    RunOnDirectory(
        $ENV{EPAGES_CARTRIDGES},
        sub {
            my $Git = DE_EPAGES::Core::API::Git->new();
            $Git->Commit($CommitString);
            return;
        }
    );
}

#========================================================================================
# function     _getGitTag
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_getGitTag($ReleaseTag)
#----------------------------------------------------------------------------------------
# description  Transforms a actual provided CVS ReleaseTag into an actual provided GitTag.
#----------------------------------------------------------------------------------------
# input        $ReleaseTag | tag for an epages release version | string
# return       $ReleaseTag | tag with underlines | string
#========================================================================================
sub _getGitTag {
    my $self = shift;
    my ($ReleaseTag) = @_;

    $ReleaseTag =~ s/test_release_//;
    $ReleaseTag =~ s/release_//;
    $ReleaseTag =~ s/\_/\./g;

    return $ReleaseTag;
}

#========================================================================================
# function     _getHeadCommitId
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_getHeadCommitId()
#----------------------------------------------------------------------------------------
# description  Get the short commit id (SHA-1) of the current HEAD.
#----------------------------------------------------------------------------------------
# return       $CommitID | return current commit ID of HEAD | string
#========================================================================================
sub _getHeadCommitId {
    my $self = shift;

    my $Git = DE_EPAGES::Core::API::Git->new();
    return $Git->GetCommitID('HEAD', ['--short']);
}

1;
