#========================================================================================
# �package      DE_EPAGES::Product::API::Object::Product
# �state        public
# �base         DE_EPAGES::Content::API::Object::ContentObject
# �description  object interface for Product.
#========================================================================================
package DE_EPAGES::Product::API::Object::Product;
use base DE_EPAGES::Content::API::Object::ContentObject;

use strict;

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::File qw ( CopyFile CleanPath );
use DE_EPAGES::Core::API::Object::NumberFormat;
use DE_EPAGES::Core::API::PerlTools qw ( AddDuration Compare ElementOf GetCurrentDateTime unique );
use DE_EPAGES::Core::API::TestParameter qw ( TestFloat TestHash TestString );
use DE_EPAGES::Customer::API::Shop qw ( GetCustomerForUser );
use DE_EPAGES::MultiStoreCommon::API::Synchronization; ## no critic qw(RequireCorrectImports) # EPG-32372
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias LoadObject LoadObjectByPath );
use DE_EPAGES::Object::API::Language qw ( GetCurrentLanguageID );
use DE_EPAGES::Object::API::Table::Currency qw ( GetInfoCurrency );
use DE_EPAGES::Object::API::TLEWithParentHash;
use DE_EPAGES::Object::API::TLEObjectHash;
use DE_EPAGES::Product::API::Constants qw ( FROM_MASTERPRODUCT_NON_INHERITABLE_ATTRIBUTES_ON_GETWITHPARENT
                                            PRODUCT_IMAGE_SIZE_NAMES
                                            UNITY_ATTRIBUTES_FROM_MASTERPRODUCT_WRITTEN_ON_SUBS_ON_CREATE
                                            UNITY_LOC_ATTRIBUTES_FROM_MASTERPRODUCT_WRITTEN_ON_SUBS_ON_CREATE
                                            UNITY_NOT_UPDATE_ATTRIBUTES_ON_SUBS_ON_UPDATE_MASTER
                                            UNITY_IMAGE_ATTRIBUTES
                                            UNITY_INHERITABLE_ATTRIBUTES );
#@UNITY_FROM_MASTERPRODUCT_INHERITABLE_ATTRIBUTES_ON_GETWITHPARENT
use DE_EPAGES::Product::API::Stock qw ( SendStockWarnings );
use DE_EPAGES::Product::API::Table::ProductAvailabilityEmail qw (
    DeleteProductAvailabilityEmail
    GetPKeyProductAvailabilityEmailsByProductID
);
use DE_EPAGES::Product::API::Table::ProductPrepayment qw (
    DeleteProductPrepayment
    ExistsProductPrepayment
    GetInfoProductPrepayment
    GetPKeyProductPrepaymentsByProductID
    InsertProductPrepayment
    UpdateProductPrepayment
);
use DE_EPAGES::Product::API::Table::ProductPrice qw (
    DeleteProductPrice
    GetInfoProductPricesByProductID
    GetPKeyProductPricesByProductID
    InsertProductPrice
    UpdateProductPrice
);
use DE_EPAGES::Product::API::Table::ProductShippingMethod qw (
    DeleteProductShippingMethod
    GetPKeyProductShippingMethodsByProductID
    InsertProductShippingMethod
);
use DE_EPAGES::Product::API::Table::ProductStockLevel qw ( UpdateStockLevel );
use DE_EPAGES::Product::API::Table::SelectedVariation qw (
    DeleteSelectedVariation
    ExistsSelectedVariationByProductID
    GetInfoSelectedVariationsByProductID
    GetPKeySelectedVariationsByProductID
);
use DE_EPAGES::Product::API::Table::TopSeller qw ( DeleteTopSeller );
use DE_EPAGES::Product::API::XML::ProductDeleteHandler;
use DE_EPAGES::Product::API::XML::ProductExportDriver;
use DE_EPAGES::Product::API::XML::ProductImportHandler;
use DE_EPAGES::Product::API::XML::VariationDeleteHandler;
use DE_EPAGES::Product::API::XML::VariationImportHandler;
use DE_EPAGES::Product::DAL::Product qw (
    DAL_GetSubProductsOrderedByPrice
    DAL_GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice
);
use DE_EPAGES::Shop::API::Constants qw (
    BASKETACTION_ADDNEWPOSITION
    BASKETACTION_ADDQUANTITY
    BASKETACTION_DONTCHANGEQUANTITY
    BASKETACTION_REPLACEQUANTITY
    TAX_MODEL_NET
    STOCK_IS_ORDERABLE
    STOCK_NOT_VISIBLE
);
use DE_EPAGES::Shop::API::Address qw ( IsEUAddress );
use DE_EPAGES::Tax::API::Shop qw( TaxMatrixByEUAddress );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use Scalar::Util qw ( weaken );

use constant POSITION_STEP => 10;
use constant PERCENT => 100;
use constant MONTH_PER_YEAR => 12;
use constant DAYS_PER_MONTH => 30;
use constant LINEITEM_NAME_MAX_LENGTH => 255;

my $FLOAT_FORMATTER = DE_EPAGES::Core::API::Object::NumberFormat->new;

sub ErrorDefinition {
    return {
        'ProductNoPrice'    => 'product ( #ProductID ) has no price.',
        'ProductNoTaxClass' => 'product ( #ProductID ) has no tax class.',
        'ProductAndBundleProductEqual' => 'product ( #ProductID ) can not bundled with itself.',
        'SubProductsCountFeatureLimitExceeded' => 'Attempt insert a new subproduct for product ( #ProductID ). But only #MaxValue subproducts per product are allowed.',
        'NoBundleBundles' => 'Product #ProductID can\'t be added to the bundle, because it is a bundle',
        'NoBundleAppointments' => 'Product #ProductID can\'t be added to the bundle, because it is an appointment',
        'BundleLoop', 'Product #ProductID is reverse bundled with product #BundledProductID',
        'NoVariationSuperProduct' => 'Product #SuperProductAlias(#SuperProductID) is a variation but can\'t be a master product of #ProductAlias',
        'WrongSuperProductPath' => 'Product #SuperProductAlias(#SuperProductID) is in wrong path to be a master product of #ProductAlias',
        'NoChangeSuperProduct' => 'changing of master product not allowed for product #ProductID',
    }
}

#========================================================================================
# �function     className
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $ClassName = $Product->className;
#----------------------------------------------------------------------------------------
# �description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# �return       $ClassName | class name | String
#========================================================================================
sub className  { 'Product' }

#========================================================================================
# �function     dbPackage
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $DbPackage = $Product->dbPackage;
#----------------------------------------------------------------------------------------
# �description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# �return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# �function     insert
# �state        private
#----------------------------------------------------------------------------------------
# �syntax       $Product = DE_EPAGES::Product::API::Object::Product->insert($hInfo);
# �example      my $Class = LoadClassByAlias('Product');
#               my $Product = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# �description  Insert this info as object, afterwards the hook 'OBJ_InsertProduct' is triggered.
#----------------------------------------------------------------------------------------
# �input        $hInfo | attributes of new object | ref.hash
# �return       $Product | new object | object
# �hook         OBJ_InsertProduct | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ProductID     - product identifier   - int
#               </ul>
#========================================================================================
sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    TriggerHook('OBJ_BeforeInsertProduct', { 'hInfo' => $hInfo });

    if ( $hInfo->{'SuperProduct'} ) {
        my $SuperProductID = $hInfo->{'SuperProduct'}->id;
        my $hErrVars = {
            'SuperProductID'    => $SuperProductID,
            'SuperProductAlias' => $hInfo->{'SuperProduct'}->alias,
            'ProductAlias'      => $hInfo->{'Alias'}
        };
        Error('NoVariationSuperProduct', $hErrVars) if defined $hInfo->{'SuperProduct'}->superProduct;
        Error('WrongSuperProductPath', $hErrVars) if $SuperProductID != $hInfo->{'Parent'}->parent->id;
    };

    $class->incrementFeature($hInfo);
    my $self = $class->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertProduct', {'Object' => $self, 'ProductID' => $self->id });
    delete $hInfo->{$_} foreach qw ( Shop TaxClass IsNew IsDailyPrice OrderUnit PriceQuantity MinOrder IntervalOrder WeightUnit Weight RefUnit RefAmount RefContentAmount _StockLevel StockLevelAlert ManufacturerID ManufacturerSKU DeliveryPeriod Length Height Width IsAvailable AvailabilityDate ImageMediumSmall ImageMediumLarge ImageLarge ImageHotDeal SuperProduct HasSubOwnPrices DefaultSubProduct UPCEAN NewnessDate IsAppointment CanPayAppointment DurationAmount DurationUnit RatingAverage PrepaymentType PrepaymentValue EcoParticipationCategory IsDownloadProduct IsCustomizable ExpiryDate _IsMaster TaxClassDigital ); ## no critic qw(ProhibitLongLines)
    return $self;
}

#========================================================================================
# �function     incrementFeature
# �state        private
#----------------------------------------------------------------------------------------
# �syntax       $Class->incrementFeature($hInfo)
#----------------------------------------------------------------------------------------
# �description  Increments Products or Appointments feature for a shop. Called from insert()
#----------------------------------------------------------------------------------------
# �input        $hInfo | attributes of new object | ref.hash
#========================================================================================
sub incrementFeature {
    my $class = shift;
    my ($hInfo) = @_;
    # increment the usage counter of the Products feature
    my $Shop = $hInfo->{'Shop'} || $hInfo->{'Parent'}->getSite();
    if ($hInfo->{'SuperProduct'}) {
        my $SubProductsCount = scalar(@{$hInfo->{'SuperProduct'}->get('SubProducts')});
        my $ProductMaxSubProducts = $Shop->featureMaxValue('ProductMaxSubProducts');
        if ($SubProductsCount >= $ProductMaxSubProducts  ){
            Error( 'SubProductsCountFeatureLimitExceeded', {
                MaxValue => $ProductMaxSubProducts ,
                NewValue => $SubProductsCount+1,
                Product => $hInfo->{'SuperProduct'},
                ProductID => $hInfo->{'SuperProduct'}->id,
            });
        }
    }
    else {
        my $IsAppointment = $hInfo->{'Parent'}->alias eq 'Appointments' ? 1 : 0;
        $IsAppointment = $hInfo->{'IsAppointment'} if exists $hInfo->{'IsAppointment'};
        my $Feature = $IsAppointment ? 'Appointments' : 'Products';
        $Shop->featureIncrementValue($Feature, 1);
    }
}

#========================================================================================
# �function     decrementFeature
# �state        private
#----------------------------------------------------------------------------------------
# �syntax       $Class->decrementFeature($hInfo)
#----------------------------------------------------------------------------------------
# �description  Decrements Products or Appointments feature for a shop. Called from delete()
#----------------------------------------------------------------------------------------
# �input        $hInfo | <ul>
#                           <li>Shop - shop object - object</li>
#                           <li>Feature - feature name: Product or Appointment - string</li>
#                           <li>IsSubProduct - flag if product is subproduct - integer</li>
#                        </ul> | ref.hash
#========================================================================================
sub decrementFeature {
    my $class = shift;
    my ($hInfo) = @_;

    my $Shop = $hInfo->{'Shop'};
    my $Feature  = $hInfo->{'Feature'};
    my $IsSubProduct  = $hInfo->{'IsSubProduct'};
    return unless defined $Shop or defined $Feature or defined $IsSubProduct;

    $Shop->featureIncrementValue($Feature, -1) unless $IsSubProduct;
}

sub clone {
    my $self = shift;
    my ($hInfo, $Recursive) = @_;

    # clone product with child objects
    if ( $self->superProduct ) {
        $hInfo->{'SuperProduct'}   = $hInfo->{'Parent'}->parent;
    } else { #Super Product
        $hInfo->{'SelectedVariations'}= $self->get('SelectedVariations');
    }
    $hInfo->{'RatingAverage'} = undef;
    my $Clone = $self->SUPER::clone($hInfo, $Recursive);

    # clone bundleproductmaps
    if ($self->get('IsBundleProduct')) {
        foreach my $BundleProductMap ( @{$self->get('BundleProductMaps')}) {
            my $hInfoMap = $BundleProductMap->get([qw (BundledQuantity BundledProduct Position) ]);
            my $BundledProduct = $hInfoMap->{'BundledProduct'};
            delete $hInfoMap->{'BundledProduct'};
            $BundledProduct->addToBundle($Clone, $hInfoMap);
        }
        #renumber it
        $Clone->renumberBundleProductMaps;
    }

    # clone downloadproductmaps
    if ($self->get('IsDownloadProduct')) {
        $self->_cloneDownloadProductMaps($Clone);
    }

    # clone list prices and manufacturer
    if (not defined $Clone->superProduct or $Clone->get('HasSubOwnPrices')) {
    # or $Clone->getSite()->get('IsUnityRegistered')) {
        my $CloneID = $Clone->id;
        InsertProductPrice({ %$_, 'ProductID'=>$CloneID }) foreach @{GetInfoProductPricesByProductID($self->id)};
    }

    # clone assigned ShippingMethods
    my $ahKeys = GetPKeyProductShippingMethodsByProductID($self->id);
    InsertProductShippingMethod({'ProductID'=> $Clone->id,'ShippingMethodID'=> $_->{'ShippingMethodID'} })  foreach @$ahKeys ;

    # set the correct ImageInformations after all files are copied to the clone
    if ($self->getSite()->get('IsUnityRegistered')){
        # update JSON attribute ImageInformations on product by using the
        # Object->addImageInformations from ContentObject
        # get Images for ImageInformations attribute from Constants for Products
        $self->addImageInformations([UNITY_IMAGE_ATTRIBUTES], undef);

        # we need to extend the attribute ImageInformations with additional informations
        # for the ImagesSlideShowString. We have to add all image formats to this ImagesSlideShowString hash
        $self->extendImageInformationsSlideShowString(undef);

    }

    return $Clone;
}

sub _cloneDownloadProductMaps {
    my $self = shift;
    my ($Clone) = @_;
    my $MapClass = LoadClassByAlias('DownloadProductMap');
    foreach my $DownloadProductMap ( @{$self->get('DirectDownloadProductMaps')}) {
        my $hInfoMap = $DownloadProductMap->get([qw (TargetUrl IsExternal FileSize Position) ]);
        # Copy files if files saved intern.
        unless ($hInfoMap->{'IsExternal'}) {
            my $SourceFile = $self->get('PrivatePath').'/'.$hInfoMap->{'TargetUrl'};
            my $TargetFile = $Clone->get('PrivatePath').'/'.$hInfoMap->{'TargetUrl'};
            CopyFile( $SourceFile, $TargetFile, {'create_directory' => 1} );
        }
        $MapClass->insertObject({
            'DownloadProduct' => $Clone,
            %$hInfoMap
        });
    }
}

#========================================================================================
# �function     delete
# �state        private
#----------------------------------------------------------------------------------------
# �syntax       $Product->delete;
#----------------------------------------------------------------------------------------
# �description  Deletes this object, therefore the hook 'OBJ_DeleteProduct' is triggered.
#----------------------------------------------------------------------------------------
# �hook         OBJ_DeleteProduct | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ProductID     - product identifier   - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteProduct', {'Object' => $self, 'ProductID' => $self->id });

    # update all line items that reference the deleted product
    my $ClassLineItemProduct = LoadClassByAlias('LineItemProduct');
    $_->set({'Product' => undef}) foreach @{ $ClassLineItemProduct->getObjects('ByProductID', [$self->id]) };

    # delete topseller
    DeleteTopSeller($self->id);

    # delete product prices
    my $ObjectID = $self->id;
    DeleteProductPrice($_) foreach @{GetPKeyProductPricesByProductID($ObjectID)};

    # delete selected variations
    DeleteSelectedVariation($_) foreach @{GetPKeySelectedVariationsByProductID( $ObjectID )};

    # select another default sub product if the current is default
    if ($self->get('IsDefault')) {

        # set default on last sub product, order of delete is normally by position
        my $SuperProduct = $self->superProduct;
        my @VisisbleSubProducts = grep { $_->id != $ObjectID && $_->get('IsVisible') } @{$SuperProduct->subProducts};
        my $NewDefaultSubProduct = scalar @VisisbleSubProducts ? pop @VisisbleSubProducts : undef;

#       see EPG-32372
#       If a MultiStore shop is not allowed to set the DefaultSubProduct attribute,
#       the default sub product can't be deleted, because it's still referred to
#       from the master product.
#       Setting the IS_MULTISTORE_EXIMPORT variable will allow the API to set the
#       attribute anyway, except if the shop is allowed to change the attribute
#       and did actually change it for this product. Let's hope this will never
#       happen.
#
        local $DE_EPAGES::MultiStoreCommon::API::Synchronization::IS_MULTISTORE_EXIMPORT = 1;

        $SuperProduct->set({
            'DefaultSubProduct' => $NewDefaultSubProduct,
        });
    }

    # delete assigned ShippingMethods
    my $ahKeys = GetPKeyProductShippingMethodsByProductID($ObjectID);
    DeleteProductShippingMethod($ObjectID, $_->{'ShippingMethodID'}) foreach @$ahKeys ;

    # delete prepayment values
    DeleteProductPrepayment($_) foreach @{GetPKeyProductPrepaymentsByProductID($ObjectID)};

    # delete product availability notification
    DeleteProductAvailabilityEmail($_) foreach @{GetPKeyProductAvailabilityEmailsByProductID($ObjectID)};

    # remember the shop
    my $Shop = $self->getSite();
    my $IsAppointment = $self->get('IsAppointment');
    my $IsSubProduct = $self->isSubProduct;

    #remember deleted product GUID, path and datetime
    $self->setDeleted('Product');

    $self->SUPER::delete;

    my $Feature = $IsAppointment ? 'Appointments' : 'Products';

    # decrement the usage counter of the Products feature
    $self->decrementFeature({
        'Shop' => $Shop,
        'Feature' => $Feature,
        'IsSubProduct' => $IsSubProduct
    });

    return;
}

#========================================================================================
# �function     deleteFiles
# �state        private
#----------------------------------------------------------------------------------------
# �syntax       $Product->deleteFiles;
#----------------------------------------------------------------------------------------
# �description  Deletes the directories PublicPath and PrivatePath unless the product is
#               a variation sub product.
#========================================================================================
sub deleteFiles {
    my $self = shift;

    # delete files if object is a product
    my $ShopPath = $self->get('Site')->get('PublicPath');
    my $Path = $self->get('FileAlias');

    CleanPath($ShopPath, $Path, {'force' => 1});

    $self->SUPER::deleteFiles unless $self->superProduct;
    return;
}

#========================================================================================
# �function     set
# �state        private
#----------------------------------------------------------------------------------------
# �syntax       $Product->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# �description  This function triggers hooks 'OBJ_BeforeUpdateProduct' and 'OBJ_AfterUpdateProduct'.
#----------------------------------------------------------------------------------------
# �input        $hValues | attribute name => value hash | ref.hash
# �input        $LanguageID | (optional; required for localized attributes) language id | int
# �input        $hAdditionalOptions | (optional) additional options to set | ref.hash
#----------------------------------------------------------------------------------------
# �hook         OBJ_BeforeUpdateProduct | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ProductID     - product identifier   - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# �hook         OBJ_AfterUpdateProduct | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ProductID     - product identifier   - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID, $hAdditionalOptions) = @_;

    my $SuperProduct = $self->superProduct;
    delete $hValues->{'SuperProduct'} if ($hValues->{'SuperProduct'} && defined $SuperProduct) && ($hValues->{'SuperProduct'}->id == $SuperProduct->id);
    Error('NoChangeSuperProduct',{'ProductID'=>$self->id}) if $hValues->{'SuperProduct'};

    my $OldClass = $self->class;
    my $OldClassID = $OldClass->id;
    my $NewClassID = $OldClassID;

    if( $hValues->{'Class'} ) {
        if ( $hValues->{'Class'}->id == $OldClassID ) {
            delete $hValues->{'Class'}
        } elsif ( 0 < scalar @{$OldClass->getObjects('BySuperProductID', [$self->id])} ) {
            delete $hValues->{'Class'};
            warn 'can\'t change class of variation product: '.$self->get('Path');
        } else {
            $NewClassID = $hValues->{'Class'}->id;
        }
    }

    TriggerHook('OBJ_BeforeUpdateProduct', {'Object'=>$self, 'Values'=> $hValues, 'ProductID' => $self->id, 'LanguageID' => $LanguageID});

    # reset attributes of old product type only when Class was changed
    # and the old one was not the shop's default product type
    if( $OldClassID != $NewClassID &&
        $OldClassID != $self->getSite()->get('ProductClass')->id )
    {
        my %Values = map { $_->alias => undef } @{$self->class->directAttributes};
        $self->set( \%Values, $_->{'LanguageID'} ) foreach @{$self->getSite()->get('Languages')};
    }

    #delete selectedvariations if producttype was change
    $hValues->{'SelectedVariations'} = [] if $OldClassID != $NewClassID ;

    # modify attribute values of sub products
    # if we have a unity subproduct we do not inherit from the master
    if ($self->getSite()->get('IsUnityRegistered')){
        $self->_setUnitySubProductAttributeValues($hValues, $LanguageID) if $self->isMaster;
        $self->_resetUnitySubProductAttributeValues($hValues, $LanguageID, $SuperProduct) if $SuperProduct;
    }else {
        $self->_resetSubProductAttributeValues($hValues, $LanguageID, $SuperProduct) if $SuperProduct;
    }

    #set LastUpdate if any related attribute touched
    $self->_updateProfiles($hValues, $LanguageID);

    $self->SUPER::set($hValues, $LanguageID, $hAdditionalOptions);
    TriggerHook('OBJ_AfterUpdateProduct',  {'Object'=>$self, 'Values'=> $hValues, 'ProductID' => $self->id, 'LanguageID' => $LanguageID});

    #Trigger the stocklevel hook for all bundle products that have the current product included
    if( defined $hValues->{'StockLevel'}) {
        foreach my $BundleProduct (@{$self->get('IncludedInBundles')})
        {
            my $hValuesForBundleProduct = {
                'StockLevel' => $BundleProduct->get('StockLevel')
            };
            TriggerHook('OBJ_AfterUpdateProduct',  {'Object'=>$BundleProduct, 'Values'=> $hValuesForBundleProduct, 'ProductID' => $BundleProduct->id, 'LanguageID' => $LanguageID});
        }
    }

    if ($self->getSite()->get('IsUnityRegistered')){
        # update JSON attribute ImageInformations on product by using the
        # Object->addImageInformations from ContentObject
        # get Images for ImageInformations attribute from Constants for Product
        $self->addImageInformations([UNITY_IMAGE_ATTRIBUTES], $hValues);

        # we need to extend the attribute ImageInformations with additional informations
        # for the ImagesSlideShowString. We have to add all image formats to this ImagesSlideShowString hash
        $self->extendImageInformationsSlideShowString($hValues);

    }

    return;
}

#========================================================================================
# �function     extendImageInformationsSlideShowString
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Object->extendImageInformationsSlideShowString($hValues);
#----------------------------------------------------------------------------------------
# �description  add/update SlideShowString image informations on attribute ImageInformations.
#----------------------------------------------------------------------------------------
# �input        $hValues | changed object values or undef for resetting | ref.hash
#========================================================================================
sub extendImageInformationsSlideShowString {
    my $self = shift;
    my ($hValues) = @_;

    return if (defined $hValues) and (not $hValues->{'ImagesSlideShowString'});

    my $hExistingImageInformations = $self->get('ImageInformations');
    return unless exists $hExistingImageInformations->{'ImagesSlideShowString'};
   
    my $ObjectFilePath =  $self->get('PublicPath');
    # loop over all SlideShow files and add extra file informations for all other sizes (s,m,xs...)
    foreach my $Element (keys(%{$hExistingImageInformations->{'ImagesSlideShowString'}})) {
        # loop over all sizes and get informations
        foreach my $Suffix (keys %{ PRODUCT_IMAGE_SIZE_NAMES() }){
          
            # try to find the corresponding files by naming convention, e.g. "*_s*"
            my ($FileName, $Extension) = ($Element =~ /(.*)(\.[^.]+)$/x);
            $FileName = $FileName . '_' . $Suffix . $Extension;
            $FileName = $Element if  $Suffix eq 'l';
            my $hFileInfo = $self->getImageInfo($ObjectFilePath,$FileName);
            $hExistingImageInformations->{'ImagesSlideShowString'}->{$Element}->{'Sizes'}->{PRODUCT_IMAGE_SIZE_NAMES->{$Suffix}} = $hFileInfo;
        }
    }
    $self->set({'ImageInformations' => $hExistingImageInformations });
}

# set attributes on sub products if changed on the master product
sub _setUnitySubProductAttributeValues {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    my $aSubProducts = $self->get('SubProducts');
    return unless scalar @$aSubProducts;

    my $hNewSubProductValues = {};
    my $hLocNewSubProductValues = {};

    # find changed elements for updating sub products
    foreach my $Key (keys %$hValues) {
        # ignore attributes we will not update on sub products
        next if ElementOf($Key,(UNITY_NOT_UPDATE_ATTRIBUTES_ON_SUBS_ON_UPDATE_MASTER,UNITY_IMAGE_ATTRIBUTES));

        if(ElementOf($Key,UNITY_ATTRIBUTES_FROM_MASTERPRODUCT_WRITTEN_ON_SUBS_ON_CREATE)){
            # collect attibutes to update on sub products
            $hNewSubProductValues->{$Key} = $hValues->{$Key};
        }elsif(ElementOf($Key,UNITY_LOC_ATTRIBUTES_FROM_MASTERPRODUCT_WRITTEN_ON_SUBS_ON_CREATE)) {
            # update localized attributes on sub products
            $hLocNewSubProductValues->{$Key} = $hValues->{$Key};
        }
    }

    # set changed attributes on each sub product
    foreach my $SubProduct (@$aSubProducts) {
        $SubProduct->set($hNewSubProductValues) if scalar( keys(%$hNewSubProductValues) );
        $SubProduct->set($hLocNewSubProductValues,$LanguageID) if scalar ( keys(%$hLocNewSubProductValues) );
    }
}

# set Unity values to undef if they can be inherited from the master product
sub _resetUnitySubProductAttributeValues {
    my $self = shift;
    my ($hValues, $LanguageID, $SuperProduct) = @_;

    # collect inheriatble values
    my $hInheritableValues = {};
    foreach my $Key (keys %$hValues) {
        $hInheritableValues->{$Key} = $hValues->{$Key} if ElementOf($Key, UNITY_INHERITABLE_ATTRIBUTES);
    }

    # reset values
    if (scalar %$hInheritableValues) {
        $self->_resetSubProductAttributeValues($hInheritableValues, $LanguageID, $SuperProduct);
        $hValues->{$_} = $hInheritableValues->{$_} foreach keys %$hInheritableValues;
    }
}

# set values to undef if they can be inherited from the master product
sub _resetSubProductAttributeValues {
    my $self = shift;
    my ($hValues, $LanguageID, $SuperProduct) = @_;

    $hValues->{'TaxClass'} = undef if (defined $hValues->{'HasSubOwnPrices'} && !$hValues->{'HasSubOwnPrices'});
    my %CompareKeys; # attribute alias => type name
    foreach my $Key (keys %$hValues) {
        next unless defined $hValues->{$Key};
        next if ($Key eq 'StockLevel' || $Key eq '_StockLevel');

        my $Attribute = $self->class->attribute($Key);
        next if $Attribute->get('IsMandatory');
        next if $Attribute->isArray;

        my $Type = $Attribute->type;
        next if $Type =~ /^Boolean|JSON|Hash|ProductPrice|PreDefMultiString|PreDefMultiLocalizedString|PreDefCustomerString$/; ## no critic qw(ProhibitComplexRegexes)

        $CompareKeys{$Key} = $Type;
    }

    my @CompareKeys = keys %CompareKeys;
    return if 0 == scalar @CompareKeys;

    my $hSuperValues = $SuperProduct->get( \@CompareKeys, $LanguageID );
    foreach my $Key (@CompareKeys) {
        my $SuperValue = $hSuperValues->{$Key};
        next unless defined $SuperValue;
        next if $SuperProduct->isVariationAttribute( $Key );
        my $SubValue = $hValues->{$Key};

        my $Type = $CompareKeys{$Key};

        # normalize newlines before Compare()
        if( $Type eq 'String' || $Type eq 'LocalizedString' ) {
            # convert line endings to Unix style, because values read from the database also
            # use Unix style line endings
            s/\x{0D}?\x{0A}/\x{0A}/gs foreach( $SuperValue, $SubValue );
        }

        my $Equal = ( 0 == Compare($Type, $SubValue, $SuperValue) );
        $hValues->{$Key} = undef if $Equal;
    }
}

#set LastUpdate if any related attribute touched
sub _updateProfiles {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    if ( exists $hValues->{'_StockLevel'} && 0 != Compare('Float', $hValues->{'_StockLevel'}, $self->get('_StockLevel')) ) {
       my $StockLevelProfile = LoadClassByAlias('Product')->lastUpdateProfile('StockLevel');
       $self->setUpdated( $StockLevelProfile );
       $_->setUpdated( $StockLevelProfile ) foreach @{ $self->get('IncludedInBundles') };
    }

    my @ContentAttributes = qw( Description Text Name Manufacturer ImageSmall ImageMedium ImageLarge ImageHotDeal );
    if (my @UpdatedAttributes = grep { exists $hValues->{$_} } @ContentAttributes) {
        my $hOldValues = $self->get(\@UpdatedAttributes, $LanguageID);
        if ( grep { Compare('String', $hValues->{$_}, $hOldValues->{$_}) != 0} @UpdatedAttributes) {
            my $ContentProfile = LoadClassByAlias('Product')->lastUpdateProfile('Content');
            $self->setUpdated( $ContentProfile );
        }
    }

    # Save last update of StockLevel
    if ( exists $hValues->{'StockLevel'} && 0 != Compare('Float', $hValues->{'StockLevel'}, $self->get('StockLevel')) ) {
       my $RestApiStockLevelProfile = LoadClassByAlias('Product')->lastUpdateProfile('RestApiStockLevel');
       $self->setUpdated( $RestApiStockLevelProfile );

       #TODO: update bundle products as well?
       #$_->setUpdated( $RestApiStockLevelProfile ) foreach @{ $self->get('IncludedInBundles') };
    }

    # Save last update of one of the attributes we return in the REST API.
    my @RestApiAttributes = qw( Alias Name Description Text Manufacturer UPCEAN MainCharacteristics Keywords );
    if (my @UpdatedAttributes = grep { exists $hValues->{$_} } @RestApiAttributes) {
        my $hOldValues = $self->get(\@UpdatedAttributes, $LanguageID);
        if ( grep { Compare('String', $hValues->{$_}, $hOldValues->{$_}) != 0} @UpdatedAttributes) {
            my $RestApiResponseAttributesProfile = LoadClassByAlias('Product')->lastUpdateProfile('RestApiResponseAttributes');
            $self->setUpdated( $RestApiResponseAttributesProfile );
        }
    }
}

#========================================================================================
# �function     variationAttributes
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $aAttributes = $SuperProduct->variationAttributes;
#----------------------------------------------------------------------------------------
# �description  Returns a list of variation attributes, like size and color.
#               Must be called on a super product, otherwise the return list will be empty.
#----------------------------------------------------------------------------------------
# �return       $aAttributes | list of variation attributes | ref.array.object
#========================================================================================
sub variationAttributes {
    my $self = shift;

    return [
        map { LoadObject( $_ ) }
        unique
        map { $_->{'AttributeID'} }
        @{ GetInfoSelectedVariationsByProductID($self->id) }
    ];
}

#========================================================================================
# �function     isVariationAttribute
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $IsVariationAttribute = $SuperProduct->isVariationAttribute( $AttributeAlias );
#----------------------------------------------------------------------------------------
# �description  Returns true if the given attribute is a variation attribute of this
#               super product.
#               Must be called on a super product, otherwise return always false.
#----------------------------------------------------------------------------------------
# �input        $AttributeAlias | attribute alias | string
# �return       $IsVariationAttribute | true if the attribute is a variation attribute | boolean
#========================================================================================
sub isVariationAttribute {
    my $self = shift;
    my ($AttributeAlias) = @_;

    return 1 if grep { $AttributeAlias eq $_->alias } @{ $self->variationAttributes };
    return 0;
}

#========================================================================================
# �function     isMaster
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $IsMaster = $Product->isMaster;
#----------------------------------------------------------------------------------------
# �description  Returns true if the product is a master product, i.e. has some selected
#               variation attribute values.
#----------------------------------------------------------------------------------------
# �return       $IsMaster | true if product is a master product | boolean
#========================================================================================
sub isMaster {
    return shift->{'Attributes'}{'_IsMaster'};
}

#========================================================================================
# �function     isSubProduct
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $IsSubProduct = $Product->isSubProduct;
#----------------------------------------------------------------------------------------
# �description  Returns true if the product is a sub product, i.e. has a master product.
#----------------------------------------------------------------------------------------
# �return       $IsSubProduct | true if product is a sub product | boolean
#========================================================================================
sub isSubProduct {
    return shift->superProduct ? 1 : 0;
}

#========================================================================================
# �function     hasVariationAttributeValue
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $HasValue = $SuperProduct->hasVariationAttributeValue( $PreDefValue );
#----------------------------------------------------------------------------------------
# �description  Returns true if the given attribute value is selected as for variations
#               Must be called on a super product, otherwise return always false.
#----------------------------------------------------------------------------------------
# �input        $PreDefValue | attribute value | DE_EPAGES::Object::API::Object::PreDefAttribute
# �return       $HasValue | true if the value is a valid variation attribute value | boolean
#========================================================================================
sub hasVariationAttributeValue {
    my $self = shift;
    my ($PreDefValue) = @_;

    return ExistsSelectedVariationByProductID($self->id, $PreDefValue->id);
}

#========================================================================================
# �function     superProduct
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $SuperProduct = $BaseObject->superProduct;
#----------------------------------------------------------------------------------------
# �description  Returns the super product of a variation product.
#               Returns undef if the product is a normal or master product.
#----------------------------------------------------------------------------------------
# �return       $SuperProduct | super product | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub superProduct {
    my $self = shift;

    my $SuperProduct = $self->{'SuperProduct'};
    if( defined $SuperProduct ) {
        return ref $SuperProduct ? $SuperProduct : undef;
    }

    my $SuperProductID = $self->{'Attributes'}{'SuperProductID'};
    if( defined $SuperProductID ) {
        $SuperProduct = LoadObject( $SuperProductID );
        $self->{'SuperProduct'} = $SuperProduct;
        weaken $self->{'SuperProduct'};
    }
    else {
        # $SuperProduct is still undefined
        $self->{'SuperProduct'} = 0;
    }

    return $SuperProduct;
}

#========================================================================================
# �function     subProducts
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $aSubProduct = $SuperProduct->subProducts;
#----------------------------------------------------------------------------------------
# �description  Returns the list of all sub products of a super product.
#----------------------------------------------------------------------------------------
# �return       $aSubProduct | list of sub products | ref.array.DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub subProducts {
    my $self = shift;

    return [ map { LoadObject( $_->{'ObjectID'} ) } @{ $self->_infoSubProducts } ];
}

#========================================================================================
# �function     findSubProducts
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $aSubProdustIDs = $SuperProduct->findSubProducts;
#----------------------------------------------------------------------------------------
# �description  Returns the product ids of all sub products of a super product.
#----------------------------------------------------------------------------------------
# �return       $aSubProdustIDs | list of product ids | ref.array.int
#========================================================================================
sub findSubProducts {
    my $self = shift;

    return $self->class->findObjects('BySuperProductID', [ $self->id ]);
}

#========================================================================================
# �function     countSubProducts
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Count = $SuperProduct->countSubProducts;
#----------------------------------------------------------------------------------------
# �description  Returns the number of all sub products of a super product.
#----------------------------------------------------------------------------------------
# �return       $Count | number of sub products | int
#========================================================================================
sub countSubProducts {
    my $self = shift;

    return scalar @{ $self->findSubProducts };
}

sub _infoSubProducts {
    my $self = shift;

    return [
        sort {$a->{'Position'} <=> $b->{'Position'} }
        @{$self->class->infoObjects('BySuperProductID', [ $self->id ])}
    ];
}

sub importXML {
    my ($ObjectPackage, $Handler, $hElement, $Class, $hAttributes) = @_;
    my $ProductTypeAlias;
    if (not $Handler->topID('IsReference')) {

        my $Parent = $Handler->topID('Object');
        my $Shop = $Parent->getSite();

        # map TaxClass name to TaxClass object
        if (exists $hAttributes->{'TaxClass'}) {
            my $TaxMatrix = $Shop->get('TaxMatrix');
            $hAttributes->{'TaxClass'} = $TaxMatrix->child($hAttributes->{'TaxClass'});
        }

        # map TaxClassDigital name to TaxClassDigital object
        if (exists $hAttributes->{TaxClassDigital}) {
            my $TaxMatrix = $Shop->get('TaxMatrixDigital');
            $hAttributes->{TaxClassDigital} = $TaxMatrix->child($hAttributes->{TaxClassDigital});
        }

        # set the class to the product type that is defined in this shop
        if( exists $hAttributes->{ProductType} ) {
            $Class = $Shop->get('ProductTypes')->child( $hAttributes->{ProductType} );
            $ProductTypeAlias = $hAttributes->{ProductType};
            delete $hAttributes->{ProductType};
        }
        else {
            $Class = $Shop->get('ProductClass');
        }

        #test if se have a sub product and
        #set class to class of super product
        if ( $Parent->alias eq 'SubProducts' ) {
            $Class = $Parent->parent->class;
            $hAttributes->{'SuperProduct'} = $Parent->parent unless $Parent->findChild($hAttributes->{'Alias'});
        }
    } else {
        my $Parent = $Handler->topID('Object');
        my $ObjectID = $Parent->findXML($hAttributes);
        $ProductTypeAlias = $hAttributes->{ProductType};
        $Class = LoadObject($ObjectID)->class if defined $ObjectID;
    }
    my $Product = $ObjectPackage->SUPER::importXML($Handler, $hElement, $Class, $hAttributes);
    if (!defined $Product->superProduct && $Product->class->id == $Product->getSite()->get('ProductClass')->id) {
        if( defined $ProductTypeAlias && $ProductTypeAlias ne 'ProductClass' ) {
            $Class = $Product->getSite()->child('ProductTypes')->child( $ProductTypeAlias );
            $Product->set({'Class' => $Class});
        }
    } elsif (defined $ProductTypeAlias && $Product->class->alias ne $ProductTypeAlias) {
        Error('WRONG_PRODUCTCLASS', {
            Product => $Product->alias,
            ProductClass => $Product->class->alias,
            ProductType => $ProductTypeAlias
        });
    }
    return $Product;
}

#========================================================================================
# �function     exportXMLAttributes
# �state        private
#----------------------------------------------------------------------------------------
# �syntax       $hAttributes = $Product->exportXMLAttributes( $Driver, $hAllAttributes );
#----------------------------------------------------------------------------------------
# �description  Returns plain attributes of object to XML export driver. This function
#               removes bit attributes with value 0.
#----------------------------------------------------------------------------------------
# �input        $Driver | xml export driver with support function addObject | object
# �input        $hAllAttributes | all attributes of object,
#                                 see $Object-><function exportableAttributes> | ref.hash
# �return       $hAttributes | plain attributes | ref.hash.string
#========================================================================================
sub exportXMLAttributes {
    my $self = shift;
    my ($Driver, $hAttributes) = @_;
    #$Driver->{'TagName'} = 'Product';
    $hAttributes = $self->SUPER::exportXMLAttributes($Driver, $hAttributes);

    # map TaxClass/TaxClassDigital object to class alias
    foreach my $Alias ( qw (TaxClass TaxClassDigital) ) {
        if (exists $hAttributes->{$Alias}) {
            my $TaxClass = $self->getWithParent($Alias);
            $hAttributes->{$Alias} = defined $TaxClass ? $TaxClass->alias : undef;
        }
    }

    if (exists $hAttributes->{'Class'}) {
        $hAttributes->{'ProductType'} = $self->class->alias;
        delete $hAttributes->{'ProductType'} if $hAttributes->{'ProductType'} eq 'ProductClass'; # dont export standard class
        delete $hAttributes->{'ProductType'} if defined $self->superProduct; # dont export product type for sub products
        delete $hAttributes->{'Class'};
    }
    return $hAttributes;
}

#========================================================================================
# �function     exportXMLElements
# �state        private
#----------------------------------------------------------------------------------------
# �syntax       $Product->exportXMLElements( $Driver, $hAttributes );
#----------------------------------------------------------------------------------------
# �description  Exports the via XML export driver the given attributes.
#----------------------------------------------------------------------------------------
# �input        $Driver | xml export driver with support function addObject | object
# �input        $hAttributes | attributes of object,
#                                 see $Object-><function exportableAttributes> | ref.hash
#========================================================================================
sub exportXMLElements {
    my $self = shift;
    my ($Driver, $hAttributes) = @_;

    if ($self->class->attribute('SelectedVariations')->get('ExportLevel') == $Driver->exportLevel) {
        $Driver->addElement( 'SelectedVariation', { 'SelectedVariation' => $_ } ) foreach @{$self->get('SelectedVariations')};
    }
    $self->SUPER::exportXMLElements($Driver, $hAttributes);
}

#========================================================================================
# �function     taxMatrix
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $TaxMatrix = $Product->taxMatrix;
# �example      $TaxMatrix = $Product->taxMatrix($User);
#----------------------------------------------------------------------------------------
# �description  Returns the tax matrix (digital) of shop of product.
#----------------------------------------------------------------------------------------
# �input        $User | user object | DE_EPAGES::Permission::API::Object::User (optional)
# �return       $TaxMatrix | tax matrix (digital) | DE_EPAGES::Tax::API::Object::TaxMatrix[Digital]
#========================================================================================
sub taxMatrix {
    my $self = shift;
    my ($User) = @_;
    my $Shop = $self->getSite;

    my $IsDigital = defined $self->get('TaxClassDigital') ? 1 : 0;
    my $TaxMatrixDigital = $Shop->get('TaxMatrixDigital');


    # check if this is a digital product and the user resides in the EU
    # else, return the standard tax matrix

    if ($IsDigital and defined $User) {
        #--- prio 1: check basket address
        my $Basket = defined $User->get('Baskets') ? $User->get('Baskets')->[0] : undef;
        if ($Basket and defined $Basket->taxAreaDigitalByAddress) {
            return $TaxMatrixDigital;
        }

        #--- prio 2: check user address
        #--- prio 3: check shop address
        my $Address = $User->get('ShippingAddress') // $User->get('BillingAddress') // $Shop->get('Address');
        if (_IsAddressDigital($TaxMatrixDigital, $Address)) {
            return $TaxMatrixDigital;
        }
    }
    elsif ($IsDigital) {
        #--- prio 3: check shop address
        if (_IsAddressDigital($TaxMatrixDigital, $Shop->get('Address'))) {
            return $TaxMatrixDigital;
        }
    }
    # find the correct Matrix on the system by the shipping address if delivery threshold is active
    elsif ($Shop->get('DeliveryThresholdIsActive') and defined $User) {
        my $Address = $User->get('ShippingAddress') // $User->get('BillingAddress') // $Shop->get('Address');
        if(IsEUAddress($Address)) {
            my $TaxMatrix = TaxMatrixByEUAddress($Address);
            return $TaxMatrix;
        }
    }

    return $Shop->get('TaxMatrix');
}

#========================================================================================
sub _IsAddressDigital {
    my ($TaxMatrixDigital, $Address) = @_;

    if (defined $Address) {
        if ($Address->get('IsEU') and defined $TaxMatrixDigital->taxAreaDigitalByCountry($Address->get('CountryID'))) {
            return 1;
        }
    }

    return 0;
}

#========================================================================================
# �function     price
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Price = $Product->price( $hVars );
#----------------------------------------------------------------------------------------
# �example      $Price = $Product->price( {
#                   CurrencyID => 'USD',
#                   TaxModel => $Product->getSite()->get('TaxModel')
#               } );
#----------------------------------------------------------------------------------------
# �description  Returns the product price by the given requirements.
#               The hook 'ProductPrice' is triggered before the standard price list is used.
#----------------------------------------------------------------------------------------
# �input        $hVars | <ul>
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price   - int
#                   <li>Quantity      - desired quantiy (optional, e.g. MinOrder) - float
#                   <li>Container     - line item container (optional)- LineItemContainer
#                   <li>LineItem      - product line item   (optional)- LineItemProduct
#                   <li>Customer      - customer            (optional)- Customer
#                   <li>User          - user                (optional)- User
#                   <li>IsBundled     - is part of a bundle (optional)- Boolean
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
# �return       $Price | price of product | money
#========================================================================================
sub price {
    my $self = shift;
    my ($hVars) = @_;

    my $RawPrice = $self->_getRawPrice($hVars);
    Error('ProductNoPrice', {'ProductID'=>$self->id}) unless defined $RawPrice;

    my $MinorUnit = GetInfoCurrency($hVars->{'CurrencyID'})->{'MinorUnit'};
    return $FLOAT_FORMATTER->round($RawPrice, $MinorUnit);
}

# return the price without rounding
# return undef if there is no price
sub _getRawPrice {
    my $self = shift;
    my ($hVars) = @_;

    my $PriceQuantity = $self->getWithParent('PriceQuantity');
    return undef unless $PriceQuantity;

    my %Vars = (
        %$hVars,
        'Object'    => $self,
        'ProductID' => $self->id,
        'Price'     => undef,
    );
    TriggerHook('ProductPrice', \%Vars);
    my $Price = $Vars{'Price'} // $self->getListPrice($hVars->{'CurrencyID'}, $hVars->{'TaxModel'}, undef, $hVars->{'User'},$hVars->{'TaxMatrix'});
    return $Price if defined $Price;

    #if we have a BundleProduct with an empty price set the price to 0
    #otherwise we can not add the whole bundle to the basket!
    return 0 if $hVars->{'IsBundled'};
    return undef;
}

#========================================================================================
# �function     hasPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $HasPrice = $Product->hasPrice ( $CurrencyID );
# �syntax       $HasPrice = $Product->hasPrice( $CurrencyID, $hVars );
#----------------------------------------------------------------------------------------
# �example      if( $Product->hasPrice ) { ... }
# �example      if( $Product->hasPrice( 'EUR', { Customer => $Customer } ) ) { ... }
# �example      if( $Product->hasPrice( 'EUR', { User => $User } ) ) { ... }
#----------------------------------------------------------------------------------------
# �description  Returns true if the product has a price.
#               The result may depend on the customer or indirectly on the user.
#               If no user or customer is given, returns whether there is a price for
#               anonymous customers.
#               For master products, returns true if any of the sub products has a price
#               and exists a "from" price (see <function fromPrice>).
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency id, e.g. 'USD' | string
# �input        $hVars | <ul>
#                   <li>Customer      - customer            (optional)- Customer
#                   <li>User          - user                (optional)- User
#               </ul> | ref.hash
# �return       $HasPrice | true if the product has a price | boolean
#========================================================================================
sub hasPrice {
    my $self = shift;
    my ($CurrencyID, $hVars) = @_;
    TestString('CurrencyID', $CurrencyID, 3, 3);

    my $Customer = $hVars->{'Customer'};
    my $User = $hVars->{'User'};
    if( !defined $Customer && defined $User && !$User->get('IsAnonymous') ) {
        $Customer = GetCustomerForUser( $User, $self->getSite );
    }

    my %Vars = (
        'CurrencyID' => $CurrencyID,
        'TaxModel'   => TAX_MODEL_NET,
        'Customer'   => $Customer,
        'User'       => $User,
    );

    return (defined $self->_getRawPrice(\%Vars) || defined $self->fromPrice(\%Vars)) ? 1 : 0;
}

#========================================================================================
# �function     discount
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Discount = $Product->discount( $hVars );
#----------------------------------------------------------------------------------------
# �example      $Discount = $Product->discount( {
#                   CurrencyID => 'USD',
#                   TaxModel => $Product->getSite()->get('TaxModel')
#               } );
#----------------------------------------------------------------------------------------
# �description  Returns the product discount by the given requirements.
#               The hook 'ProductDiscount' is triggered before the standard discount = 0.
#----------------------------------------------------------------------------------------
# �input        $hVars | <ul>
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gros or net price    - int
#                   <li>Container     - Container           (optional)- LineItemContainer
#                   <li>LineItem      - LineItemProduct     (optional)- LineItemProduct
#                   <li>User          - user                (optional)- User
#                   <li>Customer      - customer            (optional)- Customer
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
# �return       $Discount | discount of product | money
#========================================================================================
sub discount {
    my $self = shift;
    my ($hVars) = @_;
    my %Vars = (
        %$hVars,
        'Object'    => $self,
        'ProductID' => $self->id,
        'Discount'  => undef,
    );
    TriggerHook('ProductDiscount', \%Vars);
    return $Vars{'Discount'} if defined $Vars{'Discount'};
    return 0;
}

#========================================================================================
# �function     correctQuantity
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $ValidQuantity = $Product->correctQuantity( $DesiredQuantity );
# �example      print $Object->correctQuantity( 17.5 );
#----------------------------------------------------------------------------------------
# �description  Checks if a given quantity of the product can be ordered according to the
#               products minimum order amount (MinOrder) and interval amount (IntervalOrder).
#               Returns the next valid quantity value if the desired amount cannot be
#               ordered.
#               This function does not check the current stock level.
#----------------------------------------------------------------------------------------
# �input        $DesiredQuantity | quantity desired by the user | float
# �return       $ValidQuantity | quantity allowed by the product attributes
#                               MinOrder and IntervalOrder | float
#========================================================================================
sub correctQuantity {
    my $self = shift;
    my ($Quantity) = @_;

    # test min order
    my $MinOrder = $self->getWithParent('MinOrder');
    return $MinOrder if $Quantity < $MinOrder;

    # if no order interval is given or its 0 then all steps allowed
    my $IntervalOrder = $self->getWithParent('IntervalOrder');
    return $Quantity if not defined $IntervalOrder or $IntervalOrder == 0;

    # test order interval
    my $IntervalCount = $FLOAT_FORMATTER->ceiling(($Quantity - $MinOrder) / $IntervalOrder);
    return $MinOrder + $IntervalCount * $IntervalOrder;
}

#========================================================================================
# �function     availableQuantity
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $AvailableQuantity = $Product->availableQuantity( $DesiredQuantity, $ExistingQuantity );
#----------------------------------------------------------------------------------------
# �description  Ckecks if the given quantity value is valid according to the products
#               MinOrder and IntervalOrder settings.
#               Then checks if the desired amount is actually on stock. If one order
#               multiple line items of the same product, pass the quantity of the other
#               line items as second parameter ($ExistingQuantity).
#               Returns the closest quantity that can actually be ordered.
#               Returns 0 if the product cannot be ordered.
#----------------------------------------------------------------------------------------
# �input        $DesiredQuantity | user wants quantity | float
# �input        $ExistingQuantity | quantity of other lineitems (optional default=0) | float
# �return       $AvailableQuantity | product allow quantity (MinOrder and IntervalOrder )| float
#========================================================================================
sub availableQuantity {
    my $self = shift;
    my ($UserQuantity, $ExistingQuantity)= @_;
    $ExistingQuantity = 0 unless defined $ExistingQuantity;

    my $CorrectedQuantity = $self->correctQuantity($UserQuantity);

    my $StockLevel = $self->get('StockLevel');
    return $CorrectedQuantity unless defined $StockLevel;

    my $NewStockLevel = $StockLevel - $ExistingQuantity;
    return $CorrectedQuantity if $CorrectedQuantity <= $NewStockLevel;

    my $ProductLowOnStockAction = $self->getSite()->get('ProductLowOnStockAction');
    if( $ProductLowOnStockAction == STOCK_IS_ORDERABLE ) {
        return $CorrectedQuantity;
    }
    # available are fewer as min order
    return 0 if $NewStockLevel < $self->getWithParent('MinOrder');

    #set max quantity to corrected stock value -> could be more than stock because of rounding or intervals
    $CorrectedQuantity = $self->correctQuantity($NewStockLevel);

    #set corrected quantity lesser than NewStockLevel if it isn't so
    my $IntervalOrder = $self->getWithParent('IntervalOrder');
    $CorrectedQuantity -= $IntervalOrder while $NewStockLevel < $CorrectedQuantity;

    return $CorrectedQuantity;
}

#========================================================================================
# �function     lineItem
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $hLineItem = $Product->lineItem( $Container, $hInfo );
#----------------------------------------------------------------------------------------
# �description  Returns the required attributes to create a product line item object,
#               which can be added to the line item container of a basket or order.
#               Use <function addToContainer> instead of calling this function directly.
#----------------------------------------------------------------------------------------
# �input        $Container | line item container | object
# �input        $hInfo | keys <ul>
#                   <li>Quantity      - quantity            (optional default=MinOrder)- float
#                   <li>CurrencyID    - alpha currency code (optional default=Container.CurrencyID) - char(3)
#                   <li>TaxModel      - gros or net price   (optional default=Container.TaxModel)  - int
#                   <li>IsChanged - quantity was changed by product (optional default=0)  - int
#                   <li>LineItem - existing line item object (optional) - object
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
# �return       $hLineItem |  hash for lineitem insert, keys <ul>
#                   <li>SKU
#                   <li>Name
#                   <li>TaxClass
#                   <li>Quantity
#                   <li>BasePrice
#                   <li>Discount
#                   <li>LineItemPrice
#                   <li>Product
#                   <li>TaxRate
#                   <li>IsChanged
#                   <li>OrderUnit
#                   <li>WeightUnit
#                   <li>Weight
#                   <li>VariationString
#                   <li>EcoParticipation
#                   <li>Deposit
#               </ul> | ref.hash
# �hook         ProductLineItemQuantity | hook parameter keys :
#               <ul>
#                 <li>Product - this object - DE_EPAGES::Product::API::Object::Product</li>
#                 <li>LineItem - existing line item (maybe undef) - DE_EPAGES::Product::API::Object::LineItemProduct</li>
#                 <li>LineItemContainer - line item container - DE_EPAGES::Product::API::Object::LineItemProduct</li>
#                 <li>OtherLineItemQuantity - quantiy of the same products in other line items of the container - int</li>
#                 <li>Quantity - desired quantity (this value can be changed by hook functions) - int</li>
#               </ul>
#========================================================================================
sub lineItem {
    my $self = shift;
    my ($Container, $hInfo) = @_;

    $hInfo->{'Container'} //= $Container;
    $hInfo->{'Quantity'} //= $self->getWithParent('MinOrder');
    my $WantedQuantity = $hInfo->{'Quantity'};

    $hInfo->{'User'} //= $Container->user;

    my $LineItem = $hInfo->{'LineItem'};
    my $LineItemsQuantity = $self->getQuantityInOtherLineItems($Container, $LineItem);

    # EPG-21475: allow cartridges to modify the desired quantity before the correction
    # regarding minimum order quantity, interval and stock level
    my %Vars = (
        Product => $self,
        LineItem => $LineItem,
        LineItemContainer => $Container,
        OtherLineItemQuantity => $LineItemsQuantity, # quantity of the same product in other line items
        Quantity => $WantedQuantity, # this value can be changed by the hook function
    );
    TriggerHook('ProductLineItemQuantity', \%Vars);

    $hInfo->{'Quantity'} = $self->availableQuantity($Vars{'Quantity'}, $LineItemsQuantity);

    $hInfo->{'CurrencyID'} //= $Container->get('CurrencyID');
    $hInfo->{'TaxModel'}   //= $Container->get('TaxModel');
    $hInfo->{'IsBundled'}  = $Container->parent->instanceOf('LineItemProduct');


    my $PostFix = $self->getWithParent('TaxClassDigital') ? 'Digital' : '';
    $hInfo->{'TaxMatrix'} = $Container->get('TaxMatrix' . $PostFix);

    # ugly hack for getting the correct address from the object
    # we can not use the user for getting the address in ->price
    $hInfo->{'User'} = $Container->parent if $self->getSite()->get('DeliveryThresholdIsActive');

    my $PriceQuantity = $self->getWithParent('PriceQuantity');

    # ability to patch lineItemProduct Prices in the basket.
    my $Price;
    if (defined $LineItem && $LineItem->get('IsPatched')) {
        $Price = $LineItem->get('BasePrice');
    } else {
        $Price = $self->price($hInfo);
    }

    my $Discount = $self->discount($hInfo);
    my $EcoParticipation = $self->getEcoParticipation($hInfo->{'CurrencyID'}, $hInfo->{'TaxModel'});

    my $LocaleID = $Container->get('LocaleID');
    my $Locale = $Container->getSite()->siteLocale($LocaleID);
    my $TimeFactor = $self->get('IsAppointment') && defined $LineItem ? $LineItem->get('TimeFactor') : 1;

    local $DE_EPAGES::Object::API::Language::LANGUAGEID = $Container->get('LanguageID');

    my $VariationString = $self->localizedVariationString($Container->get('LanguageID'));

    my $Appointment = $self->get('IsAppointment') && defined $LineItem ? $LineItem->get('Appointment') : undef;

    my $IsShippingRequired = $self->get('IsAppointment') || $self->get('IsDownloadProduct') ? 0 : 1;
    if ($self->get('IsBundleProduct')) {
        my @NonDownloadProducts = grep {!$_->get('IsDownloadProduct')} @{$self->get('BundledProducts')};
        $IsShippingRequired = 0 if 0 == scalar @NonDownloadProducts;
    }

    my $LineItemPrice = $self->_lineItemPrice($Appointment, $Price, $hInfo->{'Quantity'},
                                              $hInfo->{'CurrencyID'}, $PriceQuantity, $Discount,
                                              $TimeFactor, $hInfo->{'TaxModel'},
                                              $self->get('IsAppointment'));
    $LineItemPrice = $Locale->roundMoney($LineItemPrice, $hInfo->{'CurrencyID'});

    delete $hInfo->{'User'};

    my $ProductTaxClassAlias;
    my $TaxMatrixID = $Container->get('TaxMatrix')->id;
    my $Shop = $Container->getSite();
    if ($Shop->get('DeliveryThresholdIsActive')) {
        my $aProductTaxClassMapping = $self->get('TaxClassMapping');
        $ProductTaxClassAlias = _GetTaxAliasFromTaxClassMapping($aProductTaxClassMapping, $TaxMatrixID);
    }
    $ProductTaxClassAlias = defined $ProductTaxClassAlias ? $ProductTaxClassAlias : $self->getWithParent('TaxClass')->alias;
    my $TaxClassForInfo = $Container->get('TaxMatrix')->getTaxMatrixTaxClassByAlias($ProductTaxClassAlias);

    my %Info = (
        %$hInfo,
        'SKU'               => $self->alias,
        'Name'              => substr($self->get('NameOrAlias'), 0, LINEITEM_NAME_MAX_LENGTH),
        'TaxClass'          => $TaxClassForInfo,
        'Quantity'          => $hInfo->{'Quantity'},
        'BasePrice'         => $Price / $PriceQuantity,
        'Discount'          => $Discount,
        'LineItemPrice'     => $LineItemPrice,
        'Product'           => $self,
        'TaxRate'           => $Container->taxRate($TaxClassForInfo),
        'IsShippingRequired' => $IsShippingRequired,
        'Weight'            => $self->getWithParent('Weight'),
        'WeightUnit'        => $self->getWithParent('WeightUnit'),
        'OrderUnit'         => $self->getWithParent('OrderUnit'),
        'VariationString'   => $VariationString,
        'WeightQuantity'    => $self->getWithParent('PriceQuantity'),
        'EcoParticipation'  => defined $EcoParticipation ? $EcoParticipation / $PriceQuantity : undef,
        'MaxDownloadTime'   => $self->getWithParent('MaxDownloadTime'),
    );

    # do we have set the digital tax class on the product, use digital tax matrix
    if ($self->getWithParent('TaxClassDigital')) {
        $Info{'TaxClassDigital'} = $self->getWithParent('TaxClassDigital');
        $Info{'TaxRate'} = $Container->taxRateDigital($self->getWithParent('TaxClassDigital'));
        unless($hInfo->{'IsBundled'}){
            $Info{'TaxRate'} = 0 unless $Container->get('BasketOrOrder')->taxAreaDigitalByAddress();
        }
    }

    if ($self->get('IsAppointment')) {
        $Info{'Prepayment'} = $self->appointmentPrepayment($hInfo->{'Appointment'}, {
            'CurrencyID' => $hInfo->{'CurrencyID'},
            'TaxModel'   => $hInfo->{'TaxModel'},
            'Quantity'   => $hInfo->{'Quantity'},
        });
        my $OriginalLineItemPrice = $Locale->roundMoney($self->_lineItemPrice(
            $Appointment, $Price, $hInfo->{'Quantity'}, $hInfo->{'CurrencyID'},
            $PriceQuantity, $Discount, $TimeFactor, $hInfo->{'TaxModel'}, 0
        ), $hInfo->{'CurrencyID'});
        $Info{'OriginalLineItemPrice'} = $OriginalLineItemPrice if $OriginalLineItemPrice != $Info{'LineItemPrice'};
    }

    # set only if changed == 1 !!
    delete $Info{$_} foreach qw (CurrencyID TaxModel LineItem Container IsChanged IsBundled IsEditOrder TaxMatrix);
    $Info{'IsChanged'} = 1 if Compare('Float', $hInfo->{'Quantity'}, $WantedQuantity);
    return \%Info;
}

#========================================================================================
# �function     localizedVariationString
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $LocalizedString = $Product->localizedVariationString($LanguageID)
#----------------------------------------------------------------------------------------
# �description  Returns a single string with all variation attributes sorted by attribute name.
#               Example "Color: Croft, Material: Goretex, Size: S"
#               If method is called on non SubProduct undef is returned.
#----------------------------------------------------------------------------------------
# �input        $LanguageID | language used for localization | int
# �return       $LocalizedString | contains list of variation attributes | string or undef
#========================================================================================
sub localizedVariationString {
    my $self = shift;
    my ($LanguageID) = @_;
    my $SuperProduct = $self->superProduct();
    return undef unless defined $SuperProduct;

    my @Attributes = map { $_->{'Attribute'} } @{$self->getWithParent('VariationTypeValues')};

    return join(', ',
    map { $_->get('NameOrAlias', $LanguageID) . ': ' . $self->getWithParent($_->alias)->get('ValueOrAlias', $LanguageID) }
    sort { $a->get('Position') <=> $b->get('Position') } @Attributes);
}

sub _lineItemPrice {
    my $self = shift;
    my ($Appointment, $Price, $Quantity, $CurrencyID, $PriceQuantity, $Discount, $TimeFactor, $TaxModel, $IsAppointment) = @_;
    if ($IsAppointment) {
        my $LineItemPrice = $self->appointmentPrepayment($Appointment, {
            'CurrencyID' => $CurrencyID,
            'TaxModel'   => $TaxModel,
            'Quantity'   => $Quantity
        });
        return $LineItemPrice if defined $LineItemPrice && $LineItemPrice != 0;
    }
    return ($Price * ($Quantity / $PriceQuantity) - $Discount) * $TimeFactor;
}


#========================================================================================
# �function     getQuantityInOtherLineItems
#----------------------------------------------------------------------------------------
# �syntax       $Product->getQuantityInOtherLineItems($Container, $CurrentLineItem);
#----------------------------------------------------------------------------------------
# �description  returns the quantity of the product in the basket, excluding $CurrentLineItem
#----------------------------------------------------------------------------------------
# �input        $Container | lineitem container | object
# �input        $CurrentLineItem | line item | object
# �return       $Quantity | quantity of the given product in other line items | float
#========================================================================================
sub getQuantityInOtherLineItems {
    my $self = shift;
    my ($Container, $CurrentLineItem) = @_;

    my $ProductID = $self->id;
    my $CurrentLineItemID = defined $CurrentLineItem ? $CurrentLineItem->id : 0;
    my (@OtherProductLineItems) =
        grep { $_->id != $CurrentLineItemID }
        grep { $_->instanceOf('LineItemProduct') }
        @{ $Container->get('LineItems') };

    my $Quantity = 0;
    foreach my $LineItem ( @OtherProductLineItems ) {

        if ($LineItem->get('IsLineItemProductBundle')) {
            my $BundleQuantity = $LineItem->get('Quantity');
            foreach my $BundledLineItem ( @{$LineItem->get('BundledProductContainer')->get('LineItems')} ) {
                my $Product = $BundledLineItem->get('Product');
                if( defined $Product && $Product->id == $ProductID ) {
                    $Quantity += $BundledLineItem->get('Quantity') * $BundleQuantity;
                }
            }
        }

        my $Product = $LineItem->get('Product');
        if( defined $Product && $Product->id == $ProductID ) {
            $Quantity += $LineItem->get('Quantity');
        }
    }

    return $Quantity;
}

#========================================================================================
# �function     addBundleProductContainer
#----------------------------------------------------------------------------------------
# �syntax       $Product->addBundleProductContainer($LineItem);
#----------------------------------------------------------------------------------------
# �description  add a sub container for bundle products to a given line item
#----------------------------------------------------------------------------------------
# �input        $LineItem | line item | object
# �return       $BundledProductContainer | line item container | object
#========================================================================================
sub addBundleProductContainer {
    my $self = shift;
    my ($LineItem) = @_;

    #add container for bundled products
    my %ContainerInfo = ( 'Alias' => 'BundledProductContainer' );
    my $Container = $LineItem->container;
    $ContainerInfo{$_} = $Container->get($_) foreach qw ( LocaleID TaxArea CurrencyID TaxModel);

    my $BundledProductContainer = $LineItem->insertChild('LineItemContainer', \%ContainerInfo);
    foreach my $BundleProductMap ( @{$self->get('BundleProductMaps')} ) {
        $BundleProductMap->get('BundledProduct')->addToContainer( $BundledProductContainer, {
            'Quantity' => $BundleProductMap->get('BundledQuantity'),
            'BasePrice' => 0,
            'Position' => $BundleProductMap->get('Position'),
        });
    }
    $LineItem->set({'BundledProductContainer' => $BundledProductContainer });
    return $BundledProductContainer;
}

#========================================================================================
# �function     addToContainer
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $LineItem = $Product->addToContainer( $Container, $hInfo );
# �example      Transaction( sub {
#                   my $Basket = GetBasket( $User, $Shop );
#                   my $LineItem = $Product->addToContainer( $Basket->container, {
#                       Quantity => 3
#                   } );
#                   $Basket->recalculate;
#               } );
#----------------------------------------------------------------------------------------
# �description  Adds the product to the line item container (of a basket or order).
#               If a line item of the same product already exists in the basket, then the
#               function acts according to the shop preference Shop.AddToBasketStyle:
#               <ul><li>a new line item is created or</li>
#                   <li>the quantity is added to the existing line item or</li>
#                   <li>the new line item replaces the existing line item or</li>
#                   <li>the previous quantity remains and the product is not added</li>
#               </ul>
#----------------------------------------------------------------------------------------
# �input        $Container | line item container | object
# �input        $hInfo | keys <ul>
#                   <li>Quantity      - quantity            (optional default=MinOrder)- float
#                   <li>CurrencyID    - alpha currency code (optional default=Container.CurrencyID) - char(3)
#                   <li>TaxModel      - gros or net price   (optional default=Container.TaxModel)  - int
#                   <li>AddProductStyle  - basket action for new added product (optional default=Shop.AddProductStyle) - int
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
# �return       $LineItem | line item | object
#========================================================================================
sub addToContainer {
    my $self = shift;
    my ($Container, $hInfo) = @_;

    my $TaxClass = $self->getWithParent('TaxClass');
    Error('ProductNoTaxClass', {'ProductID'=>$self->id}) if not defined $TaxClass;

    # get line items which are equal to product
    my $LineItems = $Container->get('LineItems');
    my @SameProductLineItems = grep { $_->instanceOf('LineItemProduct') and defined $_->get('Product') and $_->get('Product')->id == $self->id } @$LineItems;
    my $LineItem = scalar( @SameProductLineItems) ? $SameProductLineItems[0] : undef;

    # use the basket behavior from the info hash otherwise use shop settings
    my $AddProductStyle = defined $hInfo->{'AddProductStyle'} ? $hInfo->{'AddProductStyle'} : $self->getSite()->get('AddProductStyle');
    delete $hInfo->{'AddProductStyle'};

    #if line items doesnt exists add else add quantity
    if (not defined $LineItem or $AddProductStyle == BASKETACTION_ADDNEWPOSITION) {
        # test if we can add the lineitem
        my $hlineItemInfo = $self->lineItem($Container, $hInfo);
        if ( $hlineItemInfo->{'Quantity'} > 0 ) {
            my $Class = $self->get('IsAppointment')
                ? LoadClassByAlias('LineItemProductAppointment')
                : LoadClassByAlias('LineItemProduct');
            $LineItem = $Container->insertLineItem($Class, $hlineItemInfo );
            $self->addBundleProductContainer( $LineItem ) if $self->get('IsBundleProduct');

            my $LineItemDeposit = $self->addLineItemDeposit($Container, $hlineItemInfo, $LineItem->alias, $LineItem->get('Position'));
            $LineItem->set({ 'LineItemDeposit' => $LineItemDeposit });
        } else {
            Error('ProductCanNotAddToBasket');
        }
    } else {
        # add, change or ignore quantity like the settings for basketaction
        my $Quantity = $hInfo->{'Quantity'};
        my $LineItemQuantity = $LineItem->get('Quantity');
        if ($AddProductStyle == BASKETACTION_ADDQUANTITY) {
            my $IntervalOrder = $self->getWithParent('IntervalOrder');
            $Quantity = $IntervalOrder if not defined $Quantity or $Quantity < $IntervalOrder;
            $hInfo->{'Quantity'} = $Quantity + $LineItemQuantity;
        } elsif ($AddProductStyle == BASKETACTION_REPLACEQUANTITY) {
            $Quantity = $self->getWithParent('MinOrder') unless defined $Quantity;
            $hInfo->{'Quantity'} = $Quantity;
        } elsif ($AddProductStyle == BASKETACTION_DONTCHANGEQUANTITY) {
            $hInfo->{'Quantity'} = $LineItemQuantity;
        }
        $hInfo->{'LineItem'} = $LineItem;
        $LineItem->set($self->lineItem($Container, $hInfo) );
        my $LineItemDeposit = $LineItem->get('LineItemDeposit');
        $LineItemDeposit->set({'Quantity' => $hInfo->{'Quantity'}}) if defined $LineItemDeposit;
        Error('ProductCanNotChangedAtBasket') if $hInfo->{'IsChanged'};
    }

    my $LineItemEcoParticipation;
    my $EcoParticipation = $LineItem->get('EcoParticipation');
    if ($EcoParticipation) {
        foreach my $LineItemStatic (@ {$Container->get('Statics')}) {
            next unless $LineItemStatic->instanceOf('LineItemEcoParticipation');
            $LineItemEcoParticipation = $LineItemStatic if $LineItemStatic->get('TaxClass')->id eq $LineItem->get('TaxClass')->id;
        }
        if (defined $LineItemEcoParticipation) {
            $LineItemEcoParticipation->recalculate($Container);
        } else {
            $self->addLineItemEcoParticipation($Container, $LineItem->get('TaxClass'), $LineItem->get('TaxRate'), $EcoParticipation);
        }
    }
    return $LineItem;
}

#========================================================================================
# �function     addLineItemEcoParticipation
#----------------------------------------------------------------------------------------
# �syntax       $LineItemEcoParticipation = $Container->addLineItemEcoParticipation($Container, $TaxClass);
# �syntax       $LineItemEcoParticipation = $Container->addLineItemEcoParticipation($Container, $TaxClass, $TaxRate);
# �syntax       $LineItemEcoParticipation = $Container->addLineItemEcoParticipation($Container, $TaxClass, $TaxRate, $LineItemPrice);
#----------------------------------------------------------------------------------------
# �description  Adds a eco Particpation lineitem for the given tax class, if not exists yet.
#----------------------------------------------------------------------------------------
# �input        $Container | lineitem container | object
# �input        $TaxClass | tax class | object
# �input        $TaxRate | (optional) sets the tax rate of the line item,
#                   overrides the tax calculation | float
# �input        $LineItemPrice | (optional) sets the line item price,
#                   overrides the price calculation | float
# �return       $LineItemParticipation | line item eco participation | object
#========================================================================================
sub addLineItemEcoParticipation {
    my $self = shift;
    my ($Container, $TaxClass, $TaxRate, $LineItemPrice) = @_;
    $TaxRate = $Container->taxRate($TaxClass) unless defined $TaxRate;

    my $LineItemClass = LoadClassByAlias('LineItemEcoParticipation');
    local $DE_EPAGES::Object::API::Language::LANGUAGEID = $Container->get('LanguageID');
    my $LineItem = $Container->setLineItem($LineItemClass, {
        'TaxClass' => $TaxClass,
        'TaxRate' => $TaxRate,
        'Name' => $TaxClass->get('NameOrAlias'),
        'LineItemPrice' => $LineItemPrice,
    });
    return $LineItem;
}

#========================================================================================
# �function     addLineItemDeposit
#----------------------------------------------------------------------------------------
# �syntax       $LineItemDeposit = $Container->addLineItemDeposit($Container, $hInfo, $Alias, $Position);
#----------------------------------------------------------------------------------------
# �description  Adds a deposit lineitem for the given product lineitem.
#----------------------------------------------------------------------------------------
# �input        $Container | lineitem container| object
# �input        $hInfo | attributes from LineItemProduct | ref.hash
# �input        $Alias | alias of lineitemproduct | string
# �input        $Position | position of lineitemproduct | integer
# �return       $LineItemDeposit | line item deposit | object
#========================================================================================
sub addLineItemDeposit {
    my $self = shift;
    my ($Container, $hInfo, $Alias, $Position) = @_;
    return if $Container->parent->instanceOf('LineItemProduct'); # LineItem is in a bundle

    my $CurrencyID = defined $hInfo->{'CurrencyID'} ?  $hInfo->{'CurrencyID'} : $Container->get('CurrencyID');
    my $TaxModel = defined $hInfo->{'TaxModel'} ? $hInfo->{'TaxModel'} : $Container->get('TaxModel');
    my $Deposit = $self->getDeposit($CurrencyID, $TaxModel);
    return unless $Deposit;

    my $LineItemClass = LoadClassByAlias('LineItemDeposit');
    local $DE_EPAGES::Object::API::Language::LANGUAGEID = $Container->get('LanguageID');
    my $BasePrice = $Deposit / $hInfo->{'WeightQuantity'};
    my $LineItemDeposit = $Container->setLineItem($LineItemClass, {
        'LineItemPrice' => $BasePrice * $hInfo->{'Quantity'},
        'TaxClass' => $hInfo->{'TaxClass'},
        'TaxRate' => $hInfo->{'TaxRate'},
        'BasePrice' => $BasePrice,
        'Quantity' => $hInfo->{'Quantity'},
        'Alias' => $Alias . '_Deposit',
        'Product' => $hInfo->{'Product'},
        'OrderUnit' => $hInfo->{'OrderUnit'},
        'Position' => $Position,
        'PriceQuantity' => $hInfo->{'WeightQuantity'},
    });
    return $LineItemDeposit;
}

#========================================================================================
# �function     recalculateLineItem
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->recalculateLineItem( $LineItem, $hValues);
#----------------------------------------------------------------------------------------
# �description  Recalculate a product line item when something has changed on basket,
#               for example if a different tax area was chosen.
#----------------------------------------------------------------------------------------
# �input        $LineItem | line item | object
# �input        $hValues | possible keys which was changed <ul>
#                   <li>Quantity      - quantity            (optional default=MinOrder)- float
#                   <li>CurrencyID    - alpha currency code (optional default=Container.CurrencyID) - char(3)
#                   <li>TaxModel      - gros or net price   (optional default=Container.TaxModel)  - int
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
#========================================================================================
sub recalculateLineItem {
    my $self = shift;
    my ($LineItem, $hValues) = @_;

    my $Container = $LineItem->container;
    my %Info = %$hValues;
    $Info{'Quantity'}  = $LineItem->get('Quantity') unless exists $Info{'Quantity'};
    $Info{'LineItem'}  = $LineItem;
    $Info{'Container'} = $Container;


    # here we build a lineitem from the original product
    my $hProductLineItem = $self->lineItem($Container, \%Info);

    # do not use the original TaxClass of the product
    # we use the existing item tax class because we may have changed the TaxClass on the container
    if ($self->getSite()->get('DeliveryThresholdIsActive')){
        $hProductLineItem->{'TaxClass'} = $Container->get('TaxMatrix')->getTaxMatrixTaxClassByAlias($hProductLineItem->{'TaxClass'}->alias) ;
        $hProductLineItem->{'TaxRate'} = $Container->taxRate($hProductLineItem->{'TaxClass'});
    }

    $LineItem->set($hProductLineItem);

    my $LineItemDeposit = $LineItem->get('LineItemDeposit');
    if (defined $LineItemDeposit) {
        $LineItemDeposit->set({'Quantity' => $LineItem->get('Quantity')});
    }
    return;
}

#========================================================================================
# �function     getWithParent
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $hValues = $Product->getWithParent($aNames, $LanguageID);
# �syntax       $hValues = $Product->getWithParent($aNames);
# �syntax       $Values = $Product->getWithParent($Name, $LanguageID);
# �syntax       $Value = $Product->getWithParent($Name);
#----------------------------------------------------------------------------------------
# �example      my $LanguageID = GetPKeyLanguageByCode('en');
#               my $Description = $Product->getWithParent('Description', $LanguageID);
#----------------------------------------------------------------------------------------
# �description  Returns the values of product attributes. Similar to $Object->get().
#               If called on a sub product and the attribute value is not defined for the
#               sub product, then returns the attribute value of its super product.
#----------------------------------------------------------------------------------------
# �input        $aNames | list of attribute names | ref.array.string
# �input        $LanguageID | language id | integer
# �input        $Name | attribute name | string
# �return       $hValues | attribute name => value hash | ref.hash
# �return       $Values | attribute value | scalar
#========================================================================================
sub getWithParent {
    my $self = shift;
    my ($aNames, $LanguageID) = @_;
    $LanguageID = GetCurrentLanguageID() unless defined $LanguageID;

    my $SuperProduct = $self->superProduct;
    return $self->get($aNames, $LanguageID) unless defined $SuperProduct;

    # we do not inherit attributes for unity
    #my $IsUnityRegistered = $self->getSite()->get('IsUnityRegistered') ? 1 : 0;

    my $Value = $self->get($aNames, $LanguageID);
    if (ref $aNames eq 'ARRAY' ) {
        my @getSuperAttributes =();
        foreach my $key (keys %$Value ) {

            #next if $IsUnityRegistered and not ElementOf($key,@UNITY_FROM_MASTERPRODUCT_INHERITABLE_ATTRIBUTES_ON_GETWITHPARENT) ;

            my $AttributeValue = $Value->{$key};
            next if ref $AttributeValue ne 'ARRAY' and defined $AttributeValue or
                     ref $AttributeValue eq 'ARRAY' and scalar @$AttributeValue > 0;
            push @getSuperAttributes , $key;
        }
        if (scalar @getSuperAttributes ) {
            my $SuperValues = $SuperProduct->get(\@getSuperAttributes, $LanguageID);
            $Value->{$_} = $SuperValues->{$_} foreach (keys %$SuperValues );
        }
        return $Value;
    }
    else {
        # never get the Attribute from the SuperProduct
        return $Value if ElementOf($aNames,FROM_MASTERPRODUCT_NON_INHERITABLE_ATTRIBUTES_ON_GETWITHPARENT) ;

        return $Value if ref $Value ne 'ARRAY' and defined $Value  or
                         ref $Value eq 'ARRAY' and scalar @$Value > 0;

        return $SuperProduct->get($aNames, $LanguageID);
    }
}

#========================================================================================
# �function     addToBundle
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->addToBundle( $BundleProduct, $hInfo );
# �example      $Product->addToBundle( $BundleProduct, { BundledQuantity => 2 } );
#----------------------------------------------------------------------------------------
# �description  Adds this product to a product bundle.
#----------------------------------------------------------------------------------------
# �input        $BundleProduct | bundle product | object
# �input        $hInfo | keys <ul>
#                   <li>BundledQuantity - quantity (optional default=MinOrder)- float
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
#========================================================================================
sub addToBundle {
    my $self = shift;
    my ($BundleProduct, $hInfo) = @_;

    Error('ProductAndBundleProductEqual', {'ProductID'=>$self->id}) if $self->id == $BundleProduct->id;
    Error('NoBundleBundles', {'ProductID'=>$self->id}) if $self->get('IsBundleProduct');
    Error('NoBundleAppointments', {'ProductID'=>$self->id}) if $self->get('IsAppointment');
    # don't add $self if already assigned
    return if ElementOf($self, @{$BundleProduct->get('BundledProducts')});

    $hInfo->{'BundledQuantity'} = defined $hInfo->{'BundledQuantity'}
        ? $self->correctQuantity($hInfo->{'BundledQuantity'})
        : $self->getWithParent('MinOrder');

    my $MapClass = LoadClassByAlias('BundleProductMap');
    $MapClass->insertObject({'BundleProduct'  => $BundleProduct,
                             'BundledProduct' => $self,
                             %$hInfo});
    return;
}

#========================================================================================
# �function     renumberBundleProductMaps
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->renumberBundleProductMaps;
#----------------------------------------------------------------------------------------
# �description  Renumbers the bundled products assigned to a product.
#========================================================================================
sub renumberBundleProductMaps {
    my $self = shift;

    my $aMaps = $self->get('BundleProductMaps'); # sorted already
    my $position = POSITION_STEP;
    foreach my $Map(@$aMaps) {
        $Map->set({'Position', $position});
        $position += POSITION_STEP;
    }
    return;
}

#========================================================================================
# �function     updateBundleEcoParticipation
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->updateBundleEcoParticipation;
#----------------------------------------------------------------------------------------
# �description  Set ecoparticipation from the bundled products to a product.
#========================================================================================
sub updateBundleEcoParticipation {
    my $self = shift;
    my $aMaps = $self->get('BundleProductMaps');
    my %EcoParticipation;
    foreach my $Map(@$aMaps) {
        my $BundledProduct = LoadObject($Map->get('BundledProductID'));
        my $BundledQuantity = $Map->get('BundledQuantity');
            foreach my $hPrices (@{$BundledProduct->get('EcoParticipations')}) {
                $EcoParticipation{$hPrices->{'CurrencyID'}} += $hPrices->{'EcoParticipation'} * $BundledQuantity if defined $hPrices->{'EcoParticipation'};
            }
    }
    foreach my $CurrencyID (keys %EcoParticipation) {
        my $Value = $EcoParticipation{$CurrencyID};
        my @Prices = grep { $_->{'CurrencyID'} eq $CurrencyID } @{GetInfoProductPricesByProductID($self->id)};
        if (scalar @Prices) {
            if (defined $Value or defined $Prices[0]{'ListPrice'} or defined $Prices[0]{'ManufacturerPrice'}) {
                UpdateProductPrice({ %{$Prices[0]}, 'EcoParticipation'=>$Value});
            } else {
                DeleteProductPrice($Prices[0]);
            }
        } elsif(defined $Value) {
            InsertProductPrice({
                'ProductID'  => $self->id,
                'CurrencyID' => $CurrencyID,
                'EcoParticipation' => $Value
            });
        }
        if (defined $Value) {
            my $Category = LoadObjectByPath('/EcoParticipationCategories/1');
            $self->set({'EcoParticipationCategory' => $Category});
        }
    }
}

#========================================================================================
# �function     updateBundleDeposit
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->updateBundleDeposit;
#----------------------------------------------------------------------------------------
# �description  Set deposit from the bundled products to a product.
#========================================================================================
sub updateBundleDeposit {
    my $self = shift;
    my $aMaps = $self->get('BundleProductMaps');
    my %Deposit;
    foreach my $Map(@$aMaps) {
        my $BundledProduct = LoadObject($Map->get('BundledProductID'));
        my $BundledQuantity = $Map->get('BundledQuantity');
        foreach my $hPrices (@{$BundledProduct->getWithParent('Deposits')}) {
            $Deposit{$hPrices->{'CurrencyID'}} += $hPrices->{'Deposit'} * $BundledQuantity if defined $hPrices->{'Deposit'};
        }
    }
    foreach my $CurrencyID (keys %Deposit) {
        my $Value = $Deposit{$CurrencyID};
        my @Prices = grep { $_->{'CurrencyID'} eq $CurrencyID } @{GetInfoProductPricesByProductID($self->id)};
        if (scalar @Prices) {
            if (defined $Value or defined $Prices[0]{'ListPrice'} or defined $Prices[0]{'ManufacturerPrice'}) {
                UpdateProductPrice({ %{$Prices[0]}, 'Deposit'=>$Value});
            } else {
                DeleteProductPrice($Prices[0]);
            }
        } elsif(defined $Value) {
            InsertProductPrice({
                'ProductID'  => $self->id,
                'CurrencyID' => $CurrencyID,
                'Deposit' => $Value
            });
        }
    }
}

#========================================================================================
# �function     canAddToBasket
#----------------------------------------------------------------------------------------
# �syntax       $CanAddToBasket = $Product->canAddToBasket($Quantity);
#----------------------------------------------------------------------------------------
# �description  Tests if a product can be added to the basket with the given quantity.
#----------------------------------------------------------------------------------------
# �input        $Quantity | quantity (optional: default Product.MinOrder) | float
# �return       $CanAddToBasket | true if the pruduct can be added to the basket | boolean
#========================================================================================
sub canAddToBasket {
    my $self = shift;
    my ($Quantity )= @_;
    return 0 unless $self->get('IsAvailable');
    return 0 unless $self->get('IsVisible');
    return 0 if not defined $self->getWithParent('TaxClass');

    if( defined $Quantity ) {
        TestFloat('Quantity', $Quantity);
    }
    else {
        $Quantity = $self->get('MinOrder') // 1;
    }

    my $ProductLowOnStockAction = $self->getSite()->get('ProductLowOnStockAction');

    if ($self->get('IsBundleProduct')) {
        my $CanAddToBasket = 1;
        foreach my $BundledProductMap ( @{$self->get('BundleProductMaps')} ) {
            $CanAddToBasket = $BundledProductMap->get('BundledProduct')->canAddToBasketBundledProduct($Quantity * $BundledProductMap->get('BundledQuantity'));
            last if $CanAddToBasket == 0;
        }
        return $CanAddToBasket;
    } else {
        my $StockLevel = $self->get('StockLevel');
        return 1 unless defined $StockLevel;
        return 1 if $ProductLowOnStockAction == STOCK_IS_ORDERABLE;

        return 0 if $StockLevel <= 0;
        return 0 if $StockLevel < $Quantity;
        return 1;
    }
}

sub canAddToBasketBundledProduct {
    my $self = shift;
    my ($Quantity )= @_;
    return 0 unless $self->get('IsAvailable');
    return 0 if not defined $self->getWithParent('TaxClass');


    if( defined $Quantity ) {
        TestFloat('Quantity', $Quantity);
    }
    else {
        $Quantity = $self->get('MinOrder') // 1;
    }

    my $ProductLowOnStockAction = $self->getSite()->get('ProductLowOnStockAction');

    my $StockLevel = $self->get('StockLevel');
    return 1 unless defined $StockLevel;
    return 1 if $ProductLowOnStockAction == STOCK_IS_ORDERABLE;

    return 0 if $StockLevel <= 0;
    return 0 if $StockLevel < $Quantity;
    return 1;
}

#========================================================================================
# �function     isDeliveryTimeLonger
#----------------------------------------------------------------------------------------
# �syntax       $IsLonger = $Product->isDeliveryTimeLonger();
#----------------------------------------------------------------------------------------
# �description  Tests if a product has a longer delivery time, because it is not currently
#               on stock.
#               The current basket is not involved.
#----------------------------------------------------------------------------------------
# �return       $IsLonger | delivery time is longer | boolean
#========================================================================================
sub isDeliveryTimeLonger {
    my $self = shift;

    if ($self->get('IsBundleProduct')) {
        my $IsDeliveryTimeLonger = 0;
        foreach my $BundledProductMap ( @{$self->get('BundleProductMaps')} ) {
            my $BundledProduct = $BundledProductMap->get('BundledProduct');
            if( $BundledProduct->get('IsBundleProduct') ) {
                Error('BundleLoop', {
                    ProductID => $self->id,
                    BundledProductID => $BundledProduct->id,
                });
            }
            last if $BundledProduct->isDeliveryTimeLonger();
        }
        return $IsDeliveryTimeLonger;
    } else {
        my $StockLevel = $self->get('StockLevel');
        return 0 unless defined $StockLevel;
        return 1 if $StockLevel <= 0 and $self->getSite()->get('ProductLowOnStockAction') == STOCK_IS_ORDERABLE;
        return 0;
    }
}

#========================================================================================
# �function     stockAlert
#----------------------------------------------------------------------------------------
# �syntax       $StockAlert = $Product->stockAlert($Quantity);
# �syntax       $StockAlert = $Product->stockAlert;
#----------------------------------------------------------------------------------------
# �description  Gets the availability status of the product.
#               Used by the attribute 'StockLevelAlert'.
#               Possible return values are:
#               <ul>
#                 <li>NO_STOCK - sock handling is disabled (stock level is undefined)</li>
#                 <li>ON_STOCK - product is on stock, StockLevel > StockLevelAlert or StockLevelAlert is not defined</li>
#                 <li>WARN_STOCK - product is on stock, StockLevel <= StockLevelAlerta</li>
#                 <li>OUT_STOCK - product is not on stock (StockLevel < $Quantity)</li>
#               </ul>
#----------------------------------------------------------------------------------------
# �input        $Quantity | quantity (optional: default Product.MinOrder) | float
# �return       $StockAlert | stock alert string | string
#========================================================================================
sub stockAlert {
    my $self = shift;
    my ($Quantity )= @_;
    $Quantity = $self->getWithParent('MinOrder') unless defined $Quantity;

    my $hStockVars = $self->getWithParent(['IsBundleProduct', 'StockLevel', 'StockLevelAlert', 'MinOrder']);
    if ( $hStockVars->{'IsBundleProduct'} ) {
        my %StockConstants = ('NO_STOCK'   => 0,
                              'ON_STOCK'   => 1,
                              'WARN_STOCK' => 2,
                              'OUT_STOCK'  => 3, );
        my $StockAlert = 'NO_STOCK';
        foreach my $BundledProductMap ( @{$self->get('BundleProductMaps')} ) {
            my $BundledStockAlert = $BundledProductMap->get('BundledProduct')->stockAlert($Quantity * $BundledProductMap->get('BundledQuantity'));
            $StockAlert = $BundledStockAlert if ($StockConstants{$BundledStockAlert} > $StockConstants{$StockAlert} );
            last if $StockAlert eq 'OUT_STOCK'
        }
        return $StockAlert;
    } else {
        my $StockAlert;
        if ( not defined $hStockVars->{'StockLevel'} ) {
            $StockAlert = 'NO_STOCK';
        } elsif ($hStockVars->{'StockLevel'} < $Quantity ) {
            $StockAlert = 'OUT_STOCK';
        } elsif (defined $hStockVars->{'StockLevelAlert'} and $hStockVars->{'StockLevel'} <  $hStockVars->{'StockLevelAlert'} ) {
            $StockAlert = 'WARN_STOCK';
        } else {
            $StockAlert = 'ON_STOCK';
        }
        return $StockAlert;
    }
}

#========================================================================================
# �function     addToStock
#----------------------------------------------------------------------------------------
# �syntax       $Product->addToStock($Quantity);
#----------------------------------------------------------------------------------------
# �description  Increments the StockLevel by the given Quantity.
#----------------------------------------------------------------------------------------
# �input        $Quantity | product quantity | float
#========================================================================================
sub addToStock {
    my $self = shift;
    my ($Quantity )= @_;
    my $StockLevel = $self->get('StockLevel');
    return unless defined $StockLevel;

    $self->set({'StockLevel'=> $StockLevel + $Quantity });
    return;
}

#========================================================================================
# �function     addBundleToStock
#----------------------------------------------------------------------------------------
# �syntax       $Product->addBundleToStock($Quantity);
#----------------------------------------------------------------------------------------
# �description  Adds the given Quantity to the Stocklevel of each bundled product.
#               See also <function addToStock>($Quantity). If the product is no
#               bundle then this function is called for this product.
#----------------------------------------------------------------------------------------
# �input        $Quantity | product quantity | float
#========================================================================================
sub addBundleToStock {
    my $self = shift;
    my ($Quantity )= @_;

    if( $self->get('IsBundleProduct') ) {
        foreach my $bundledProductMap (@{$self->get('BundleProductMaps')}) {
            $bundledProductMap->get('BundledProduct')->addToStock(
                $Quantity * $bundledProductMap->get('BundledQuantity')
            );
        }
    }
    else {
        $self->addToStock($Quantity);
    }
    return;
}

#========================================================================================
# �function     substractFromStock
#----------------------------------------------------------------------------------------
# �syntax       $Product->substractFromStock($Quantity);
#----------------------------------------------------------------------------------------
# �description  Decrements the stock level by the given quantity
#               and sets the visibility of a product if choosen and sends a stock warning.
#----------------------------------------------------------------------------------------
# �input        $Quantity | product quantity | float
#========================================================================================
sub substractFromStock {
    my $self = shift;
    my ($Quantity)= @_;

    my $StockLevel = $self->get('StockLevel');
    return unless defined $StockLevel;

    my $NewStockLevel = $StockLevel - $Quantity;

    # Set quantity negative to adapt the relative change
    $Quantity = ($Quantity * -1);

    my $Shop = $self->getSite();

    # Set the relative stocklevel as quantity to substract from the actual stocklevel value.
    # It's necessary to have the absolut and relative stocklevel for either setting the stocklevel
    # to a fix value or decrease it.
    if( $NewStockLevel <= 0 and
        $Shop->get('ProductLowOnStockAction') == STOCK_NOT_VISIBLE){
        $self->set({
            'IsVisible'  => 0,
            'StockLevel' => $NewStockLevel
        }, undef, {'RelativeStockLevel' => $Quantity});
        # set bundles invisible which the product is assigned to
        my $aBundleProductIDs = $self->get('IncludedInBundles');
        $_->set({'IsVisible' => 0}) foreach @$aBundleProductIDs;
    } else {
        $self->set({'StockLevel' => $NewStockLevel}, undef, {'RelativeStockLevel' => $Quantity});
    }

    SendStockWarnings($self);
    return;
}

#========================================================================================
# �function     substractBundleFromStock
#----------------------------------------------------------------------------------------
# �syntax       $Product->substractBundleFromStock($Quantity);
#----------------------------------------------------------------------------------------
# �description  Decrements the stock level of each bundled Product by given quantity.
#               See also <function substractFromStock>($Quantity). If the product is no
#               bundle then this function is called for this product.
#----------------------------------------------------------------------------------------
# �input        $Quantity | product quantity | float
#========================================================================================
sub substractBundleFromStock {
    my $self = shift;
    my ($Quantity )= @_;

    if( $self->get('IsBundleProduct') ) {
        foreach my $bundledProductMap (@{$self->get('BundleProductMaps')}) {
            $bundledProductMap->get('BundledProduct')->substractFromStock(
                $Quantity * $bundledProductMap->get('BundledQuantity')
            );
        }
    }
    else {
        $self->substractFromStock($Quantity);
    }

    return;
}

#========================================================================================
# �function     addCustomizationTextOption
#----------------------------------------------------------------------------------------
# �syntax       $Product->addCustomizationTextOption($hParams);
#----------------------------------------------------------------------------------------
# �description  Adds a text option for a customizable product.
#----------------------------------------------------------------------------------------
# �input        $Quantity | product quantity | float
# �return       $TextOption | text option object | object
#========================================================================================
sub addCustomizationTextOption {
    my $self = shift;
    my ($hParams)= @_;
    TestHash('hParams', $hParams, {
        'IsMandatory' => 'Integer;NotNull',
        'MaxLength'    => 'Integer;NotNull',
        'MultiLine'   => 'Integer;NotNull',
    });

    my $OptionsFolder = $self->folder('CustomizationOptions');
    $hParams->{'Alias'} = $OptionsFolder->nextAlias('CustomizationTextOption%s');
    $hParams->{'Product'} = $self;
    delete $hParams->{'ProductID'};
    my $TextOption = $OptionsFolder->insertChild('TextOption', $hParams);

    return $TextOption;
}

#========================================================================================
# �function     addCustomizationFileOption
#----------------------------------------------------------------------------------------
# �syntax       $Product->addCustomizationFileOption($hParams);
#----------------------------------------------------------------------------------------
# �description  Adds a file option for a customizable product.
#----------------------------------------------------------------------------------------
# �input        $Quantity | product quantity | float
# �return       $FileOption | file option object | object
#========================================================================================
sub addCustomizationFileOption {
    my $self = shift;
    my ($hParams)= @_;
    TestHash('hParams', $hParams, {
        'IsMandatory' => 'Integer;NotNull',
        'MinWidth'    => 'Integer;NotNull',
        'MinHeight'   => 'Integer;NotNull',
    });

    my $OptionsFolder = $self->folder('CustomizationOptions');
    $hParams->{'Alias'} = $OptionsFolder->nextAlias('CustomizationFileOption%s');
    $hParams->{'Product'} = $self;
    delete $hParams->{'ProductID'};
    my $FileOption = $OptionsFolder->insertChild('FileOption', $hParams);

    return $FileOption;
}

#addCustomizableArea
sub addCustomizableArea {
    my $self = shift;
    my ($hParams)= @_;
    TestHash('hParams', $hParams, {
        'AreaID' => 'Integer;NotNull',
        'FromTop' => 'Integer;NotNull',
        'FromLeft' => 'Integer;NotNull',
        'Width' => 'Integer;NotNull',
        'Height' => 'Integer;NotNull',
    });

    my $AreasFolder = $self->folder('CustomizableAreas');
    $hParams->{'Alias'} = $hParams->{'AreaID'};
    $hParams->{'Product'} = $self;
    delete $hParams->{'AreaID'};
    delete $hParams->{'ProductID'};
    my $CustomizableArea = $AreasFolder->insertChild('CustomizableArea', $hParams);

    return $CustomizableArea;
}

#deleteCustomizableArea
sub deleteCustomizableArea {
    my $self = shift;
    my ($Alias)= @_;
    TestString('Alias', $Alias);

    if ($self->existsCustomizableArea($Alias)) {
        my $AreasFolder = $self->folder('CustomizableAreas');
        $AreasFolder->child($Alias)->delete;
    }
    return;
}

#existsCustomizableArea
sub existsCustomizableArea {
    my $self = shift;
    my ($Alias)= @_;
    TestString('Alias', $Alias);

    return $self->folder('CustomizableAreas')->existsChild($Alias);
}

#updateCustomizableArea
sub updateCustomizableArea {
    my $self = shift;
    my ($hParams)= @_;
    TestHash('hParams', $hParams, {
        'AreaID' => 'Integer;NotNull',
        'FromTop' => 'Integer;NotNull',
        'FromLeft' => 'Integer;NotNull',
        'Width' => 'Integer;NotNull',
        'Height' => 'Integer;NotNull',
    });

    my $Alias = $hParams->{'AreaID'};

    if ($self->existsCustomizableArea($Alias)) {
        my $AreasFolder = $self->folder('CustomizableAreas');
        $hParams->{'Product'} = $self;
        delete $hParams->{'AreaID'};
        delete $hParams->{'ProductID'};
        my $CustomizableArea = $AreasFolder->child($Alias);
        $CustomizableArea->set($hParams);
    }

    return;
}


#========================================================================================
# �function     filePath
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Object->filePath;
#----------------------------------------------------------------------------------------
# �description  Returns the relative file path of object from the System file path.
#               If the product is an variation the filePath of the SuperProduct will returned.
#----------------------------------------------------------------------------------------
# �return       $Path | relative file path | string
#========================================================================================
sub filePath {
    my $self = shift;
    my $SuperProduct = $self->superProduct;
    if ( $SuperProduct ) {
        return $SuperProduct->filePath;
    } else {
        return $self->SUPER::filePath;
    }
}

#========================================================================================
# �function     tleHash
# �state        private
#----------------------------------------------------------------------------------------
# �syntax       $hVars = $Object->tleHash;
#----------------------------------------------------------------------------------------
# �description  Returns the tied hash for tle compiler.
#               Unlike the TLE hash of regular objects, this one also considers the
#               super product if the product is a sub product unless we have a unity shop.
#----------------------------------------------------------------------------------------
# �return       $hVars | TLE variables | ref.hash
#========================================================================================
sub tleHash {
    my $self = shift;

    if($self->getSite()->get('IsUnityRegistered')) {
        return DE_EPAGES::Object::API::TLEObjectHash::TieObject($self);
    } else {
        return DE_EPAGES::Object::API::TLEWithParentHash::TieObject($self);
    }
}

#========================================================================================
# �function     setListPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->setListPrice($CurrencyID, $TaxModel, $Value);
# �example      $Product->setListPrice('EUR', 1, 19.99);
#               $Product->set({ HasSubOwnPrices => 1 }) if defined $Product->superProduct;
# �example      $Product->setListPrice('EUR', 1, undef);
#----------------------------------------------------------------------------------------
# �description  Sets the list price of this product.
#               If $Product is a sub product, then the price overrides the price of the
#               super product. In this case, you should also set the attribute
#               HasSubOwnPrices of the sub product to true.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $Value | price. A value of undef deletes the price. | float
#========================================================================================
sub setListPrice {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $Value) = @_;
    TestString('CurrencyID', $CurrencyID);

    # get the default tax model from the shop
    my $Shop = $self->getSite();
    $TaxModel = $Shop->get('InputTaxModel') unless defined $TaxModel;

    if( defined $Value ) {
        TestFloat('Value', $Value);
        # convert gross price to net if necessary
        my $Postfix = $self->taxMatrix->instanceOf('TaxMatrixDigital') ? 'Digital' : '';
        my $TaxClass = $self->getWithParent('TaxClass'.$Postfix);

        $Value = $self->taxMatrix->gross2net($CurrencyID, $Value,
                                            {'TaxModel' => $TaxModel,
                                             'TaxClass' => $TaxClass,
                                             'Shop' => $Shop,
                                            });
    }

    my $SetUpdated = sub {
        my $ListPriceProfile = LoadClassByAlias('Product')->lastUpdateProfile('ListPrice');
        $self->setUpdated( $ListPriceProfile );
    };

    # set list price
    my @Prices = grep { $_->{'CurrencyID'} eq $CurrencyID } @{GetInfoProductPricesByProductID($self->id)};
    if (scalar @Prices) {
        if (defined $Value or defined $Prices[0]{'ManufacturerPrice'} or defined $Prices[0]{'EcoParticipation'} or defined $Prices[0]{'Deposit'}) {
            UpdateProductPrice({ %{$Prices[0]}, 'ListPrice'=>$Value});
        } else {
            DeleteProductPrice($Prices[0]);
        }
        # update only if price has changed
        if(Compare('Float', $Prices[0]{'ListPrice'}, $Value) != 0) {
            $SetUpdated->();
        }
    }
    elsif(defined $Value) {
        InsertProductPrice({
            'ProductID'  => $self->id,
            'CurrencyID' => $CurrencyID,
            'ListPrice'  => $Value
        });
        $SetUpdated->();
    }
}

#========================================================================================
# �function     getListPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Value = $Product->getListPrice($CurrencyID, $TaxModel);
# �syntax       $Value = $Product->getListPrice($CurrencyID, $TaxModel, $ExtraPrecision);
# �example      $Value = $Product->getListPrice('EUR', TAX_MODEL_NET);
#----------------------------------------------------------------------------------------
# �description  Returns the list price of this product.
#               Returns the list price of the super product if $Product is a sub product
#               that does not have its own price, i.e. the attribute HasSubOwnPrices is
#               false.
#               If available, the user's address is used to determine a specific price
#               for digital goods and ebooks.
#               Returns undef if the product does not have a list price.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $ExtraPrecision | (optional, default: false) extra precision | int
# �input        $User | user object | DE_EPAGES::Permission::API::Object::User (optional)
# �input        $TaxMatrix | tax matrix object | DE_EPAGES::Tax::API::Object::TaxMatrix (optional)
# �return       $Value | price | float
#========================================================================================
sub getListPrice {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $ExtraPrecision, $User, $TaxMatrix) = @_;
    TestString('CurrencyID', $CurrencyID);
    $TaxModel = $self->getSite()->get('InputTaxModel') unless defined $TaxModel;

    my $SuperProduct = $self->superProduct;
    return $self->getDirectListPrice($CurrencyID, $TaxModel, $ExtraPrecision, $User, $TaxMatrix ) unless defined $SuperProduct;
    return $self->getDirectListPrice($CurrencyID, $TaxModel, $ExtraPrecision, $User, $TaxMatrix) if $self->get('HasSubOwnPrices');
    return $SuperProduct->getDirectListPrice($CurrencyID, $TaxModel, $ExtraPrecision, $User, $TaxMatrix);
}

#========================================================================================
# �function     getReferencePrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Value = $Product->getReferencePrice($CurrencyID, $Price, $Discount);
# �example      $Value = $Product->getReferencePrice('EUR', 12);
#----------------------------------------------------------------------------------------
# �description  Calculates the price of a product relative to the reference amout.
#               The following formula is used:
#               <br/>#Price * (#Product.RefAmount / #Product.RefContentAmount)
#               Returns undef if the product does not have a list price.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $Price | price for the default amount | float
# �input        $Discount | (optional, default=0) applicable discount in percent | float
# �return       $Value | price for the reference amount | float
#========================================================================================
sub getReferencePrice {
    my $self = shift;
    my ($CurrencyID, $Price, $Discount) = @_;
    return undef unless defined $Price;
    TestString('CurrencyID', $CurrencyID);

    my $Shop = $self->getSite();
    my $RefAmount;
    my $RefContentAmount;

    # Unity sub products have to be independent of their master product
    if($Shop->get('IsUnityRegistered') && $self->isSubProduct) {
        $RefAmount = $self->get('RefAmount');
        $RefContentAmount = $self->get('RefContentAmount');
    } else {
        $RefAmount = $self->getWithParent('RefAmount');
        $RefContentAmount = $self->getWithParent('RefContentAmount');
    }

    return undef if !defined $RefContentAmount || !defined $RefAmount || $RefContentAmount <= 0;

    my $ReferencePrice = $Price * ($RefAmount / $RefContentAmount);
    if( defined $Discount ) {
        $ReferencePrice = $ReferencePrice * (100-$Discount)/100; ## no critic qw(ProhibitMagicNumbers)
    }

    TestString('CurrencyID', $CurrencyID , 3, 3);
    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new();
    return $Formatter->round($ReferencePrice, GetInfoCurrency($CurrencyID)->{'MinorUnit'});
}

#========================================================================================
# �function     getDirectListPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Value = $Product->getDirectListPrice($CurrencyID, $TaxModel);
# �example      $Value = $Product->getDirectListPrice('EUR', 1);
#               $Value = $Product->getDirectListPrice('EUR', 1, undef, $User);
#----------------------------------------------------------------------------------------
# �description  Returns the list price that is defined on directly on this product object.
#               Returns undef if $Product is a sub product that does not have its own
#               price and inherits its price from the super product.
#               If available, the user's address is used to determine a specific price
#               for digital goods and ebooks.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $ExtraPrecision | extra precision | int
# �input        $User | user object | DE_EPAGES::Permission::API::Object::User (optional)
# �input        $TaxMatrix | tax matrix object | DE_EPAGES::Tax::API::Object::TaxMatrix (optional)
# �return       $Value | price | float
#========================================================================================
sub getDirectListPrice {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $ExtraPrecision, $User, $TaxMatrix) = @_;

    TestString('CurrencyID', $CurrencyID);
    $TaxModel = $self->getSite()->get('InputTaxModel') unless defined $TaxModel;

    my @Prices = grep { $_->{'CurrencyID'} eq $CurrencyID } @{GetInfoProductPricesByProductID($self->id)};
    return undef unless scalar @Prices;
    my $Price = $Prices[0]{'ListPrice'};
    return undef unless defined $Price;


    # use the given tax matrix for calculation or use the tax matrix from the given user
    $TaxMatrix = defined $TaxMatrix ? $TaxMatrix : $self->taxMatrix($User);

    my $TaxClass;
    my $TaxArea;

    if ($TaxMatrix->instanceOf('TaxMatrixDigital')) {
        if (defined $User) {
            my $Address = $User->get('ShippingAddress') // $User->get('BillingAddress') // $self->getSite->get('Address');
            $TaxArea = $TaxMatrix->taxAreaDigitalByCountry($Address->get('CountryID'));
        }
        $TaxClass = $self->getWithParent('TaxClassDigital');
    } else {
        my $ProductTaxClassAlias;
        my $TaxMatrixID = $TaxMatrix->id;
        my $Shop = $self->getSite();
        if ($Shop->get('DeliveryThresholdIsActive')) {
            my $aProductTaxClassMapping = $self->get('TaxClassMapping');
            $ProductTaxClassAlias = _GetTaxAliasFromTaxClassMapping($aProductTaxClassMapping, $TaxMatrixID);
        }
        $ProductTaxClassAlias = defined $ProductTaxClassAlias ? $ProductTaxClassAlias : $self->getWithParent('TaxClass')->alias;
        # prevent using the TaxClass from the wrong TaxMatrix if changed
        # important for delivery threshold
        $TaxClass = $TaxMatrix->getTaxMatrixTaxClassByAlias($ProductTaxClassAlias);
    }

    # convert net price to gross if necessary
    $Price = $TaxMatrix->net2gross($CurrencyID, $Price, {
        'TaxModel' => $TaxModel,
        'TaxClass' => $TaxClass,
        'TaxArea' => $TaxArea,
        'Shop'    => $self->getSite(),

    }, $ExtraPrecision);

    return $Price;
}

#========================================================================================
# �function     existSubProductPrices
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->existSubProductPrices();
# �example      return if $Product->existSubProductPrices();
#----------------------------------------------------------------------------------------
# �description  Returns true if a cheapest sub product exists that is currently available.
#               Returns false if $Product is not a super product.
#----------------------------------------------------------------------------------------
# �return       $ExistsPrices | true if a cheaper sub product exists  | boolean
#========================================================================================
sub existSubProductPrices {
    my $self = shift;

    my $aInfoSubProducts = $self->_infoSubProducts;
    my @AvailableSubProducts = grep {
        $_->{'HasSubOwnPrices'} && $_->{'IsAvailable'} && $_->{'IsVisible'}
    } @$aInfoSubProducts;
    return scalar @AvailableSubProducts ? 1 : 0;
}

#========================================================================================
# �function     subProductWithLowestPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $SubProduct = $SuperProduct->subProductWithLowestPrice($hParams);
#----------------------------------------------------------------------------------------
# �description  Returns the cheapest sub product that is currently available in the
#               desired quantity (using <function canAddToBasket>).
#               Returns undef if no sub product is available.
#               Must be called on a super product, otherwise returns always undef.
#----------------------------------------------------------------------------------------
# �input        $hParams | <ul>
#                   <li>Quantity      - desired quantiy (e.g. MinOrder) - float
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>IgnoreCanAddToBasket - disable add to basket check (optional)- boolean
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
# �return       $SubProduct | cheapest sub product | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub subProductWithLowestPrice {
    my $self = shift;
    my ($hParams) = @_;

    my $Iterator = DAL_GetSubProductsOrderedByPrice( $self->id, $hParams->{'CurrencyID'}, 'ASC' );
    while( my $SubProductID = $Iterator->next ) {
        my $SubProduct = LoadObject( $SubProductID );
        next unless ($hParams->{'IgnoreCanAddToBasket'} || $SubProduct->canAddToBasket($hParams->{'Quantity'}));
        return $SubProduct;
    }

    return undef;
}

#========================================================================================
# �function     subProductWithLowestPriceThatHasOwnPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $SubProduct = $SuperProduct->subProductWithLowestPriceThatHasOwnPrice($CurrencyID);
#----------------------------------------------------------------------------------------
# �description  Returns the cheapest sub product that is currently available in the
#               desired quantity (using <function canAddToBasket>) and that has it's own price.
#               Returns undef if no sub product is available.
#               Must be called on a super product, otherwise returns always undef.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | alpha currency code (ISO 4217) - char(3) | string
# �return       $SubProduct | cheapest sub product | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub subProductWithLowestPriceThatHasOwnPrice {
    my $self = shift;
    my ($CurrencyID) = @_;

    my $Iterator = DAL_GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice($self->id, $CurrencyID, 'ASC');
    while( my $SubProductID = $Iterator->next ) {
        my $SubProduct = LoadObject( $SubProductID );
        next unless ($SubProduct->canAddToBasket());

        return $SubProduct;
    }

    return undef;
}

#========================================================================================
# �function     subProductWithHighestPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $SubProduct = $SuperProduct->subProductWithHighestPrice($hParams);
#----------------------------------------------------------------------------------------
# �description  Returns the most expensive sub product that is currently available in the
#               desired quantity (using <function canAddToBasket>).
#               Returns undef if no sub product is available.
#               Must be called on a super product, otherwise returns always undef.
#----------------------------------------------------------------------------------------
# �input        $hParams | <ul>
#                   <li>Quantity      - desired quantiy (e.g. MinOrder) - float
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>IgnoreCanAddToBasket - disable add to basket check (optional)- boolean
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
# �return       $SubProduct | most expensive sub product | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub subProductWithHighestPrice {
    my $self = shift;
    my ($hParams) = @_;

    my $Iterator = DAL_GetSubProductsOrderedByPrice( $self->id, $hParams->{'CurrencyID'}, 'DESC' );
    while( my $SubProductID = $Iterator->next ) {
        my $SubProduct = LoadObject( $SubProductID );
        next unless ($hParams->{'IgnoreCanAddToBasket'} || $SubProduct->canAddToBasket($hParams->{'Quantity'}));
        return $SubProduct;
    }

    return undef;
}

#========================================================================================
# �function     subProductWithHighestPriceThatHasOwnPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $SubProduct = $SuperProduct->subProductWithHighestPriceThatHasOwnPrice($CurrencyID);
#----------------------------------------------------------------------------------------
# �description  Returns the most expensive sub product that is currently available in the
#               desired quantity (using <function canAddToBasket>) and that has it's own price.
#               Returns undef if no sub product is available.
#               Must be called on a super product, otherwise returns always undef.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | alpha currency code (ISO 4217) - char(3) | string
# �return       $SubProduct | most expensive sub product | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub subProductWithHighestPriceThatHasOwnPrice {
    my $self = shift;
    my ($CurrencyID) = @_;

    my $Iterator = DAL_GetVisibleAndAvailableAndWithOwnPriceSubProductsOrderedByPrice($self->id, $CurrencyID, 'DESC');
    while( my $SubProductID = $Iterator->next ) {
        my $SubProduct = LoadObject( $SubProductID );
        next unless ($SubProduct->canAddToBasket());

        return $SubProduct;
    }

    return undef;
}

#========================================================================================
# �function     lowestSubProductPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $LowestPrice = $SuperProduct->lowestSubProductPrice($hParams);
#----------------------------------------------------------------------------------------
# �description  Returns the cheapest sub product price that is currently available in the
#               minimum orderable quantity (using <function canAddToBasket>).
#               Returns undef if no sub product is available.
#               Must be called on a super product, otherwise returns always undef.
#----------------------------------------------------------------------------------------
# �input        CurrencyID | alpha currency code (ISO 4217) | char(3)
# �input        $TaxModel | gross or net price | int
# �return       $LowestPrice | lowest price of product | money
#========================================================================================
sub lowestSubProductPrice {
    my $self = shift;
    my ($CurrencyID, $TaxModel) = @_;

    my $SubProduct = $self->subProductWithLowestPriceThatHasOwnPrice($CurrencyID);
    return unless defined($SubProduct);

    my $SubProductPrice = $SubProduct->getListPrice($CurrencyID, $TaxModel);
    my $ProductPrice = $self->getListPrice($CurrencyID, $TaxModel);

    if (not defined $ProductPrice or $SubProductPrice <= $ProductPrice) {
        return $SubProduct->displayPrice({
            'CurrencyID' => $CurrencyID,
            'TaxModel'   => $TaxModel
        });
    } else {
        return $self->displayPrice({
            'CurrencyID' => $CurrencyID,
            'TaxModel'   => $TaxModel
        });
    }
}

#========================================================================================
# �function     highestSubProductPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $HighestPrice = $SuperProduct->highestSubProductPrice($hParams);
#----------------------------------------------------------------------------------------
# �description  Returns the highest sub product price that is currently available in the
#               minimum orderable quantity (using <function canAddToBasket>).
#               Returns undef if no sub product is available.
#               Must be called on a super product, otherwise returns always undef.
#----------------------------------------------------------------------------------------
# �input        CurrencyID | alpha currency code (ISO 4217) | char(3)
# �input        $TaxModel | gross or net price | int
# �return       $HighestPrice | highest price of product | money
#========================================================================================
sub highestSubProductPrice {
    my $self = shift;
    my ($CurrencyID, $TaxModel) = @_;

    my $SubProduct = $self->subProductWithHighestPriceThatHasOwnPrice($CurrencyID);
    return unless defined($SubProduct);

    my $SubProductPrice = $SubProduct->getListPrice($CurrencyID, $TaxModel);
    my $ProductPrice = $self->getListPrice($CurrencyID, $TaxModel);

    if (not defined $ProductPrice or $SubProductPrice >= $ProductPrice) {
        return $SubProduct->displayPrice({
            'CurrencyID' => $CurrencyID,
            'TaxModel'   => $TaxModel
        });
    } else {
        return $self->displayPrice({
            'CurrencyID' => $CurrencyID,
            'TaxModel'   => $TaxModel
        });
    }
}

#========================================================================================
# �function     fromPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $FromPrice = $SuperProduct->fromPrice($hParams);
#----------------------------------------------------------------------------------------
# �description  Price of the cheapest sub product that is currently available in the
#               desired quantity (using <function canAddToBasket>).
#               Returns undef if no sub product is available.
#               Must be called on a super product, otherwise returns always undef.
#----------------------------------------------------------------------------------------
# �input        $hParams | <ul>
#                   <li>Quantity      - desired quantiy (e.g. MinOrder) - float
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price   - int
#                   <li>Customer      - customer            (optional)- Customer
#                   <li>IgnoreCanAddToBasket - disable add to basket check (optional) - boolean
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
# �return       $FromPrice | price of the cheapest sub product | float
#========================================================================================
sub fromPrice {
    my $self = shift;
    my ($hParams) = @_;

    my $SubProduct = $self->subProductWithLowestPrice($hParams);

    return defined $SubProduct
        ? $SubProduct->displayPrice( $hParams )
        : undef;
}

#========================================================================================
# �function     toPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $FromPrice = $SuperProduct->toPrice($hParams);
#----------------------------------------------------------------------------------------
# �description  Price of the most expensive sub product that is currently available in the
#               desired quantity (using <function canAddToBasket>).
#               Returns undef if no sub product is available.
#               Must be called on a super product, otherwise returns always undef.
#----------------------------------------------------------------------------------------
# �input        $hParams | <ul>
#                   <li>Quantity      - desired quantiy (e.g. MinOrder) - float
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price   - int
#                   <li>Customer      - customer            (optional)- Customer
#                   <li>IgnoreCanAddToBasket - disable add to basket check (optional) - boolean
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
# �return       $FromPrice | price of the most expensive sub product | float
#========================================================================================
sub toPrice {
    my $self = shift;
    my ($hParams) = @_;

    my $SubProduct = $self->subProductWithHighestPrice($hParams);

    return defined $SubProduct
        ? $SubProduct->displayPrice( $hParams )
        : undef;
}

#========================================================================================
# �function     displayPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $DisplayPrice = $Product->displayPrice($hVars);
#----------------------------------------------------------------------------------------
# �example      $DisplayPrice = $Product->displayPrice({
#                   CurrencyID => 'USD',
#                   TaxModel => $Product->getSite()->get('TaxModel')
#               });
#----------------------------------------------------------------------------------------
# �description  Returns the product display price by the given requirements.
#               (not for now shops) Uses the list price unless $Shop->get('ShowCustomerSpecificPrices').
#----------------------------------------------------------------------------------------
# �input        $hVars | <ul>
#                   <li>Quantity      - desired quantiy (e.g. MinOrder) - float
#                   <li>CurrencyID    - alpha currency code (ISO 4217)  - char(3)
#                   <li>TaxModel      - gross or net price              - int
#                   <li>Customer      - customer             (optional) - Customer
#                   <li>User          - user                 (optional) - User
#               </ul> | ref.hash
# �return       $DisplayPrice | price of product | money
#========================================================================================
sub displayPrice {
    my $self = shift;
    my ($hParams) = @_;
    my $Shop = $self->getSite;

    # we use always customer specific prices for now shops
    my $hShopInfo = $Shop->get([qw(IsUnityRegistered ShowCustomerSpecificPrices)]);
    if ( (not $hShopInfo->{'IsUnityRegistered'}) and (not $hShopInfo->{'ShowCustomerSpecificPrices'})){
        return $self->getListPrice($hParams->{'CurrencyID'}, $hParams->{'TaxModel'}, undef, $hParams->{'User'});
    }

    my $DisplayPrice = eval {
        my $Price = $self->price({
            'CurrencyID' => $hParams->{'CurrencyID'},
            'TaxModel'   => $hParams->{'TaxModel'},
            'Customer'   => $hParams->{'Customer'},
            'Quantity'   => $hParams->{'Quantity'},
            'User'       => $hParams->{'User'},
        });

        my $Discount = $self->discount({
            'CurrencyID' => $hParams->{'CurrencyID'},
            'TaxModel'   => $hParams->{'TaxModel'},
            'Customer'   => $hParams->{'Customer'},
            'Quantity'   => $hParams->{'Quantity'},
        });
        return $Price - $Discount;
    };
    if (ExistsError()) {
        my $Error = GetError();
        return undef if $Error->code eq 'ProductNoPrice';
        Error($Error);
    }
    return $DisplayPrice;
}

#========================================================================================
# �function     canInvalidatePageCache
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $CanInvalidate = $Object->canInvalidatePageCache;
#----------------------------------------------------------------------------------------
# �description  Overwrites method DE_EPAGES::Object::API::Object::Object->canInvalidatePageCache
#               to invalidate page cache.
#----------------------------------------------------------------------------------------
# �return       $CanInvalidate | always true | boolean
#========================================================================================
sub canInvalidatePageCache {
    return 1;
}

#========================================================================================
# �function     default
# �state        private
#----------------------------------------------------------------------------------------
# �syntax       $hValues = DE_EPAGES::Product::API::Object::Product->default( $Parent, $hInfo );
# �example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# �description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# �input        $Parent | parent object of the new object | object
# �input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# �return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    my $Shop = $Parent->getSite();
    my $SuperProduct = $hInfo->{'SuperProduct'};
    my $IsAppointment = defined $hInfo->{'IsAppointment'} ? $hInfo->{'IsAppointment'} : ($Parent->alias eq 'Appointments' ? 1 : 0);
    my %DefaultAttributes = (
        %{$class->SUPER::default($Parent, $hInfo)},
        'Shop'          => $Shop,
        'IsNew'         => 0,
        'IsDefault'     => 0,
        'IsDailyPrice'  => 0,
        'OrderUnit'     => $SuperProduct && !$Shop->isUnity() ? undef : LoadObjectByPath('/Units/piece'),    #use attributes from Super product
        'TaxClass'      => $SuperProduct ? undef : $Shop->get('TaxMatrix')->defaultTaxClass,
        'PriceQuantity' => $SuperProduct && !$Shop->isUnity() ? undef : 1,
        'MinOrder'      => $SuperProduct ? undef : 1,
        'IntervalOrder' => $SuperProduct ? undef : 1,
        'IsAvailable'   => 1,
        'HasSubOwnPrices' => 0,
        'IsAppointment' => $IsAppointment,
        'CanPayAppointment' => 0,
        'DurationAmount' => 1,
        'DurationUnit' => $IsAppointment ? LoadObjectByPath('/Units/second/hour') : undef,
        'IsDownloadProduct' => 0,
        'IsBundleProduct' => 0,
        'IsCustomizable' => 0,
        '_IsMaster' => 0,
    );
    my %DefaultAppointmentAttributes =(
        'StockLevel' => 1,
    );
    %DefaultAttributes = (%DefaultAttributes, %DefaultAppointmentAttributes) if $IsAppointment;
    $DefaultAttributes{'StockLevel'} = $SuperProduct->get('StockLevel') if ( defined $SuperProduct );

    return \%DefaultAttributes;
}

#========================================================================================
# �function     getDirectManufacturerPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Value = $Product->getDirectManufacturerPrice($CurrencyID, $TaxModel);
# �example      $Value = $Product->getDirectManufacturerPrice('EUR', 1);
#----------------------------------------------------------------------------------------
# �description  Returns the manufacturer's suggested retail price (MSRP) that is defined
#               directly on this product object.
#               Returns undef if $Product is a sub product that does not have its own
#               MSRP and inherits its MSRP from the super product.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $ExtraPrecision | extra precision | int
# �return       $Value | price | float
#========================================================================================
sub getDirectManufacturerPrice {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $ExtraPrecision) = @_;
    TestString('CurrencyID', $CurrencyID);
    $TaxModel = $self->getSite()->get('InputTaxModel') unless defined $TaxModel;

    my @Prices = grep { $_->{'CurrencyID'} eq $CurrencyID } @{GetInfoProductPricesByProductID($self->id)};
    return undef unless scalar @Prices;
    my $Price = $Prices[0]{'ManufacturerPrice'};
    return undef unless defined $Price;

    # convert net price to gross if necessary
    my $Postfix = $self->taxMatrix->instanceOf('TaxMatrixDigital') ? 'Digital' : '';
    my $TaxClass = $self->getWithParent('TaxClass'.$Postfix);

    $Price = $self->taxMatrix->net2gross($CurrencyID, $Price,{'TaxModel' => $TaxModel,
                                                              'TaxClass' => $TaxClass,
                                                              'Shop' => $self->getSite(),
                                                             }, $ExtraPrecision);
    return $Price;
}

#========================================================================================
# �function     getManufacturerPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Value = $Product->getManufacturerPrice($CurrencyID, $TaxModel);
# �example      $Value = $Product->getManufacturerPrice('EUR', 1);
#----------------------------------------------------------------------------------------
# �description  Returns the manufacturer's suggested retails price (MSRP) of this product.
#               Returns the MSRP of the super product if $Product is a sub product
#               that does not have its own price, i.e. the attribute HasSubOwnPrices is
#               false.
#               Returns undef if the product does not have an MSRP.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.TaxModel) | int
# �input        $ExtraPrecision | extra precision | int
# �return       $Value | price | float
#========================================================================================
sub getManufacturerPrice {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $ExtraPrecision) = @_;
    TestString('CurrencyID', $CurrencyID);
    $TaxModel = $self->getSite()->get('TaxModel') unless defined $TaxModel;

    my $SuperProduct = $self->superProduct;

    return $self->getDirectManufacturerPrice($CurrencyID, $TaxModel, $ExtraPrecision) unless defined $SuperProduct;
    return $self->getDirectManufacturerPrice($CurrencyID, $TaxModel, $ExtraPrecision) if $self->get('HasSubOwnPrices');
    return $SuperProduct->getDirectManufacturerPrice($CurrencyID, $TaxModel, $ExtraPrecision);
}

#========================================================================================
# �function     setManufacturerPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->setManufacturerPrice($CurrencyID, $TaxModel, $Value);
# �example      $Product->setManufacturerPrice('EUR', 1, 19.99);
# �example      $Product->setManufacturerPrice('EUR', 1, undef);
#----------------------------------------------------------------------------------------
# �description  Sets the manufacturer's suggested retail price (MSRP) of this product.
#               If $Product is a sub product, then the MSRP overrides the MSRP of the
#               super product.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $Value | price. A value of undef deletes the price. | float
#========================================================================================
sub setManufacturerPrice {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $Value) = @_;
    TestString('CurrencyID', $CurrencyID);

    # get the default tax model from the shop
    my $Shop = $self->getSite();
    $TaxModel = $Shop->get('InputTaxModel') unless defined $TaxModel;

    if( defined $Value ) {
        TestFloat('Value', $Value);
        # convert gross price to net IF NECESSARY

        my $Postfix = $self->taxMatrix->instanceOf('TaxMatrixDigital') ? 'Digital' : '';
        my $TaxClass = $self->getWithParent('TaxClass'.$Postfix);
        $Value = $self->taxMatrix->gross2net($CurrencyID, $Value,
                                            {'TaxModel' => $TaxModel,
                                             'TaxClass' => $TaxClass,
                                             'Shop'     => $Shop,
                                            });
    }
    # set manufacturer price
    my @Prices = grep { $_->{'CurrencyID'} eq $CurrencyID } @{GetInfoProductPricesByProductID($self->id)};
    if (scalar @Prices) {
        if (defined $Value or defined $Prices[0]{'ListPrice'} or defined $Prices[0]{'EcoParticipation'} or defined $Prices[0]{'Deposit'}) {
            UpdateProductPrice({ %{$Prices[0]}, 'ManufacturerPrice'=>$Value});
        } else {
            DeleteProductPrice($Prices[0]);
        }
    } elsif(defined $Value) {
        InsertProductPrice({
            'ProductID'  => $self->id,
            'CurrencyID' => $CurrencyID,
            'ManufacturerPrice'  => $Value
        });
    }
}

#========================================================================================
# �function     getFieldContent
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $FieldValue = $Product->getFieldContent($FieldName, $LanguageID);
# �example      $FieldValue = $Product->getFieldContent('Text', 'de');
#               $FieldValue = $Product->getFieldContent('Description');
#----------------------------------------------------------------------------------------
# �description  Returns the value of a text field in product. If a subproduct doesn't
#               define any new value, the value of the master product will be used.
#               LanguageID is optional. It is used to get the translation of the value.
#               This function exists, because ContentObject doesn't have getWithParent,
#               but a unified interface was needed to get a text from a UI Object.
#               See EPC-342
#----------------------------------------------------------------------------------------
# �input        $FieldName | name of the field | string
# �input        $LanguageID | language code of a language (optional) | string
# �return       $FieldValue | fields text | string
#========================================================================================
sub getFieldContent {
    my $self = shift;
    my ($FieldName, $LanguageID) = @_;

    return $self->getWithParent($FieldName, $LanguageID);
}

#========================================================================================
# �function     getDirectEcoParticipation
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Value = $Product->getDirectEcoParticipation($CurrencyID, $TaxModel);
# �example      $Value = $Product->getDirectEcoParticipation('EUR', 1);
#----------------------------------------------------------------------------------------
# �description  Returns the eco participation that is defined on directly on this product object.
#               Returns undef if $Product is a sub product that does not have its own
#               price and inherits its price from the super product.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $ExtraPrecision | extra precision | int
# �return       $Value | price | float
#========================================================================================
sub getDirectEcoParticipation {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $ExtraPrecision) = @_;
    TestString('CurrencyID', $CurrencyID);
    $TaxModel = $self->getSite()->get('InputTaxModel') unless defined $TaxModel;

    my @Prices = grep { $_->{'CurrencyID'} eq $CurrencyID } @{GetInfoProductPricesByProductID($self->id)};
    return undef unless scalar @Prices;
    my $Price = $Prices[0]{'EcoParticipation'};
    return undef unless defined $Price;

    # convert net price to gross if necessary
    my $Postfix = $self->taxMatrix->instanceOf('TaxMatrixDigital') ? 'Digital' : '';
    my $TaxClass = $self->getWithParent('TaxClass'.$Postfix);

    $Price = $self->taxMatrix->net2gross($CurrencyID, $Price,{'TaxModel' => $TaxModel,
                                                              'TaxClass' => $TaxClass,
                                                              'Shop'     => $self->getSite(),
                                                             }, $ExtraPrecision);
    return $Price;
}

#========================================================================================
# �function     getEcoParticipation
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Value = $Product->getEcoParticipation($CurrencyID, $TaxModel);
# �example      $Value = $Product->getEcoParticipation('EUR', 1);
#----------------------------------------------------------------------------------------
# �description  Returns the eco participation of this product.
#               Returns the eco participation of the super product if $Product is a sub product.
#               Returns undef if the product does not have a eco participation.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $ExtraPrecision | extra precision | int
# �return       $Value | price | float
#========================================================================================
sub getEcoParticipation {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $ExtraPrecision) = @_;
    TestString('CurrencyID', $CurrencyID);
    $TaxModel = $self->getSite()->get('InputTaxModel') unless defined $TaxModel;

    my $SuperProduct = $self->superProduct;
    return $self->getDirectEcoParticipation($CurrencyID, $TaxModel, $ExtraPrecision) unless defined $SuperProduct;
    return $SuperProduct->getDirectEcoParticipation($CurrencyID, $TaxModel, $ExtraPrecision);
}

#========================================================================================
# �function     setEcoParticipation
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->setEcoParticipation($CurrencyID, $TaxModel, $Value);
# �example      $Product->setEcoParticipation('EUR', 1, 19.99);
#               $Product->set({ HasSubOwnPrices => 1 }) if defined $Product->superProduct;
# �example      $Product->setEcoParticipation('EUR', 1, undef);
#----------------------------------------------------------------------------------------
# �description  Sets the eco participation of this product.
#               If $Product is a sub product, then the price overrides the price of the
#               super product. In this case, you should also set the attribute
#               HasSubOwnPrices of the sub product to true.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $Value | price. A value of undef deletes the price. | float
#========================================================================================
sub setEcoParticipation {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $Value) = @_;
    TestString('CurrencyID', $CurrencyID);

    # get the default tax model from the shop
    my $Shop = $self->getSite();
    $TaxModel = $Shop->get('InputTaxModel') unless defined $TaxModel;

    if( defined $Value ) {
        TestFloat('Value', $Value);
        # convert gross price to net IF NECESSARY
        my $Postfix = $self->taxMatrix->instanceOf('TaxMatrixDigital') ? 'Digital' : '';
        my $TaxClass = $self->getWithParent('TaxClass'.$Postfix);

        $Value = $self->taxMatrix->gross2net($CurrencyID, $Value,
                                            {'TaxModel' => $TaxModel,
                                             'TaxClass' => $TaxClass,
                                             'Shop'     => $Shop,
                                            });
    }
    # set eco participation
    my @Prices = grep { $_->{'CurrencyID'} eq $CurrencyID } @{GetInfoProductPricesByProductID($self->id)};
    if (scalar @Prices) {
        if (defined $Value or defined $Prices[0]{'ListPrice'} or defined $Prices[0]{'ManufacturerPrice'} or defined $Prices[0]{'Deposit'}) {
            UpdateProductPrice({ %{$Prices[0]}, 'EcoParticipation'=>$Value});
        } else {
            DeleteProductPrice($Prices[0]);
        }
    } elsif(defined $Value) {
        InsertProductPrice({
            'ProductID'  => $self->id,
            'CurrencyID' => $CurrencyID,
            'EcoParticipation' => $Value
        });
    }
}

#========================================================================================
# �function     getDirectDeposit
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Value = $Product->getDirectDeposit($CurrencyID, $TaxModel);
# �example      $Value = $Product->getDirectDeposit('EUR', 1);
#----------------------------------------------------------------------------------------
# �description  Returns the eco participation that is defined on directly on this product object.
#               Returns undef if $Product is a sub product that does not have its own
#               price and inherits its price from the super product.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $ExtraPrecision | extra precision | int
# �return       $Value | price | float
#========================================================================================
sub getDirectDeposit {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $ExtraPrecision) = @_;
    TestString('CurrencyID', $CurrencyID);
    $TaxModel = $self->getSite()->get('InputTaxModel') unless defined $TaxModel;

    my @Prices = grep { $_->{'CurrencyID'} eq $CurrencyID } @{GetInfoProductPricesByProductID($self->id)};
    return undef unless scalar @Prices;
    my $Price = $Prices[0]{'Deposit'};
    return undef unless defined $Price;

    # convert net price to gross if necessary
    my $Postfix = $self->taxMatrix->instanceOf('TaxMatrixDigital') ? 'Digital' : '';
    my $TaxClass = $self->getWithParent('TaxClass'.$Postfix);

    $Price = $self->taxMatrix->net2gross($CurrencyID, $Price,{'TaxModel' => $TaxModel,
                                                              'TaxClass' => $TaxClass,
                                                              'Shop'     => $self->getSite(),
                                                             }, $ExtraPrecision);
    return $Price;
}

#========================================================================================
# �function     getDeposit
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Value = $Product->getDeposit($CurrencyID, $TaxModel);
# �example      $Value = $Product->getDeposit('EUR', 1);
#----------------------------------------------------------------------------------------
# �description  Returns the eco participation of this product.
#               Returns the eco participation of the super product if $Product is a sub product
#               that does not have its own price, i.e. the attribute HasSubOwnPrices is
#               false.
#               Returns undef if the product does not have a eco participation.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $ExtraPrecision | extra precision | int
# �return       $Value | price | float
#========================================================================================
sub getDeposit {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $ExtraPrecision) = @_;
    TestString('CurrencyID', $CurrencyID);
    $TaxModel = $self->getSite()->get('InputTaxModel') unless defined $TaxModel;

    my $SuperProduct = $self->superProduct;
    return $self->getDirectDeposit($CurrencyID, $TaxModel, $ExtraPrecision) unless defined $SuperProduct;
    return $self->getDirectDeposit($CurrencyID, $TaxModel, $ExtraPrecision) if $self->get('HasSubOwnPrices');
    return $SuperProduct->getDirectDeposit($CurrencyID, $TaxModel, $ExtraPrecision);
}

#========================================================================================
# �function     setDeposit
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->setDeposit($CurrencyID, $TaxModel, $Value);
# �example      $Product->setDeposit('EUR', 1, 19.99);
#               $Product->set({ HasSubOwnPrices => 1 }) if defined $Product->superProduct;
# �example      $Product->setDeposit('EUR', 1, undef);
#----------------------------------------------------------------------------------------
# �description  Sets the deposit of this product.
#               If $Product is a sub product, then the price overrides the price of the
#               super product. In this case, you should also set the attribute
#               HasSubOwnPrices of the sub product to true.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $Value | price. A value of undef deletes the price. | float
#========================================================================================
sub setDeposit {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $Value) = @_;
    TestString('CurrencyID', $CurrencyID);

    # get the default tax model from the shop
    my $Shop = $self->getSite();
    $TaxModel = $Shop->get('InputTaxModel') unless defined $TaxModel;

    if( defined $Value ) {
        TestFloat('Value', $Value);
        # convert gross price to net IF NECESSARY
        my $Postfix = $self->taxMatrix->instanceOf('TaxMatrixDigital') ? 'Digital' : '';
        my $TaxClass = $self->getWithParent('TaxClass'.$Postfix);

        $Value = $self->taxMatrix->gross2net($CurrencyID, $Value,
                                            {'TaxModel' => $TaxModel,
                                             'TaxClass' => $TaxClass,
                                             'Shop'     => $Shop,
                                            });
    }
    # set deposit
    my @Prices = grep { $_->{'CurrencyID'} eq $CurrencyID } @{GetInfoProductPricesByProductID($self->id)};
    if (scalar @Prices) {
        if (defined $Value or defined $Prices[0]{'ListPrice'} or defined $Prices[0]{'ManufacturerPrice'} or defined $Prices[0]{'EcoParticipation'}) {
            UpdateProductPrice({ %{$Prices[0]}, 'Deposit'=>$Value});
        } else {
            DeleteProductPrice($Prices[0]);
        }
    } elsif(defined $Value) {

        InsertProductPrice({
            'ProductID'  => $self->id,
            'CurrencyID' => $CurrencyID,
            'Deposit' => $Value
        });
    }
}

#========================================================================================
# �function     businessHours
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Calendar = $Product->businessHours;
#----------------------------------------------------------------------------------------
# �description  Returns product calendar (child BusinessHours), if not exists then the
#               calendar will be created.
#----------------------------------------------------------------------------------------
# �return       $Calendar | calendar object | object
#========================================================================================
sub businessHours {
    my $self = shift;
    return $self->child('BusinessHours') if $self->existsChild('BusinessHours');
    my $Calendar = $self->getSite()->businessHours;
    return $self->insertChild('Calendar', {
        Alias=>'BusinessHours',
        BaseCalendar=>$Calendar
    });
}

#========================================================================================
# �function     adaptLayout
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->adaptLayout($hInfo);
#----------------------------------------------------------------------------------------
# �description  Changes selected templates to 0 if product has no bundles or variation.
#----------------------------------------------------------------------------------------
# �input        $hInfo | template type info, used key is 'Templates' | ref.hash.hash
#========================================================================================
sub adaptLayout {
    my $self = shift;
    my ($hInfo) = @_;
    my $Templates = $hInfo->{'Templates'};
    $Templates->{'Content-Bundles'} = 0 unless scalar @{$self->get('BundledProducts')};
    $Templates->{'Content-Variation'} = 0 unless $self->isSubProduct || $self->isMaster;
}

#========================================================================================
# �function     timeFactor
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       my $TimeFactor = $Product->timeFactor($StartDate, $EndDate);
#----------------------------------------------------------------------------------------
# �description  Returns factor to calculate price as quotient for
#               appointment duration devided by default appointment duration.
#               e.g. If default duration is 1 hour and  appointment duration is 30 minutes
#               than timefactor = 0.5
#----------------------------------------------------------------------------------------
# �input        $StartDate | start of appointment | DateTime
# �input        $EndDate | end of appointment | DateTime
# �return       $TimeFactor | factor for price | float
#========================================================================================
sub timeFactor {
    my $self = shift;
    my ($StartDate, $EndDate) = @_;

    my ($DefaultDurationUnit, $DefaultDurationAmount) = $self->businessHours->defaultDuration;
    my $Unit = $self->get('DurationUnit');
    $Unit = $DefaultDurationUnit unless defined $Unit;
    my $TimeFactor = $Unit->get('Factor');
    my $UnitAlias = $Unit->alias;
    $TimeFactor = 1 if $UnitAlias eq 'second';

    my $Amount = $self->get('DurationAmount');
    $Amount = $DefaultDurationAmount unless $Amount;

    return $FLOAT_FORMATTER->round(($EndDate->epoch - $StartDate->epoch) / ($TimeFactor * $Amount)) if defined $TimeFactor;
    # unit month and year have no timefactor (every month has different count of days; years could have different count auf days)

    my $Duration = $EndDate->clone->set_time_zone('UTC') - $StartDate->clone->set_time_zone('UTC'); # ignore glock change

    # started days in the started month
    my $MDays = $Duration->delta_days;
    $MDays++ if $Duration->seconds + $Duration->minutes + $Duration->hours;
    $MDays = DAYS_PER_MONTH if $MDays > DAYS_PER_MONTH;

    my $Factor = ($Duration->years * MONTH_PER_YEAR + $Duration->months) + $MDays/DAYS_PER_MONTH;
    return $FLOAT_FORMATTER->round(($UnitAlias eq 'month' ? $Factor / $Amount : $Factor / MONTH_PER_YEAR / $Amount));
}

#========================================================================================
# �function     appointmentPrice
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Price = $Product->appointmentPrice($Appointment, $hInfo);
#----------------------------------------------------------------------------------------
# �description  Returns product price includes discount and duration data.
#----------------------------------------------------------------------------------------
# �input        $Appointment | calendar appointment (optional if Quantity and Duration set) | object
# �input        $hInfo | session info, possible keys <ul>
#                   <li>Quantity      - quantity            (optional default=MinOrder)- float
#                   <li>CurrencyID    - alpha currency code (optional default=Shop.CurrencyID) - char(3)
#                   <li>TaxModel      - gros or net price   (optional default=Shop.TaxModel)  - int
#                   <li>Duration      - duration of appointment in seconds - int
#                   <li>User          - user how use appointment (optional) - object
#               </ul> | ref.hash
# �return       $Price | calculated price | float
#========================================================================================
sub appointmentPrice {
    my $self = shift;
    my ($Appointment, $hInfo) = @_;
    my $Shop = $self->getSite();
    $hInfo->{'CurrencyID'} = $Shop->get('CurrencyID') unless defined $hInfo->{'CurrencyID'};
    $hInfo->{'TaxModel'} = $Shop->get('TaxModel') unless defined $hInfo->{'TaxModel'};

    if (defined $Appointment) {
        $hInfo->{'Quantity'} = $Appointment->get('Quantity');
        $hInfo->{'StartDate'} = $Appointment->get('StartDate');
        $hInfo->{'EndDate'} = $Appointment->get('EndDate');
    } else {
        $hInfo->{'Quantity'} = $self->get('MinOrder') unless defined $hInfo->{'Quantity'};
        $hInfo->{'StartDate'} = GetCurrentDateTime() unless defined $hInfo->{'StartDate'};
        if (not defined $hInfo->{'EndDate'}) {
            my ($DurationUnit, $DurationAmount) = (defined $hInfo->{'DurationUnit'} and defined $hInfo->{'Duration'})
                ? ($hInfo->{'DurationUnit'}, $hInfo->{'Duration'})
                : $self->businessHours->defaultDuration;
            my $DurationUnits = $DurationUnit->durationUnitPlural;
            $hInfo->{'EndDate'} = AddDuration($hInfo->{'StartDate'}->clone, $DurationUnits, $DurationAmount);
        }
    }

    my $PriceQuantity = $self->getWithParent('PriceQuantity');
    my $Currency = GetInfoCurrency($hInfo->{'CurrencyID'});

    my $Price = ( $self->price($hInfo) * ($hInfo->{'Quantity'}/$PriceQuantity) - $self->discount($hInfo))
                * $self->timeFactor($hInfo->{'StartDate'}, $hInfo->{'EndDate'});
    return $FLOAT_FORMATTER->round($Price, $Currency->{'MinorUnit'});
}

#========================================================================================
# �function     templatePath
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $aObjects = $Product->templatePath;
#----------------------------------------------------------------------------------------
# �description  Extends the Object.templatePath with the super product.
#----------------------------------------------------------------------------------------
# �return       $aObjects | list of objects, classes | ref.array.object
#========================================================================================
sub templatePath {
    my $self = shift;
    my $aPath = $self->SUPER::templatePath;
    my $SuperProduct = $self->superProduct;
    if (defined $SuperProduct) {
        pop @$aPath; # remove object
        push @$aPath, $SuperProduct, $self;
    }
    return $aPath;
}

sub effectedVisibilityAttributes {
    my $class = shift;
    return {
        %{$class->SUPER::effectedVisibilityAttributes},
        'VisibleInNavigation'=>0,
        'ReadPermissionFor'=>0
    };
}

#========================================================================================
# �function     isAllowedChild
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $isAllowed = $Object->isAllowedChild($PossibleChild);
#----------------------------------------------------------------------------------------
# �description  Returns true if an Object could be the parent of a PossibleChild.
#----------------------------------------------------------------------------------------
# �input        $PossibleChild | possible child object | object
# �return       $isAllowed | always false | boolean
#========================================================================================
sub isAllowedChild($) {
    my $self = shift;
    my ($PossibleChild) = @_;
    return 0;
}

#========================================================================================
# �function     createURI
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $URI = $Object->createURI($Name, $Parent, $LanguageID);
#----------------------------------------------------------------------------------------
# �description  Returns a suggestion for a short url.
#----------------------------------------------------------------------------------------
# �input        $Name | localized name | string
# �input        $Parent | parent object | object
# �input        $LanguageID | language id | integer
# �return       $URI | suggested URI | string
#========================================================================================
sub createURI {
    my $class = shift;
    my ($Name, $Parent, $LanguageID) = @_;

    return '' unless defined $Name;
    return '' if $Name eq '';
    return '' if $Parent->alias eq 'SubProducts';
    return $class->SUPER::createURI($Name, $Parent, $LanguageID);
}

#========================================================================================
# �function     defaultObject
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $ContentObject = $Object->defaultObject();
#----------------------------------------------------------------------------------------
# �description  Returns the default object (has to be overwritten for products with variation).
#----------------------------------------------------------------------------------------
# �return       $ContentObject | product or
#                                default subproduct for product with variation | object
#========================================================================================
sub defaultObject {
    my $self = shift;
    my $DefaultSubProduct = $self->get('DefaultSubProduct');
    return $DefaultSubProduct if defined $DefaultSubProduct;
    return $self;
}

#========================================================================================
# �function     setPrepayment
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->setPrepayment($CurrencyID, $TaxModel, $Value);
# �example      $Product->setPrepayment('EUR', 1, 19.99);
# �example      $Product->setPrepayment('EUR', 1, undef);
#----------------------------------------------------------------------------------------
# �description  Sets the prepayment of this product.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $Value | price. A value of undef deletes the price. | float
#========================================================================================
sub setPrepayment {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $Value) = @_;
    TestString('CurrencyID', $CurrencyID);

    # get the default tax model from the shop
    my $Shop = $self->getSite();
    $TaxModel = $Shop->get('InputTaxModel') unless defined $TaxModel;

    if( defined $Value ) {
        TestFloat('Value', $Value);
        # convert gross price to net if necessary
        my $Postfix = $self->taxMatrix->instanceOf('TaxMatrixDigital') ? 'Digital' : '';
        my $TaxClass = $self->getWithParent('TaxClass'.$Postfix);

        $Value = $self->taxMatrix->gross2net($CurrencyID, $Value,
                                            {'TaxModel' => $TaxModel,
                                             'TaxClass' => $TaxClass,
                                             'Shop'     => $Shop,
                                            });
    }

    # set prepayment
    my $ProductID = $self->id;
    if ( ExistsProductPrepayment($ProductID, $CurrencyID) ) {
        if (defined $Value) {
            #Update
            UpdateProductPrepayment($ProductID, $CurrencyID, {'Value' => $Value});
        } else {
            #Delete
            DeleteProductPrepayment($ProductID, $CurrencyID);
        }
    } elsif (defined $Value) {
        #Insert
        InsertProductPrepayment({'ProductID' => $ProductID, 'CurrencyID' => $CurrencyID, 'Value' => $Value});
    }
}

#========================================================================================
# �function     getPrepayment
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Value = $Product->getPrepayment($CurrencyID, $TaxModel);
# �example      $Value = $Product->getPrepayment('EUR', 1);
#----------------------------------------------------------------------------------------
# �description  Returns the prepayment of this product.
#               Returns undef if the product does not have a prepayment.
#----------------------------------------------------------------------------------------
# �input        $CurrencyID | currency code | string
# �input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax)
#                       (optional, default=Product.Shop.InputTaxModel) | int
# �input        $ExtraPrecision | extra precision | int
# �return       $Prepayment | price | float
#========================================================================================
sub getPrepayment {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $ExtraPrecision) = @_;
    TestString('CurrencyID', $CurrencyID);
    $TaxModel = $self->getSite()->get('InputTaxModel') unless defined $TaxModel;

    my $hProductPrepayment = ExistsProductPrepayment($self->id, $CurrencyID) ? GetInfoProductPrepayment($self->id, $CurrencyID) : undef;
    return undef unless defined $hProductPrepayment;
    my $Prepayment = $hProductPrepayment->{'Value'};
    return undef unless defined $Prepayment;

    # convert net price to gross if necessary
    my $Postfix = $self->taxMatrix->instanceOf('TaxMatrixDigital') ? 'Digital' : '';
    my $TaxClass = $self->getWithParent('TaxClass'.$Postfix);

    $Prepayment = $self->taxMatrix->net2gross($CurrencyID, $Prepayment,{'TaxModel' => $TaxModel,
                                                              'TaxClass' => $TaxClass,
                                                              'Shop'     => $self->getSite(),
                                                             }, $ExtraPrecision);
    return $Prepayment;
}

#========================================================================================
# �function     appointmentPrepayment
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Price = $Product->appointmentPrepayment($Appointment, $hInfo);
#----------------------------------------------------------------------------------------
# �description  Returns product price includes discount and duration data.
#----------------------------------------------------------------------------------------
# �input        $Appointment | calendar appointment (optional if Quantity and Duration set) | object
# �input        $hInfo | session info, possible keys <ul>
#                   <li>Quantity      - quantity            (optional default=MinOrder)- float
#                   <li>CurrencyID    - alpha currency code (optional default=Shop.CurrencyID) - char(3)
#                   <li>TaxModel      - gros or net price   (optional default=Shop.TaxModel)  - int
#                   <li>Duration      - duration of appointment in seconds - int
#                   <li>User          - user how use appointment (optional) - object
#               </ul> | ref.hash
# �return       $Price | calculated price | float
#========================================================================================
sub appointmentPrepayment {
    my $self = shift;
    my ($Appointment, $hInfo) = @_;
    my $Shop = $self->getSite();
    $hInfo->{'CurrencyID'} = $Shop->get('CurrencyID') unless defined $hInfo->{'CurrencyID'};
    $hInfo->{'TaxModel'} = $Shop->get('TaxModel') unless defined $hInfo->{'TaxModel'};

    my $PrepaymentType = $self->get('PrepaymentType');
    return undef unless defined $PrepaymentType;
    my $Currency = GetInfoCurrency($hInfo->{'CurrencyID'});
    if ($PrepaymentType eq 'FIX') {
        my $Prepayment = $self->getPrepayment($hInfo->{'CurrencyID'}, $hInfo->{'TaxModel'}, 1);
        return undef unless defined $Prepayment;
        return $FLOAT_FORMATTER->round($Prepayment * $hInfo->{'Quantity'}, $Currency->{'MinorUnit'});
    } elsif ($PrepaymentType eq 'PERCENT') {
        my $PercentValue = $self->get('PrepaymentValue');
        return undef unless defined $PercentValue;
        return $FLOAT_FORMATTER->round($self->appointmentPrice($Appointment, $hInfo) * $PercentValue / PERCENT, $Currency->{'MinorUnit'});
    }
    return undef;
}

#========================================================================================
# �function     renumberDownloadProductMaps
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Product->renumberDownloadProductMaps;
#----------------------------------------------------------------------------------------
# �description  Renumbers the downloadfiles assigned to a product.
#========================================================================================
sub renumberDownloadProductMaps {
    my $self = shift;

    my $aMaps = $self->get('DownloadProductMaps'); # sorted already
    my $position = POSITION_STEP;
    foreach my $Map(@$aMaps) {
        $Map->set({'Position', $position});
        $position += POSITION_STEP;
    }
    return;
}

#========================================================================================
# �function     classTemplate
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Name = $Object->classTemplate($Name, $ObjectPageType);
#----------------------------------------------------------------------------------------
# �description  This function is used to overwrite the given template for the purpose that
#               each content object can defined there own template (visualisation).
#               Function overwrites DE_EPAGES::Object::API::Object::Object->classTemplate
#               for EPG-16237. By merchand created subclasses of product should not used
#               for TemplateName.
#----------------------------------------------------------------------------------------
# �input        $Name | include name | string
# �input        $ObjectPageType | object related pagetype | DE_EPAGES::Presentation::API::Object::ObjectPageType
# �return       $Name | include name | string
#========================================================================================
sub classTemplate {
    my $self = shift;
    my ($TemplateName, $ObjectPageType) = @_;

    my $ExistsTemplate = 0;
    my $Class = LoadClassByAlias('Product');
    my $PageType = $ObjectPageType->get('PageType');
    while (not $ExistsTemplate and defined $Class) {
        my $ClassAlias = $Class->alias;
        if ($PageType->existsTemplate($TemplateName.$ClassAlias)) {
            $ExistsTemplate = 1;
            $TemplateName.=$ClassAlias;
        }
        $Class = $Class->baseClass;
    }
    return $TemplateName;
}

#========================================================================================
# �function     exportableChildren
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $aChildren = $Object->exportableChildren
#----------------------------------------------------------------------------------------
# �example      $Description = $Object->exportableChildren;
#----------------------------------------------------------------------------------------
# �description  Returns all children of the object which should be exported.
#----------------------------------------------------------------------------------------
# �return       $aChildren | children of the object | array ref
#========================================================================================
sub exportableChildren {
    my $self = shift;

    my $aChildren = $self->SUPER::exportableChildren;
    my $hParams = {
        'Children' => $aChildren,
    };
    TriggerHook('OBJ_ProductExportableChildren', $hParams);

    return $hParams->{'Children'};
}

#========================================================================================
# �function     getWebUrl
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $WebUrl = $Object->getWebUrl($hParams);
#----------------------------------------------------------------------------------------
# �description  Construct the web url of a given object dependend of the following parameters:
#----------------------------------------------------------------------------------------
# �input        $hParams | attributes of new object
#                       <ul>
#                         <li>SSL - SSL web url - boolean</li>
#                         <li>Section - Storefront | Mobile | Admin - string</li>
#                       </ul> | hash ref
# �return       $WebUrl | web url of the object | string
#========================================================================================
sub getWebUrl {
    my $self = shift;
    my ($hParams) = @_;
    my $WebUrl;

    if ($self->getSite()->get('IsUnityRegistered')) {
        $WebUrl = $self->getWebUrlUnity($hParams);
    } else {
        $WebUrl = $self->SUPER::getWebUrl($hParams);
    }

    return $WebUrl;
}

#========================================================================================
# �function     getShopUrl
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $ShopUrl = $Object->getShopUrl($hParams, $hOptions, $Servlet);
#----------------------------------------------------------------------------------------
# �description  Construct the shop url of a given object dependend of the following parameters:
#----------------------------------------------------------------------------------------
# �input        $hParams | url get parameters; values may be strings or arrays of strings | ref.hash
# �input        $hOptions | with following keys:
#               <ul>
#                   <li>Type          - one of 'sf', 'admin', 'mobile', 'sf/mobile',
#                   if type is 'sf/mobile' a .mobile will be generated if the servlet is defined and the client
#                   is a mobile device, otherwise a .sf url is generated; defaults to 'sf' - string
#                   <li>UseSSL        - use https as protocol if SSL enabled for shop / database (optional, defaults to 0) - boolean
#                   <li>UseObjectPath - use object path rather than objectid (optional, defaults to 0) - boolean
#                   <li>AddToken      - add security token or not (optional, defaults to 0) - boolean
#                   <li>Token         - the security token (required AddToken is true) - String
#                   <li>Cookies       - add cookies to url or not (optional, defaults to 0) - boolean
#                   <li>AddCookies    - cookie values (required if AddCookies is true) - hash.ref
#                   <li>NoUTF8        - do not use utf-8 for encoding get parameters (optional, defaults to 0) - boolean
#                   <li>IsMobileServlet - is the client a mobile device or not - boolean
#               </ul>  | ref.hash
# �input        $Shop | the current shop | object
# �return       $url | the url | string
#========================================================================================
sub getShopUrl {
    my $self = shift;
    my ($hParams, $hOptions, $Shop) = @_;
    my $ShopUrl;

    if ($hOptions->{'Type'} ne 'admin' && $Shop->get('IsUnityRegistered')) {
        $ShopUrl = $self->get('WebUrlUnitySSL');
    } else {
        $ShopUrl = $self->SUPER::getShopUrl($hParams, $hOptions, $Shop);
    }

    return $ShopUrl;
}

#========================================================================================
# �function     setRelativeStockLevel
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $Object->setRelativeStockLevel($Object, $RelativeStockLevel);
#----------------------------------------------------------------------------------------
# �description  Change the StockLevel of the given product by the relative value and
#               update the cache.
#----------------------------------------------------------------------------------------
# �input        $Object | url get parameters; values may be strings or arrays of strings | ref.hash
# �input        $RelativeStockLevel | with following keys:
#               <ul>
#                   <li>RelativeStockLevel - the relative value of a stock level change
#               </ul>  | ref.hash
#========================================================================================
sub setRelativeStockLevel {
    my $self = shift;
    my($Object, $RelativeStockLevel) = @_;

    my $StockLevel = UpdateStockLevel($Object, $RelativeStockLevel);

    $Object->{'Attributes'}{'_StockLevel'} = $StockLevel;

    # Reset DAL cache for stocklevel
    my $DAL_PACKAGE = DE_EPAGES::Database::API::DAL->load('DE_EPAGES::Product::API::Object::Product');
    $DAL_PACKAGE->cacheUpdate($Object->id, $Object->{'Attributes'}, 0);

    return;
}

#========================================================================================
# �function     collectAllNonImageFiles
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $aAllNonImageFiles = $Product->collectAllNonImageFiles();
#----------------------------------------------------------------------------------------
# �description  Collect the non image files (EnergyLabel and ProductDataSheets)
#               for all related products (Sub and Master).
#----------------------------------------------------------------------------------------
# �return       $aAllNonImageFiles | all non image files | ref.array
#========================================================================================
sub collectAllNonImageFiles {
    my $self = shift;

    my @AllNonImageFileNames;
    my $SuperProduct = $self->get('SuperProduct') // $self;

    push @AllNonImageFileNames,
        @{ $SuperProduct->collectNonImageFiles() };

    foreach my $SubProduct (@{ $SuperProduct->get('SubProducts') }) {
        push @AllNonImageFileNames,
            @{ $SubProduct->collectNonImageFiles() };
    }

    return \@AllNonImageFileNames;
}

#========================================================================================
# �function     collectNonImageFiles
# �state        public
#----------------------------------------------------------------------------------------
# �syntax       $aNonImageFiles = $Product->collectNonImageFiles();
#----------------------------------------------------------------------------------------
# �description  Collect the non image files (EnergyLabel and ProductDataSheets)
#               of a single product.
#----------------------------------------------------------------------------------------
# �return       $aAllNonImageFiles | all non image files | ref.array
#========================================================================================
sub collectNonImageFiles {
    my $self = shift;

    my $Shop = $self->getSite();
    my @ShopLanguageIDs = map { $_->{'LanguageID'} } @{ $Shop->get('Languages') };

    my @NonImageFileNames;

    my $EnergyLabelSourceFile = $self->get('_EnergyLabelSourceFile');
    push @NonImageFileNames, $EnergyLabelSourceFile if defined $EnergyLabelSourceFile;

    foreach my $LanguageID (@ShopLanguageIDs) {
        my $ProductDataSheet = $self->get('_ProductDataSheet', $LanguageID);
        push @NonImageFileNames, $ProductDataSheet if defined $ProductDataSheet;
    }

    return \@NonImageFileNames;
}

sub setTaxClassMapping {
    my $self = shift;
    my ($Shop, $hTaxClassMapping) = @_;

    my $TaxMatrix = $Shop->get('TaxMatrix');
    $hTaxClassMapping = $TaxMatrix->buildTaxClassMappingStructure($self, $Shop) unless $hTaxClassMapping;

    # Delete TaxRates before save
    foreach my $hTaxClassMappingEntry (@{$hTaxClassMapping->{'preselected'}}) {
        delete $hTaxClassMappingEntry->{'taxClass'}->{'rate'} if defined $hTaxClassMappingEntry->{'taxClass'}->{'rate'};
    }
    foreach my $hTaxClassMappingEntry (@{$hTaxClassMapping->{'optional'}}) {
        delete $hTaxClassMappingEntry->{'taxClass'}->{'rate'} if defined $hTaxClassMappingEntry->{'taxClass'}->{'rate'};
    }

    $self->set({'TaxClassMapping' => $hTaxClassMapping});
}

sub _GetTaxAliasFromTaxClassMapping {
    my ($ProductTaxClassMapping, $TaxMatrixID) = @_;
    my $TaxClassAlias;
    foreach my $hTaxMapping (@{$ProductTaxClassMapping->{'preselected'}}) {
        my $ProductTaxMatrixID = $hTaxMapping->{'taxMatrixId'};
        if ($TaxMatrixID == $ProductTaxMatrixID) {
            $TaxClassAlias = $hTaxMapping->{'taxClass'}->{'alias'};
        }
    }
    foreach my $hTaxMapping (@{$ProductTaxClassMapping->{'optional'}}) {
        my $ProductTaxMatrixID = $hTaxMapping->{'taxMatrixId'};
        if ($TaxMatrixID == $ProductTaxMatrixID) {
            $TaxClassAlias = $hTaxMapping->{'taxClass'}->{'alias'};
        }
    }

    return $TaxClassAlias;
}


1;
