#========================================================================================
# package      DE_EPAGES::ShopMigration::API::MigrationShop
# state        public
#----------------------------------------------------------------------------------------
# description  This package provides the functions regarding migration shop and its prepshop.
#========================================================================================
package DE_EPAGES::ShopMigration::API::MigrationShop;
use base Exporter;
use strict;
use 5.010; # Needed because we use state vars

our @EXPORT_OK = qw(
    CreatePrepShop
    ExistsPrepShop
    SetShopFeatureValues
    TransferEpagesUI
    FindAllShopsWithMigrationStatus
    ExportShopDataForMigration
    RunPreFlightChecks
    RunPreFlightChecksInitial
    RunPreFlightChecksFinal
    ImportShopDataForMigration
    PreMigrationImportTasks
    PreMigrationImportTasksInitial
    PreMigrationImportTasksFinal
    PostMigrationImportTasks
    PostMigrationImportTasksInitial
    PostMigrationImportTasksFinal
    SendMigrationConfirmationMail
    ExistsFileExportDirectory
);

use DE_EPAGES::Core::API::Error qw(Error ExistsError GetError);
use DE_EPAGES::Core::API::Log qw(LogError LogDebug);
use DE_EPAGES::Core::API::TestParameter qw(TestObject TestArray TestString);
use DE_EPAGES::Database::API::Config qw(DatabaseConnectionData);
use DE_EPAGES::Database::API::Connection qw(RunOnStore GetCurrentDBHandle);
use DE_EPAGES::Core::API::File qw(RenameDirectory ExistsDirectory AbsolutePath);
use DE_EPAGES::Object::API::Domain qw(ConstructUnityDomainName GetUnitySubDomain);
use DE_EPAGES::Object::API::Factory qw(LoadObjectByPath Transaction LoadRootObject LoadObject LoadClassByAlias ExistsObject);
use DE_EPAGES::Object::API::Table::Language qw(GetInfoLanguage);
use DE_EPAGES::Object::API::Table::Locale qw(GetInfoLocale);
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw(ImportObjects);
use DE_EPAGES::Object::API::XML::AttributeImportHandler qw(ImportAttributes);
use DE_EPAGES::ShopMigration::API::Cartridge;
use DE_EPAGES::ShopMigration::API::Constants qw(
    PREPSHOPS_FOLDER_ALIAS
    @SHOPMIGRATION_PREPSHOP_FEATURES
    %SHOPMIGRATION_PREPSHOP_MODIFIED_FEATURES
    %SHOPMIGRATION_SHOPTYPE_MAPPING
    SHOPMIGRATION_INITIAL_IN_PROGRESS
    SHOPMIGRATION_INITAL_PREPARED
    SHOPMIGRATION_FINAL_ALLOW
    SHOPMIGRATION_FINAL_IN_PROGRESS
    SHOPMIGRATION_FINAL_SHOPTYPECHANGE_ALLOW
    EXPORT_IMPORT_ORDER
);
use DE_EPAGES::RemoteSearch::API::RemoteSearchImpl qw(GetOrCreateSearchIndex); ##no critic (RequireCorrectImports)
use DE_EPAGES::ShopMigration::API::ShopCommonData qw(GetBasicShopData SetBasicShopData);
use DE_EPAGES::ShopMigration::API::CustomerGroupsCommon qw (DeleteCustomerGroupsPrepShop CorrectCustomerGroupsPriceListsPrepShop);
use DE_EPAGES::ShopMigration::API::PriceListsCommon qw (DeletePriceListsPrepShop CreateAdditionalPriceListsPrepShop);
use DE_EPAGES::ShopMigration::API::PriceListQuantityDiscountsCommon qw (RemoveQuantityDiscountsForUnityBulkPrices);
##no critic (RequireCorrectImports)
use DE_EPAGES::ShopMigration::API::JSONLines::ShopDataExportImport qw(ImportShopDataJSON ExportShopDataJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::MailTypesExportImport qw(ImportMailTypesJSON ExportMailTypesJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::LegalPagesExportImport qw(ImportLegalPagesJSON ExportLegalPagesJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::CategoriesExportImport qw(ImportCategoriesJSON ExportCategoriesJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::ProductTypesExportImport qw(ImportProductTypesJSON ExportProductTypesJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::ProductsExportImport qw(ImportProductsJSON ExportProductsJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::CustomerGroupsExportImport qw(ImportCustomerGroupsJSON ExportCustomerGroupsJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::PriceListsExportImport qw(ImportPriceListsJSON ExportPriceListsJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::PriceListScalePricesExportImport qw(ImportPriceListScalePricesJSON ExportPriceListScalePricesJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::PriceListQuantityDiscountsExportImport qw(ImportPriceListQuantityDiscountsJSON ExportPriceListQuantityDiscountsJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::PriceListGlobalValueDiscountsExportImport qw(ImportPriceListGlobalValueDiscountsJSON ExportPriceListGlobalValueDiscountsJSON);
use DE_EPAGES::ShopMigration::API::JSONLines::CategoryProductMapsExportImport qw(
    ImportCategoryProductMapsJSON ExportCategoryProductMapsJSON
);
use DE_EPAGES::ShopMigration::API::JSONLines::CrossSellingsExportImport qw(ImportCrossSellingsJSON ExportCrossSellingsJSON);
use DE_EPAGES::ShopMigration::API::ShortURL qw(AddDeletedShortUrlRedirections);
## use critic
use DE_EPAGES::ShopMigration::API::MigrationReason qw(
    GetMaximumFeatureValuesForShopMigration
    HasReasonDomainWithoutSSL
    HasReasonMaxCategoryDepth
    HasReasonCurrencies
    HasReasonCurrencyMatchesLocale
    HasReasonVariations
    HasReasonMaxCustomerGroups);
use DE_EPAGES::ShopMigration::API::ResetShopHandler qw(ResetBaseShop);
use DE_EPAGES::Trigger::API::Trigger qw(TriggerHook);

our $DELETE_PREP_SHOP = 0;

#ERRORDEFINITION
sub ErrorDefinition() {
    return {
        'NOT_A_BASESHOP'               => 'The shop #Alias is not a base shop.',
        'PREPSHOP_EXISTS'              => 'The shop #Alias has already a prepshop at prepshops folder.',
        'NO_PREPSHOP'                  => 'The shop #Alias has no prepshop.',
        'MIGRATIONSTATUS_NOT_EXISTING' => 'The attribute MigrationStatus does not exists in class shop.',
    };
}
#ERRORDEFINITION END

#========================================================================================
# function     SetShopFeatureValues
# state        intern
#----------------------------------------------------------------------------------------
# syntax       SetShopFeatureValues($Shop, $ahFeatureValues);
#----------------------------------------------------------------------------------------
# description  sets the shopfeatures for shop
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | object
# input        $ahFeatureValues | array of ShopFeatures to set | ref.array.hash.string
#========================================================================================
sub SetShopFeatureValues {
    my ($Shop, $ahFeatureValues) = @_;
    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');
    TestArray('ahFeatureValues', $ahFeatureValues);

    my $Features = LoadObjectByPath('/Features');
    Transaction(sub {
        foreach my $hFeatureValue (@$ahFeatureValues) {
            my $MaxValue = $hFeatureValue->{'MaxValue'};
            my $IsVisible = exists $hFeatureValue->{'IsVisible'} ? $hFeatureValue->{'IsVisible'} : 1;
            my $Feature;
            if (exists $hFeatureValue->{'Feature'}) {
                $Feature = $hFeatureValue->{'Feature'};
            } elsif (exists $hFeatureValue->{'Alias'}) {
                $Feature = $Features->child($hFeatureValue->{'Alias'}) if ($Features->existsChild($hFeatureValue->{'Alias'}));
            } elsif (exists $hFeatureValue->{'FeatureID'}) {
                $Feature = LoadObject($hFeatureValue->{'FeatureID'}) if ExistsObject($hFeatureValue->{'FeatureID'});
            }
            if (defined $Feature) {
                $Shop->featureMaxValue($Feature, $MaxValue, $IsVisible);
            } elsif ($MaxValue) {
                Error('FeatureNotAvailable', {
                    'FeatureAlias' => $hFeatureValue->{'Alias'},
                    'FeatureID'    => $hFeatureValue->{'FeatureID'},
                });
            }
        }
    });
    return;
}

#========================================================================================
# function     ExistsPrepShop
#----------------------------------------------------------------------------------------
# syntax       $Exists = ExistsPrepShop( $Shop );
#----------------------------------------------------------------------------------------
# description  Returns true if a PrepShop for this Shop exists in the database.
#----------------------------------------------------------------------------------------
# input        $Shop | base shop - shop object | object
# return       $Exists | true if the prep shop exists | boolean
#========================================================================================
sub ExistsPrepShop {
    my ($Shop) = @_;
    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');

    my $Folder = _getPrepShopFolder();
    if ($Folder->existsChild(_getPrepShopAlias($Shop->alias))) {
        return 1;
    } else {
        return 0;
    }
}

#========================================================================================
# function     CreatePrepShop
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $PrepShop = CreatePrepShop($Shop);
#----------------------------------------------------------------------------------------
# description  creates a PrepShop if not existing and returns it
#----------------------------------------------------------------------------------------
# input        $Shop | base shop - shop object | object
#----------------------------------------------------------------------------------------
# return       $PrepShop | prep shop - shop object | object
#========================================================================================
sub CreatePrepShop {
    my ($Shop) = @_;

    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');
    if ($Shop->parent->alias ne 'Shops' or $Shop->get('IsUnityRegistered')) {
        Error('NOT_A_BASESHOP', { Alias => $Shop->alias });
    }
    if (ExistsPrepShop($Shop)) {
        Error('PREPSHOP_EXISTS', { Alias => $Shop->alias });
    }
    my $hShopData = GetBasicShopData($Shop);

    $Shop->set({ 'MigrationStatus' => SHOPMIGRATION_INITIAL_IN_PROGRESS });
    my $PrepShopAlias = _getPrepShopAlias($hShopData->{'Attributes'}{'Alias'});
    #use UnitySlug of Base Shop
    my $UnitySlug = GetUnitySubDomain($hShopData->{'Attributes'}{'Alias'});
    #create prepshop
    my $PrepShopFolder = _getPrepShopFolder();
    my $ShopTypeRef = _getShopTypeMappingForShop($Shop);
    my $ahFeatures = _getPrepShopFeaturesByShopTypeRef($ShopTypeRef);
    _modifyPrepShopFeatureMaxValues($ahFeatures);
    # UNITY-6905 check max languages available in shoptype
    my $MaxFeatureLanguages = 0;
    foreach my $hFeatureValue (@$ahFeatures) {
        next if $hFeatureValue->{'Alias'} ne 'Languages';
        $MaxFeatureLanguages = $hFeatureValue->{'MaxValue'};
        last;
    }
    my $ActiveLangs = scalar @{$hShopData->{'ActiveLanguages'}};
    if ($ActiveLangs > $MaxFeatureLanguages) {
        Error(sprintf('Feature Languages not large enough (required: %s is: %s)', $ActiveLangs, $MaxFeatureLanguages));
    }
    my $PrepShop = $PrepShopFolder->insertChild('Shop', {
        'Alias'                    => $PrepShopAlias,
        'WebServerScriptNamePart'  => $PrepShopAlias,
        'UnitySlug'                => $UnitySlug,
        'ShopTypeRef'              => $ShopTypeRef,
        'InitialShopWizardStartup' => 0,
        'DomainName'               => ConstructUnityDomainName($hShopData->{'Attributes'}{'Alias'}),
        'HasSSLCertificate'        => 1,
        'IsMigrationPrepShop'      => 1,
        'LocaleID'                 => $hShopData->{'Attributes'}{'LocaleID'}
    });
    SetShopFeatureValues($PrepShop, $ahFeatures);
    _importPrepShopDefaults($PrepShop);
    SetBasicShopData($PrepShop, $hShopData);
    sleep 10; ## no critic qw( ProhibitMagicNumbers )
    eval {$PrepShop->registerUnity;};
    if (ExistsError()) {
        my $Error = GetError();
        #if registerunity does not work, delete prepshop
        $PrepShop->delete;
        Error($Error);
    }
    #set it again because we need to add additional languages again after registerUnity
    SetBasicShopData($PrepShop, $hShopData);

    _removePagesinUI($UnitySlug);

    # Add now domain to pooldb (works only after registerunity with urlschema domain)
    $PrepShop->set({ 'WebServerScriptNamePart' => $PrepShopAlias, 'MigrationShop' => $Shop });

    my $hInfo = {
        'Shop' => $Shop,
        'PrepShop' => $PrepShop,
    };
    TriggerHook('API_CreatePrepShop', $hInfo);

    $Shop->set({ 'MigrationPrepShop' => $PrepShop });
    return $PrepShop;
}

#========================================================================================
# function     TransferEpagesUI
# state        public
#----------------------------------------------------------------------------------------
# syntax       TransferEpagesUI($Shop);
#----------------------------------------------------------------------------------------
# description  assigns epages-ui from prepshop to migrationshop
#               unset IsUnityRegistered from prepshop
#----------------------------------------------------------------------------------------
# input        $Shop | base shop - shop object | object
#========================================================================================
sub TransferEpagesUI {
    my ($Shop) = @_;

    my $PrepShop = $Shop->get('MigrationPrepShop');
    Error('NO_PREPSHOP', { Alias => $Shop->alias }) unless defined $PrepShop;

    #add unity feature to base shop
    my $Features = LoadObjectByPath('/Features');
    my $Feature = $Features->child('Unity');
    $Shop->featureMaxValue($Feature, 1, 1);

    #do nearly the same like DE_EPAGES/Shop/API/Object/Shop.pm sub registerUnity
    my %Content;
    #add all the things e.g. ShopNotifications, uneditable and prefilled attributes
    TriggerHook('API_BeforeCreateUnityShop', {
        'Object'  => $Shop,
        'Content' => \%Content
    });
    eval {
        #delete locale (this cannot patched here, this only possible for create shops)
        delete $Content{'locale'};
        # send credentials to the Unity database (via now shop app)
        $PrepShop->sendUnityRequest('PATCH', '/shop', undef, \%Content);
    };
    if (ExistsError()) {
        my $Error = GetError();
        #todo:
        # what should happen in this case ?
        # remove the unity feature? some data is already overwritten (uneditable, prefilled attributes) :-(
        LogError($Error);
        Error($Error);
    } else {
        # unassign now shop:
        $PrepShop->set({ 'DomainName' => undef, 'IsUnityRegistered' => 0 });

        #set domainname again to add unity wildcard domain to pooldb
        my $DomainName = $Shop->get('DomainName');
        $Shop->set({
            'DomainName'        => $DomainName,
            'IsUnityRegistered' => 1,
        });
    }
    return;
}

#========================================================================================
# function     SetMaxFeatureValuesBase
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetMaxFeatureValuesBase($Shop);
#----------------------------------------------------------------------------------------
# description  set maxvalue of features of base shop to prepshop maxvalue,
#               if prepshop max value is higher
#----------------------------------------------------------------------------------------
# input        $Shop | base shop - shop object | object
#========================================================================================
sub SetMaxFeatureValuesBase {
    my ($Shop) = @_;

    my $PrepShop = $Shop->get('MigrationPrepShop');
    Error('NO_PREPSHOP', { Alias => $Shop->alias }) unless defined $PrepShop;
    my $hShopFeatureValues = $Shop->get('FeatureMaxValue');
    my $hPrepShopFeatureValues = $PrepShop->get('FeatureMaxValue');
    foreach my $key (keys %$hPrepShopFeatureValues) {
        if ($hShopFeatureValues->{$key} < $hPrepShopFeatureValues->{$key}) {
            $Shop->featureMaxValue($key, $hPrepShopFeatureValues->{$key});
        }
    }
    return;
}

sub _getShopTypeMappingForShop {
    my ($Shop) = @_;

    my $ShopTypeRefFolder = LoadRootObject()->folder('ShopTypeRefs');
    my $ShopTypeRef = $Shop->get('ShopTypeRef');
    my $ShopTypeRefAlias = defined $ShopTypeRef ? $ShopTypeRef->alias : '_default';
    if (exists $SHOPMIGRATION_SHOPTYPE_MAPPING{$ShopTypeRefAlias}) {
        $ShopTypeRefAlias = $SHOPMIGRATION_SHOPTYPE_MAPPING{$ShopTypeRefAlias};
    } else {
        $ShopTypeRefAlias = $SHOPMIGRATION_SHOPTYPE_MAPPING{'_default'};
    }
    return $ShopTypeRefFolder->child($ShopTypeRefAlias);
}

sub _getPrepShopFeaturesByShopTypeRef {
    my ($ShopTypeRef) = @_;
    my $aShopIDs = LoadClassByAlias('Shop')->findObjects('ByShopTypeRefID', [ $ShopTypeRef->id ]);
    my $FirstShopWithShopTypeRef = scalar(@$aShopIDs) ? LoadObject($aShopIDs->[0]) : undef;
    my $aFeatureValues;
    if (defined $FirstShopWithShopTypeRef) {
        $aFeatureValues = $FirstShopWithShopTypeRef->get('FeatureValues');
        foreach my $hFeature (@$aFeatureValues) {
            $hFeature->{'Alias'} = $hFeature->{'Feature'}->alias;
            $hFeature->{'FeatureID'} = $hFeature->{'Feature'}->id;
        }
    } else {
        $aFeatureValues = \@SHOPMIGRATION_PREPSHOP_FEATURES;
    }
    return $aFeatureValues;
}

sub _modifyPrepShopFeatureMaxValues {
    my ($ahFeatures) = @_;
    foreach my $hFeature (@$ahFeatures) {
        if (exists $SHOPMIGRATION_PREPSHOP_MODIFIED_FEATURES{$hFeature->{'Alias'}}) {
            $hFeature->{'MaxValue'} = $SHOPMIGRATION_PREPSHOP_MODIFIED_FEATURES{$hFeature->{'Alias'}};
        }
    }
    return;
}

sub _importPrepShopDefaults {
    my ($PrepShop) = @_;

    my $Cartridge = DE_EPAGES::ShopMigration::API::Cartridge->new;
    my $Directory = join('/', $ENV{EPAGES_CARTRIDGES}, $Cartridge->cartridgeDirectory, 'Data/PrepShopImport.d');
    foreach my $file (qw(BasicMinUnity.xml)) {
        _importXML($Directory . '/' . $file, $PrepShop->pathString);
    }
    return;
}

sub _removePagesinUI {
    my ($UnitySlug) = @_;

    my $Store = LoadRootObject()->get('StoreName');
    my $UIStore = DatabaseConnectionData($Store)->{InstanceUnityUI};
    if (!defined $UIStore || !$UIStore) {
        Error('UI_STORE_NOT_DEFINED', { Store => $Store })
    } else {
        RunOnStore(
            Store => $UIStore,
            Sub   => sub {
                my $dbi = GetCurrentDBHandle();
                #remove default pages (type='blank') from ui db
                my $sth = $dbi->executeStatement(<<SQL, [ $UnitySlug ]);
DELETE FROM pages
WHERE type='blank'
  AND shopId IN (
    SELECT id FROM shops WHERE shopGateID = (
        SELECT id FROM shopGates WHERE slug=?
    )
)
SQL
                LogDebug(sprintf("Deleting in %s.pages [%s]: affected %s rows\n",
                    $UIStore, $UnitySlug, $sth->rows), 1);
            }
        );
    }
    return;
}

sub _importXML {
    my ($FileName, $Path) = @_;
    TestString('FileName', $FileName);
    TestString('Path', $Path);

    if ($FileName =~ /\bTranslation\.(\w+\.)+(\w\w|\w\w-\w\w)\.xml$/) {
        ImportAttributes($FileName, $Path);
    } else {
        ImportObjects($FileName, $Path);
    }
    return;
}

sub _getPrepShopAlias {
    my ($ShopAlias) = @_;
    return GetUnitySubDomain('PrepShop' . $ShopAlias);
}

sub _getPrepShopFolder {
    return LoadRootObject()->child(PREPSHOPS_FOLDER_ALIAS);
}

#========================================================================================
# function     RunPreFlightChecks($Shop)
# state        public
#----------------------------------------------------------------------------------------
# syntax       RunPreFlightChecks($Shop);
#----------------------------------------------------------------------------------------
# description  runs sanity checks before starting initial/final migration
#----------------------------------------------------------------------------------------
# input        $Shop      | shop object | object
#========================================================================================
sub RunPreFlightChecks {
    my ($Shop) = @_;

    my $Status = $Shop->get('MigrationStatus');
    if ($Status == SHOPMIGRATION_INITIAL_IN_PROGRESS) {
        RunPreFlightChecksInitial($Shop);
    } elsif ($Status == SHOPMIGRATION_FINAL_ALLOW || $Status == SHOPMIGRATION_FINAL_IN_PROGRESS) {
        RunPreFlightChecksFinal($Shop);
    }
}

#========================================================================================
# function     RunPreFlightChecksInitial($Shop)
# state        public
#----------------------------------------------------------------------------------------
# syntax       RunPreFlightChecksInitial($Shop);
#----------------------------------------------------------------------------------------
# description  runs sanity checks before starting initial migration
#----------------------------------------------------------------------------------------
# input        $Shop      | shop object | object
#========================================================================================
sub RunPreFlightChecksInitial {
    my ($Shop) = @_;

    LogDebug('RunPreFlightChecksInitial', 1);
    my $hMaxFeatureValues = GetMaximumFeatureValuesForShopMigration();
    Error('ReasonDomainWithoutSSL') if (HasReasonDomainWithoutSSL($Shop, $hMaxFeatureValues));
    Error('ReasonMaxCategoryDepth') if (HasReasonMaxCategoryDepth($Shop, $hMaxFeatureValues));
    Error('ReasonCurrencies') if (HasReasonCurrencies($Shop, $hMaxFeatureValues));
    Error('ReasonCurrencyMatchesLocale') if (HasReasonCurrencyMatchesLocale($Shop, $hMaxFeatureValues));
    Error('ReasonVariations') if (HasReasonVariations($Shop, $hMaxFeatureValues));
    Error('ReasonCustomerGroups') if (HasReasonMaxCustomerGroups($Shop, $hMaxFeatureValues));
    return;
}

#========================================================================================
# function     RunPreFlightChecksFinal($Shop)
# state        public
#----------------------------------------------------------------------------------------
# syntax       RunPreFlightChecksFinal($Shop);
#----------------------------------------------------------------------------------------
# description  runs sanity checks before starting final migration
#----------------------------------------------------------------------------------------
# input        $Shop      | shop object | object
#========================================================================================
sub RunPreFlightChecksFinal {
    my ($Shop) = @_;

    LogDebug('RunPreFlightChecksFinal', 1);

    my $hMaxFeatureValues = GetMaximumFeatureValuesForShopMigration();
    Error('HasNoSSLCertifcate') if (HasReasonDomainWithoutSSL($Shop, $hMaxFeatureValues));
    my $Prep = $Shop->get('MigrationPrepShop');
    foreach my $Setting (qw(TaxModel InputTaxModel)) {
        Error('TaxModelDifferent', {Setting => $Setting}) if ($Prep->get($Setting) != $Shop->get($Setting));
    }
    return;
}

#========================================================================================
# function     FindAllShopsWithMigrationStatus
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $aShops = FindAllShopsWithMigrationStatus($Status);
#----------------------------------------------------------------------------------------
# description  returns an array of shops with the given migration status
#               see DE_EPAGES::ShopMigration::API::Constants for a list of
#               available status
#----------------------------------------------------------------------------------------
# input        $Status | wanted migration status | integer
#----------------------------------------------------------------------------------------
# return       $aShops | array of shop objects with given status | ref.array.object
#========================================================================================
sub FindAllShopsWithMigrationStatus {
    my ($Status) = @_;

    my $ExistsAttribute = LoadClassByAlias('Shop')->existsAttribute('MigrationStatus');
    Error('MIGRATIONSTATUS_NOT_EXISTING') unless $ExistsAttribute;
    my $MigrationStatusAttributeID = LoadClassByAlias('Shop')->attribute('MigrationStatus')->id;
    my $dbi = GetCurrentDBHandle();
    my $Query = 'SELECT objectid from integerattribute WHERE attributeid=? and value=?';
    my @Params = ($MigrationStatusAttributeID, $Status);
    my $aaRows = $dbi->execute($Query, \@Params, [ 'Integer' ]);
    return [ map {LoadObject($_->[0])} @$aaRows ];
}

#========================================================================================
# function     ExportShopDataForMigration
# state        public
#----------------------------------------------------------------------------------------
# syntax       ExportShopDataForMigration($Shop, $Directory);
#----------------------------------------------------------------------------------------
# description  exports all the JSON data file for a shop (prepshop or baseshop)
#               to the given directory
#----------------------------------------------------------------------------------------
# input        $Shop      | shop object | object
# input        $Directory | output directory | string
#========================================================================================
sub ExportShopDataForMigration {
    my ($Shop, $Directory) = @_;

    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');
    TestString('Directory', $Directory);
    state $Dispatch = { map {$_, \&{'Export' . $_ . 'JSON'}} @{EXPORT_IMPORT_ORDER()} };
    foreach my $DataType (@{EXPORT_IMPORT_ORDER()}) {
        LogDebug(sprintf("Exporting %s to %s\n", $DataType, $Directory), 1);
        $Dispatch->{$DataType}->($Shop, $Directory);
    }
}

#========================================================================================
# function     ImportShopDataForMigration
# state        public
#----------------------------------------------------------------------------------------
# syntax       ImportShopDataForMigration($Shop, $Directory);
#----------------------------------------------------------------------------------------
# description  imports all the JSON data file for a shop (prepshop or baseshop)
#               from a given directory
#----------------------------------------------------------------------------------------
# input        $Shop      | shop object | object
# input        $Directory | output directory | string
# input        $aImportOverride | limit import to these types | ref array string
#========================================================================================
sub ImportShopDataForMigration {
    my ($Shop, $Directory, $aImportOrder) = @_;

    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');
    TestString('Directory', $Directory);
    $aImportOrder //= [];
    my @ImportOrder = scalar @$aImportOrder ? @$aImportOrder : @{EXPORT_IMPORT_ORDER()};
    state $Dispatch = { map {$_, \&{'Import' . $_ . 'JSON'}} @ImportOrder };
    foreach my $DataType (@ImportOrder) {
        LogDebug(sprintf("Importing %s from %s\n", $DataType, $Directory), 1);
        $Dispatch->{$DataType}->($Shop, $Directory);
    }
}

#========================================================================================
# function     PreMigrationImportTasks
# state        public
#----------------------------------------------------------------------------------------
# syntax       PreMigrationImportTasks($Shop);
#----------------------------------------------------------------------------------------
# description  runs pre data import tasks necessary depending on whatever phase
#               this shop is in
#----------------------------------------------------------------------------------------
# input        $Shop      | base shop object | object
#========================================================================================
sub PreMigrationImportTasks {
    my ($Shop) = @_;

    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');
    if ($Shop->get('MigrationStatus') == SHOPMIGRATION_INITIAL_IN_PROGRESS) {
        PreMigrationImportTasksInitial($Shop);
    } elsif ($Shop->get('MigrationStatus') == SHOPMIGRATION_FINAL_ALLOW) {
        PreMigrationImportTasksFinal($Shop);
    }
}

#========================================================================================
# function     PreMigrationImportTasksInitial
# state        public
#----------------------------------------------------------------------------------------
# syntax       PreMigrationImportTasksInitial($Shop);
#----------------------------------------------------------------------------------------
# description  runs pre data import tasks for initial prepshop migration
#----------------------------------------------------------------------------------------
# input        $Shop      | base shop object | object
#========================================================================================
sub PreMigrationImportTasksInitial {
    my ($Shop) = @_;

    my $PrepShop = $Shop->get('MigrationPrepShop');
    LogDebug('Running [PreMigrationImportTasksInitial] DeletePriceListsPrepShop(PrepShop)', 1);
    DeletePriceListsPrepShop($PrepShop);

    LogDebug('Running [PreMigrationImportTasksInitial] DeleteCustomerGroupsPrepShop(PrepShop)', 1);
    DeleteCustomerGroupsPrepShop($PrepShop);
}

#========================================================================================
# function     PreMigrationImportTasksFinal
# state        public
#----------------------------------------------------------------------------------------
# syntax       PreMigrationImportTasksFinal($Shop);
#----------------------------------------------------------------------------------------
# description  runs pre data import tasks for final migration
#----------------------------------------------------------------------------------------
# input        $Shop      | base shop object | object
#========================================================================================
sub PreMigrationImportTasksFinal {
    my ($Shop) = @_;

    LogDebug('Running [PreMigrationImportTasks] MigrationStatus => SHOPMIGRATION_FINAL_IN_PROGRESS', 1);
    $Shop->set({ 'MigrationStatus' => SHOPMIGRATION_FINAL_IN_PROGRESS });

    #move old MediaGallery before it gets deleted by ResetHandler
    LogDebug('Running [PreMigrationImportTasks] BackupMediaGallery()', 1);
    BackupMediaGallery($Shop);

    #delete / reset obsolete data in baseshop
    LogDebug('Running [PreMigrationImportTasks] ResetBaseShop()', 1);
    ResetBaseShop($Shop);

    #transfer epages-ui of prepshop to baseshop -> set IsUnityRegistered
    LogDebug('Running [PreMigrationImportTasks] TransferEpagesUI()', 1);
    TransferEpagesUI($Shop);

    #set maxfeature values of prepshop if they are higher than current base shop values
    LogDebug('Running [PreMigrationImportTasks] SetMaxFeatureValuesBase()', 1);
    SetMaxFeatureValuesBase($Shop);
}

#========================================================================================
# function     PostMigrationImportTasks
# state        public
#----------------------------------------------------------------------------------------
# syntax       PostMigrationImportTasks($Shop);
#----------------------------------------------------------------------------------------
# description  runs post data import tasks necessary depending on whatever this is a
#               prepshop or transfer shop
#----------------------------------------------------------------------------------------
# input        $Shop      | shop object | object
#========================================================================================
sub PostMigrationImportTasks {
    my ($Shop) = @_;

    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');
    if ($Shop->get('MigrationStatus') == SHOPMIGRATION_INITIAL_IN_PROGRESS) {
        PostMigrationImportTasksInitial($Shop);
    } elsif ($Shop->get('MigrationStatus') == SHOPMIGRATION_FINAL_IN_PROGRESS) {
        PostMigrationImportTasksFinal($Shop);
    }
}

#========================================================================================
# function     PostMigrationImportTasksInitial
# state        public
#----------------------------------------------------------------------------------------
# syntax       PostMigrationImportTasksInitial($Shop);
#----------------------------------------------------------------------------------------
# description  runs post data import tasks necessary for initial prepshop migration
#----------------------------------------------------------------------------------------
# input        $Shop      | shop object | object
#========================================================================================
sub PostMigrationImportTasksInitial {
    my ($Shop) = @_;

    my $PrepShop = $Shop->get('MigrationPrepShop');

    LogDebug('Running [PostMigrationImportTasks] GetOrCreateSearchIndex(PrepShop)->initialize', 1);
    GetOrCreateSearchIndex($PrepShop)->initialize();

    LogDebug('Running [PostMigrationImportTasks] CreateAdditionalPriceListsPrepShop(PrepShop)->initialize', 1);
    CreateAdditionalPriceListsPrepShop($PrepShop);

    LogDebug('Running [PostMigrationImportTasks] CorrectCustomerGroupsPriceListsPrepShop(PrepShop)', 1);
    CorrectCustomerGroupsPriceListsPrepShop($PrepShop);

    LogDebug('Running [PostMigrationImportTasks] RemoveQuantityDiscountsForUnityBulkPrices(PrepShop)', 1);
    RemoveQuantityDiscountsForUnityBulkPrices($PrepShop);

    LogDebug('Running [PostMigrationImportTasks] MigrationStatus => SHOPMIGRATION_INITAL_PREPARED', 1);
    $Shop->set({ 'MigrationStatus' => SHOPMIGRATION_INITAL_PREPARED });
}

#========================================================================================
# function     PostMigrationImportTasksFinal
# state        public
#----------------------------------------------------------------------------------------
# syntax       PostMigrationImportTasksFinal($Shop);
#----------------------------------------------------------------------------------------
# description  runs post data import tasks necessary for final migration
#----------------------------------------------------------------------------------------
# input        $Shop      | shop object | object
#========================================================================================
sub PostMigrationImportTasksFinal {
    my ($Shop) = @_;

    LogDebug('Running [PostMigrationImportTasks] GetOrCreateSearchIndex(Shop)->initialize', 1);
    GetOrCreateSearchIndex($Shop)->initialize();

    LogDebug('Running [PostMigrationImportTasks] PrepShop->delete', 1);
    my $PrepShop = $Shop->get('MigrationPrepShop');
    $PrepShop->delete if (defined $PrepShop && $DELETE_PREP_SHOP);

    LogDebug('Running [PostMigrationImportTasks] MigrationStatus => SHOPMIGRATION_FINAL_SHOPTYPECHANGE_ALLOW', 1);
    $Shop->set({ 'IsClosed' => 0, 'MigrationStatus' => SHOPMIGRATION_FINAL_SHOPTYPECHANGE_ALLOW });

    LogDebug('Running [PostMigrationImportTasks] AddDeletedShortUrlRedirections', 1);
    AddDeletedShortUrlRedirections($Shop);

    LogDebug('Running [PostMigrationImportTasks] saveDefaultStyle', 1);
    $Shop->get('DefaultStyle')->writeCSS;

    LogDebug('Running [PostMigrationImportTasks] SendMigrationConfirmationMail', 1);
    SendMigrationConfirmationMail($Shop);
}

#========================================================================================
# function     SendMigrationConfirmationMail
# state        public
#----------------------------------------------------------------------------------------
# syntax       SendMigrationConfirmationMail($Shop);
#----------------------------------------------------------------------------------------
# description  send email to supermerchant and inform that the migration
#               is finished
#----------------------------------------------------------------------------------------
# input        $Shop | base shop - shop object | object
#----------------------------------------------------------------------------------------
# return       $Success | true if successful | boolean
#========================================================================================
sub SendMigrationConfirmationMail {
    my ($Shop) = @_;
    my %MailParams = {};
    my $System = LoadRootObject();
    my $MailType = $System->folder('MailTypes')->child('MigrationConfirmation');
    return unless $MailType->get('IsActive');

    #get SuperMerchant object
    my $SuperMerchant = $Shop->get('SuperMerchant');
    my $Email = $SuperMerchant->get('EMail') ? $SuperMerchant->get('EMail') : $Shop->get('MailFrom');

    my $hVars;
    $hVars->{'User'} = $SuperMerchant->tleHash;

    my $LanguageID = $SuperMerchant->get('LanguageID');
    $MailParams{'TO'} = $Email;
    $MailParams{'FROM'} = $Email;
    $MailParams{'Subject'} = $MailType->get('Subject', $LanguageID);
    $MailParams{'HTMLAllowed'} = 1;

    my $Locale = $SuperMerchant->get('LocaleID') ? $SuperMerchant->get('LocaleID') : $Shop->get('LocaleID');;
    my %RegionalParams = (
        'Locale'   => $Locale,
        'Language' => GetInfoLanguage(GetInfoLocale($Locale)->{'LanguageID'})->{'Code2'}
    );

    eval {
        $MailType->send($SuperMerchant->getSite(), \%MailParams, \%RegionalParams, $hVars);
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogError($Error);
        return 0;
    }

    return 1;
}

sub BackupMediaGallery {
    my ($Shop) = @_;

    my $SourceDir = $Shop->get('MediaGallery')->get('PublicPath');
    my $BackupDir = sprintf('%s.%s', $SourceDir, 'PreFinalMigration');
    RenameDirectory($SourceDir, $BackupDir) if (ExistsDirectory($SourceDir) && !ExistsDirectory($BackupDir));
    return;
}

#========================================================================================
# function     ExistsFileExportDirectory
# state        public
#----------------------------------------------------------------------------------------
# syntax       ExistsFileExportDirectory($Shop, $Dir);
#----------------------------------------------------------------------------------------
# description  checks if the corresponding file export directory exists
#               this is either Files/Base2Prep or Files/Prep2Base depending
#               on $Shop
#----------------------------------------------------------------------------------------
# input        $Shop      | base shop object | object
# input        $Dir       | base directory   | string
# return       $Exists    | diretory exists  | boolean
#========================================================================================
sub ExistsFileExportDirectory {
    my ($Shop, $Dir) = @_;

    my $Exporter = DE_EPAGES::ShopMigration::API::JSONLines::ProductsExportDriver->new((Shop => $Shop, DirName => $Dir));
    my $FilesDir = AbsolutePath($Exporter->dirNameFiles());
    my $Exists =  ExistsDirectory($FilesDir);
    LogDebug(sprintf("Export files dir %s exists: %s\n", $FilesDir, $Exists ? 'yes' : 'no'), 1);
    return $Exists;
}

1;