use strict;
use Getopt::Long;
use Time::HiRes qw(time);
use DE_EPAGES::Core::API::Error qw(ExistsError GetError);
use DE_EPAGES::Core::API::Script qw(RunScript);
use DE_EPAGES::Core::API::Warning qw(Warning);
use DE_EPAGES::Database::API::Connection qw(RunOnStore);
use DE_EPAGES::Object::API::Factory qw(LoadRootObject);
use DE_EPAGES::WebInterface::API::MessageCenter qw(SynchronizeCache);
use DE_EPAGES::ShopMigration::API::ResetShopHandler qw(UnsetLegacyAddressValues);

RunScript('Sub' => \&Main, 'Message' => 'long');

sub Main {

    local $| = 1;
    my %Options = (output => 1);
    my @GetOpt = qw(help output! storename=s passwd=s shop=s test);
    usage("Unkown parameter specified.\n") unless GetOptions(\%Options, @GetOpt);
    usage() if $Options{help};
    for (qw(storename shop)) {
        usage(sprintf("Missing parameter %s\n", $_)) unless $Options{$_};
    }

    my $Output = $Options{output};
    my $time1 = time;
    RunOnStore(
        Store      => $Options{storename},
        DBPassword => $Options{passwd},
        Sub        => sub {
            my $Shop = findShop($Options{shop});
            UnsetLegacyAddressValues($Shop);
        },
    );
    my $time2 = time;
    printf "need %.3f seconds\n", $time2 - $time1 if $Output;

    eval {SynchronizeCache();};
    Warning(GetError()) if ExistsError();
}

sub findShop {
    my ($ShopAlias) = @_;

    my $ShopFolder = LoadRootObject()->child('Shops');
    usage(sprintf('No such shop: %s', $ShopAlias)) unless $ShopFolder->existsChild($ShopAlias);
    my $Shop = $ShopFolder->child($ShopAlias);
    usage(sprintf('"%s" is not a Now shop!', $ShopAlias)) unless $Shop->get('IsUnityRegistered');
    return $Shop;
}

sub usage {
    my ($ErrorNote) = @_;

    print STDERR $ErrorNote . "\n\n" if $ErrorNote;
    print <<END_USAGE;
Description:

Unsets unsupported attributes (from RegistrationSet) on all User/Customer addresses if the shop is a Now shop.

Usage:
    perl $0 [options] [flags] [storename] attribute=value ...

options:
    -passwd     database user password (optional)
    -storename  name of store (uses [storename] if not specified)
    -shop       shop alias
    -nooutput   suppress non-error output

flags:
    -help       show the command line options
    -test       only print out what would be done without actually doing it

Example:
    perl $0 -storename Store -shop DemoShop
END_USAGE
    exit 2;
}

