#========================================================================================
# package      DE_EPAGES::Zettle::API::Object::ZettleEnvironment
# state        public
#----------------------------------------------------------------------------------------
# description  object for calling functions on the ZettleEnvironment
#========================================================================================
package DE_EPAGES::Zettle::API::Object::ZettleEnvironment;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;

use Readonly;

# Names of the attributes used to store token information in ZettleEnvironment object
Readonly my $ACCESS_TOKEN_KEY => 'AccessToken';
Readonly my $REFRESH_TOKEN_KEY => 'RefreshToken';
Readonly my $ACCESS_TOKEN_EXPIRATION_KEY => 'AccessTokenExpiryDate';

use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::PerlTools qw ( GetCurrentDateTime );

#ERRORDEFINITION
sub ErrorDefinition () {
    return {
        'CURRENCY_MISMATCH' =>
          'Shop currency (#ShopCurrency) does not match Zettle currrency (#ZettleCurrency)',
    };
}
#ERRORDEFINITION END

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::Zettle::API::Object::ZettleEnvironment->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;

    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'ReplaceZettleLibrary'          => 0,
        'ExportScheduleInterval'        => 3,
        'ProductIsExporting'            => 0,
        'ExportProductPrices'           => 0,
        'PriceIsExporting'              => 0,
        'InitialExportStatus'           => 0,
        'MarkedForDeletion'             => undef,
    };
}

#========================================================================================
# function     accessToken
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->accessToken();
# syntax       $self->accessToken($AccessToken);
#----------------------------------------------------------------------------------------
# description  Get or set access token.
#----------------------------------------------------------------------------------------
# input        $AccessToken | optional. Access token retrieved from Zettle | string
#----------------------------------------------------------------------------------------
# return       $Result | Stored access token, if $AccessToken argument is not present | string
#========================================================================================
sub accessToken {
    my $self = shift;
    my ($AccessToken) = @_;
    
    return $self->get($ACCESS_TOKEN_KEY) unless defined $AccessToken;

    $self->set({$ACCESS_TOKEN_KEY => $AccessToken});
}

#========================================================================================
# function     refreshToken
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->refreshToken();
# syntax       $self->refreshToken($RefreshToken);
#----------------------------------------------------------------------------------------
# description  Get or set refresh token.
#----------------------------------------------------------------------------------------
# input        $RefreshToken | optional. Refresh token retrieved from Zettle | string
#----------------------------------------------------------------------------------------
# return       $Result | Stored refresh token, if $RefreshToken argument is not present | string
#========================================================================================
sub refreshToken {
    my $self = shift;
    my ($RefreshToken) = @_;
    
    return $self->get($REFRESH_TOKEN_KEY) unless defined $RefreshToken;

    $self->set({$REFRESH_TOKEN_KEY => $RefreshToken});
}

#========================================================================================
# function     accessTokenExpiration
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->accessTokenExpiration();
# syntax       $self->accessTokenExpiration($AccessTokenExpiration);
#----------------------------------------------------------------------------------------
# description  Get or set access token expiration date.
#----------------------------------------------------------------------------------------
# input        $AccessTokenExpiration | optional. Access token expiration date retrieved from Zettle | Datetime
#----------------------------------------------------------------------------------------
# return       $Result | Stored access token expiration date, if $AccessTokenExpiration argument is not present | Datetime
#========================================================================================
sub accessTokenExpiration {
    my $self = shift;
    my ($AccessTokenExpiration) = @_;
    
    return $self->get($ACCESS_TOKEN_EXPIRATION_KEY) unless defined $AccessTokenExpiration;

    $self->set({$ACCESS_TOKEN_EXPIRATION_KEY => $AccessTokenExpiration});
}

#========================================================================================
# function     updateTokenData
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->updateTokenData($AccessToken, $RefreshToken, $ExpirationDate);
#----------------------------------------------------------------------------------------
# description  Update token-related data in Zettle environment.
#----------------------------------------------------------------------------------------
# input        $AccessToken    | Access token | string
# input        $RefreshToken   | Refresh token | string
# input        $ExpirationDate | Access token expiration date | Datetime
#========================================================================================
sub updateTokenData {
    my $self = shift;
    my ($AccessToken, $RefreshToken, $ExpirationDate) = @_;

    $self->set({
        $ACCESS_TOKEN_KEY     => $AccessToken,
        $REFRESH_TOKEN_KEY    => $RefreshToken,
        $ACCESS_TOKEN_EXPIRATION_KEY  => $ExpirationDate,
    });
}

#========================================================================================
# function     setupInitialExport
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->setupInitialExport($InitialExportUUID, $ExportPrices);
#----------------------------------------------------------------------------------------
# description  Update initial export attributes:
#               <ul>
#                 <li> Assign $InitialExportUUID (this enables us to query export status afterwards)
#                 <li> Set 'LastProductExport' to current datetime.
#                 <li> Set 'LastPriceExport' to current datetime if $ExportPrices is TRUE.
#               </ul>
#----------------------------------------------------------------------------------------
# input        $InitialExportUUID | UUID of initial export | string
# input        $ExportPrices | TRUE if prices are to be exported with products, false otherwise | boolean
#========================================================================================
sub setupInitialExport {
    my $self = shift;
    my ($InitialExportUUID, $ExportPrices) = @_;

    if($ExportPrices){
        $self->set({
            'InitialExportUUID' => $InitialExportUUID,
            'LastProductExport' => GetCurrentDateTime(),
            'LastPriceExport'   => GetCurrentDateTime(),
        });
    }else{
        $self->set({
            'InitialExportUUID' => $InitialExportUUID,
            'LastProductExport' => GetCurrentDateTime(),
        });
    }
}

sub lastPriceExport {
    my $self = shift;
    my ($LastPriceExport) = @_;
    
    return $self->get('LastPriceExport') unless defined $LastPriceExport;

    $self->set({'LastPriceExport' => $LastPriceExport});
}

sub lastProductExport {
    my $self = shift;
    my ($LastProductExport) = @_;
    
    return $self->get('LastProductExport') unless defined $LastProductExport;

    $self->set({'LastProductExport' => $LastProductExport});
}

#========================================================================================
# function     initialExportStatus
# state        public
#----------------------------------------------------------------------------------------
# syntax       $InitialExportStatus = $self->initialExportStatus();
# syntax       $self->initialExportStatus($Status);
#----------------------------------------------------------------------------------------
# description  Get or update initial export status.
#----------------------------------------------------------------------------------------
# input        $Status | optional. Status value. See DE_EPAGES::Zettle::API::Constants::%INITIAL_EXPORT_STATUS | integer
#----------------------------------------------------------------------------------------
# return       $Result | Stored initial export status, if $Status argument is not present | integer
#========================================================================================
sub initialExportStatus {
    my $self = shift;
    my ($Status) = @_;
    
    return $self->get('InitialExportStatus') unless defined $Status;

    $self->set({'InitialExportStatus' => $Status});
}

#========================================================================================
# function     lastStockSync
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LastStockSync = $self->lastStockSync();
# syntax       $self->lastStockSync($LastStockSync);
#----------------------------------------------------------------------------------------
# description  Get or update last local stock update datetime.
#----------------------------------------------------------------------------------------
# input        $LastStockSync | optional. Last stock sync datetime | Datetime
#----------------------------------------------------------------------------------------
# return       $Result | Stored last stock sync datetime, if $LastStockSync argument is not present | Datetime
#========================================================================================
sub lastStockSync {
    my $self = shift;
    my ($LastStockSync) = @_;
    
    return $self->get('ProductStockLevelUpdateTime') unless defined $LastStockSync;

    $self->set({'ProductStockLevelUpdateTime' => $LastStockSync});
}

#========================================================================================
# function     lastStockSyncZettle
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LastStockSync = $self->lastStockSyncZettle();
# syntax       $self->lastStockSyncZettle($LastStockSync);
#----------------------------------------------------------------------------------------
# description  Get or update last Zettle stock update datetime.
#----------------------------------------------------------------------------------------
# input        $LastStockSync | optional. Last stock sync datetime | Datetime
#----------------------------------------------------------------------------------------
# return       $Result | Stored last stock sync datetime, if $LastStockSync argument is not present | Datetime
#========================================================================================
sub lastStockSyncZettle {
    my $self = shift;
    my ($LastStockSync) = @_;
    
    return $self->get('ProductStockLevelUpdateTimeZettle') unless defined $LastStockSync;

    $self->set({'ProductStockLevelUpdateTimeZettle' => $LastStockSync});
}

#========================================================================================
# function     initialExportUUID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ExportUUID = $self->initialExportUUID();
# syntax       $self->initialExportUUID($ExportUUID);
#----------------------------------------------------------------------------------------
# description  Get or update initial export UUID.
#----------------------------------------------------------------------------------------
# input        $ExportUUID | optional. Initial export UUIDv1 | string
#----------------------------------------------------------------------------------------
# return       $ExportUUID | Stored initial export UUID if $ExportUUID parameter is not present | string
#========================================================================================
sub initialExportUUID {
    my $self = shift;
    my ($ExportUUID) = @_;

    return $self->get('InitialExportUUID') unless defined $ExportUUID;

    $self->set({'InitialExportUUID' => $ExportUUID});
}

#========================================================================================
# function     setMerchantInfo
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->setMerchantInfo($hOptions);
#----------------------------------------------------------------------------------------
# description  Configure merchant account information in Zettle environment.
#----------------------------------------------------------------------------------------
# input        $hOptions | options setup info with keys:
#                          <ul>
#                              <li>uuid - organization UUID - string
#                              <li>email - zettle login email - string
#                              <li>country - ISO3166-1 code (i.e: GB) - char(2)
#                              <li>currency - ISO4217 code (i.e: GBP) - char(3)
#                              <li>language - ISO639-1 code (i.e: en) - char(2)
#                          <ul> | ref.hash
#========================================================================================
sub setMerchantInfo {
    my $self = shift;
    my %hOptions = @_;

    my $UUID = $hOptions{'uuid'};
    my $Email = $hOptions{'email'};
    my $CountryCode = $hOptions{'country'};
    my $LanguageCode = $hOptions{'language'};
    my $CurrencyCode = $hOptions{'currency'};
    
    # Check shop currency matches zettle currency. If currencies don't match we'll throw an error and integration won't be setup
    my $Shop = $self->getSite();
    my $CurrencyID = $Shop->get('CurrencyID');
    Error('CURRENCY_MISMATCH', {
        'ShopCurrency' => $CurrencyID,
        'ZettleCurrency' => $CurrencyCode
    }) if $CurrencyID ne $CurrencyCode;

    $self->set({
        'ZettleUUID'         => $UUID,
        'ZettleEmail'        => $Email,
        'ZettleCountryCode'  => $CountryCode,
        'ZettleCurrencyCode' => $CurrencyCode,
        'ZettleLanguageCode' => $LanguageCode,
    });
}

#========================================================================================
# function     getMerchantUUID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $OrganizationUUID = $self->getMerchantUUID();
#----------------------------------------------------------------------------------------
# description  Retrieve organization UUID.
#----------------------------------------------------------------------------------------
# return       $OrganizationUUID | Organization UUID associated to merchant in Zettle site  | string
#========================================================================================
sub getMerchantUUID {
    return shift->get('ZettleUUID');
}

#========================================================================================
# function     getMerchantEmail
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Email = $self->getMerchantEmail();
#----------------------------------------------------------------------------------------
# description  Get Zettle merchant email.
#----------------------------------------------------------------------------------------
# return       $Email | Email merchant uses to login to Zettle site  | string
#========================================================================================
sub getMerchantEmail {
    return shift->get('ZettleEmail');
}

#========================================================================================
# function     getMerchantCountry
# state        public
#----------------------------------------------------------------------------------------
# syntax       $CountryCode = $self->getMerchantCountry();
#----------------------------------------------------------------------------------------
# description  Retrieve merchant country.
#----------------------------------------------------------------------------------------
# return       $CountryCode | ISO3166-1 code for country setup by merchant in Zettle site (i.e: GB)  | char(2)
#========================================================================================
sub getMerchantCountry {
    return shift->get('ZettleCountryCode');
}

#========================================================================================
# function     getMerchantCurrency
# state        public
#----------------------------------------------------------------------------------------
# syntax       $CurrencyCode = $self->getMerchantCurrency();
#----------------------------------------------------------------------------------------
# description  Retrieve merchant currency.
#----------------------------------------------------------------------------------------
# return       $CurrencyCode | ISO4217 code for currency setup by merchant in Zettle site (i.e: GBP)  | char(3)
#========================================================================================
sub getMerchantCurrency {
    return shift->get('ZettleCurrencyCode');
}

#========================================================================================
# function     getMerchantLanguage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LanguageCode = $self->getMerchantLanguage();
#----------------------------------------------------------------------------------------
# description  Retrieve merchant language.
#----------------------------------------------------------------------------------------
# return       $LanguageCode | ISO639-1 code for language setup by merchant in Zettle site (i.e: en)  | char(2)
#========================================================================================
sub getMerchantLanguage {
    return shift->get('ZettleLanguageCode');
}

#========================================================================================
# function     markForDeletion
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->markForDeletion();
#----------------------------------------------------------------------------------------
# description  Mark current environment for deletion.
#----------------------------------------------------------------------------------------
# input       $Delete | TRUE to mark environment to be deleted by batch job, false otherwise  | boolean
#========================================================================================
sub markForDeletion {
    my $self = shift;
    my ($Delete) = @_;

    if($Delete){
        $self->set({'MarkedForDeletion', GetCurrentDateTime()});
    }else{
        $self->set({'MarkedForDeletion', undef});        
    }
}

#========================================================================================
# function     setupInventorySubscription
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->setupInventorySubscription($InventorySubscriptionUUID, $InventorySubscriptionSigningKey);
#----------------------------------------------------------------------------------------
# description  Configure inventory subscription with data received from Zettle after creating event subscription.
#----------------------------------------------------------------------------------------
# input        $InventorySubscriptionUUID | Inventory subscription UUIDv1 | string
# input        $InventorySubscriptionSigningKey | Inventory subscription signing key | string
#========================================================================================
sub setupInventorySubscription {
    my $self = shift;
    my ($InventorySubscriptionUUID, $InventorySubscriptionSigningKey) = @_;

    $self->set({
        'InventorySubscriptionUUID' => $InventorySubscriptionUUID,
        'InventorySubscriptionSigningKey' => $InventorySubscriptionSigningKey,
    });
}

#========================================================================================
# function     getInventorySubscriptionUUID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $InventorySubscriptionUUID = $self->getInventorySubscriptionUUID();
#----------------------------------------------------------------------------------------
# description  Retrieve inventory subscription UUID.
#----------------------------------------------------------------------------------------
# return       $InventorySubscriptionUUID | Subscription UUID | string
#========================================================================================
sub getInventorySubscriptionUUID {
    return shift->get('InventorySubscriptionUUID');
}

#========================================================================================
# function     getInventorySubscriptionSigningKey
# state        public
#----------------------------------------------------------------------------------------
# syntax       $InventorySubscriptionSigningKey = $self->getInventorySubscriptionSigningKey();
#----------------------------------------------------------------------------------------
# description  Retrieve inventory subscription signing key.
#----------------------------------------------------------------------------------------
# return       $InventorySubscriptionSigningKey | Subscription signing key | string
#========================================================================================
sub getInventorySubscriptionSigningKey {
    return shift->get('InventorySubscriptionSigningKey');
}

#========================================================================================
# function     locationBinUuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationBinUUID = $self->locationBinUuid();
# syntax       $self->locationBinUuid($LocationBinUUID);
#----------------------------------------------------------------------------------------
# description  Get or set BIN location UUID.
#----------------------------------------------------------------------------------------
# input        $LocationBinUUID | optional. Location UUID | string
#----------------------------------------------------------------------------------------
# return       $LocationBinUUID | UUID for BIN location in Zettle site, if $LocationBinUUID argument is not present  | string
#========================================================================================
sub locationBinUuid {
    my $self = shift;
    my ($LocationBinUUID) = @_;
    
    return $self->get('LocationBinUUID') unless defined $LocationBinUUID;

    $self->set({'LocationBinUUID' => $LocationBinUUID});
}

#========================================================================================
# function     locationSoldUuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationSoldUUID = $self->locationSoldUuid();
# syntax       $self->locationSoldUuid($LocationSoldUUID);
#----------------------------------------------------------------------------------------
# description  Get or set SOLD location UUID.
#----------------------------------------------------------------------------------------
# input        $LocationSoldUUID | optional. Location UUID | string
#----------------------------------------------------------------------------------------
# return       $LocationSoldUUID | UUID for SOLD location in Zettle site, if $LocationSoldUUID argument is not present  | string
#========================================================================================
sub locationSoldUuid {
    my $self = shift;
    my ($LocationSoldUUID) = @_;
    
    return $self->get('LocationSoldUUID') unless defined $LocationSoldUUID;

    $self->set({'LocationSoldUUID' => $LocationSoldUUID});
}

#========================================================================================
# function     locationStoreUuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationStoreUUID = $self->locationStoreUuid();
# syntax       $self->locationStoreUuid($LocationStoreUUID);
#----------------------------------------------------------------------------------------
# description  Get or set STORE location UUID.
#----------------------------------------------------------------------------------------
# input        $LocationStoreUUID | optional. Location UUID | string
#----------------------------------------------------------------------------------------
# return       $LocationStoreUUID | UUID for STORE location in Zettle site, if $LocationStoreUUID argument is not present  | string
#========================================================================================
sub locationStoreUuid {
    my $self = shift;
    my ($LocationStoreUUID) = @_;
    
    return $self->get('LocationStoreUUID') unless defined $LocationStoreUUID;

    $self->set({'LocationStoreUUID' => $LocationStoreUUID});
}

#========================================================================================
# function     locationSupplierUuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationSupplierUUID = $self->locationSupplierUuid();
# syntax       $self->locationSupplierUuid($LocationSupplierUUID);
#----------------------------------------------------------------------------------------
# description  Get or set SUPPLIER location UUID.
#----------------------------------------------------------------------------------------
# input        $LocationSupplierUUID | optional. Location UUID | string
#----------------------------------------------------------------------------------------
# return       $LocationSupplierUUID | UUID for SUPPLIER location in Zettle site, if $LocationSupplierUUID argument is not present  | string
#========================================================================================
sub locationSupplierUuid {
    my $self = shift;
    my ($LocationSupplierUUID) = @_;
    
    return $self->get('LocationSupplierUUID') unless defined $LocationSupplierUUID;

    $self->set({'LocationSupplierUUID' => $LocationSupplierUUID});
}

1;
