#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::DTO::DTOFactory
# state        public
#----------------------------------------------------------------------------------------
# description  Factory for creating different DTO objects.
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::DTO::DTOFactory;

use strict;

use DE_EPAGES::Core::API::Error qw( Error );
use DE_EPAGES::Core::API::TestParameter qw( TestObject );
use DE_EPAGES::Object::API::Table::Country qw ( GetPKeyCountryByCode2 );
use DE_EPAGES::Zettle::API::Constants qw ( ZETTLE_ENVIRONMENT );
use DE_EPAGES::Zettle::API::Tools qw ( GenerateUUID );
use DE_EPAGES::Zettle::API::WebService::DTO::ProductDTO;
use DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO;
use DE_EPAGES::Zettle::API::WebService::VariationOptionHelper qw ( BuildOptionDefinitions BuildVariantOptions );

#ERRORDEFINITION
sub ErrorDefinition {
    return {
        'NoProductUuidError' => 'ZettleUUID missing for product ID=#ProductId',
        'NoProductPriceError' => 'No list price available for product ID=#ProductId',
    }
}
#ERRORDEFINITION END

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::WebService::DTO::DTOFactory->new($Shop);
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::WebService::DTO::DTOFactory
#========================================================================================
sub new {
    my $class = shift;
    my ($Shop) = @_;
    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');

    my $self = bless {
        'Shop' => $Shop,
        'Environment' => $Shop->get(ZETTLE_ENVIRONMENT),
    }, $class;
    
    return $self;
}

#========================================================================================
# function     buildProductDTO
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->buildProductDTO($Product, $IncludePrice, $hImageUploadResults);
#----------------------------------------------------------------------------------------
# description  build product DTO required for Zettle updates based on the given product
#----------------------------------------------------------------------------------------
# input        $Product | a product object | DE_EPAGES::Product::API::Object::Product
# input        $IncludePrice | determines whether price information will be included (optional) | boolean
# input        $hImageUploadResults | a map of ImageUploadResult objects indexed by product GUID
#                                      Image lookupKeys for products to be exported will be searched here.
#                                      See <DE_EPAGES::Zettle::API::Object::ImageUploadResult> | ref.hash.object
#========================================================================================
sub buildProductDTO {
    my $self = shift;
    my ($Product, $IncludePrice, $hImageUploadResults) = @_;
    TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');

    my $Shop = $self->_getSite();
    my $LanguageID = $Shop->get('LanguageID');
    my $Environment = $self->_getEnvironment();
    my $ZettleCurrencyCode = $Environment->getMerchantCurrency();
    my $ZettleCountryCode = $Environment->getMerchantCountry();
    
    my $ProductDTO = DE_EPAGES::Zettle::API::WebService::DTO::ProductDTO->new();

    my $Uuid = $self->_getProductUUID($Product);
    $ProductDTO->uuid($Uuid);

    my $ProductName = $Product->get('Name', $LanguageID);
    $ProductDTO->name($ProductName);

    my $ProductDescription = $Product->get('Description', $LanguageID);
    $ProductDTO->description($ProductDescription);

    my $ProductGuid = $Product->guid();

    # NOTE: ImageUploadResult might be undefined if product does not have image or image failed to be uploaded to Zettle
    my $aImageUploadResult = $hImageUploadResults->{$ProductGuid};
    my @ProductImageKeys = ();
    foreach my $ImageUploadResult ( @{$aImageUploadResult} ){
        my $ProductImageLookupKey = $ImageUploadResult->lookupKey();
        push(@ProductImageKeys, $ProductImageLookupKey);
    }
    # NOTE: if $ProductImageLookupKey is undefined @ProductImageKeys will be an empty array, resulting in image being removed from product in zettle (if previous image present)
    $ProductDTO->imageLookupKeys(\@ProductImageKeys);

    my $ProductVariations = $self->_buildProductVariations($Product, $IncludePrice, $LanguageID, $ZettleCurrencyCode);
    $ProductDTO->variations($ProductVariations);

    $ProductDTO->externalReference($Product->guid);

    my $OrderUnit = $Product->get('OrderUnit');
    my $Unit;
    $Unit = $OrderUnit->alias if defined $OrderUnit && $OrderUnit->alias ne 'piece';
    $ProductDTO->unitName($Unit) if defined $Unit;

    #TODO: confirm that we should use ZettleCountryCode (what if shop country does not match zettle country?)
    my $VatPercentage = $self->_getVatPercentage($Product, $ZettleCountryCode);
    $ProductDTO->vatPercentage($VatPercentage);

    my $VariantOptionDefinitions;
    my $aSubProducts = $Product->get('SubProducts');
    if(scalar @{$aSubProducts} > 0){
        $VariantOptionDefinitions = BuildOptionDefinitions($Product);
    }else{
        $VariantOptionDefinitions = [
            {
                "name" => "FakeOption",
                "properties" => [
                    {
                        "value" => "FakeValue",
                        "imageUrl" => undef
                    }
                ]
            }
        ];
    }
    $ProductDTO->variantOptionDefinitions($VariantOptionDefinitions);

    # NOTE: for unused fields see DE_EPAGES/Zettle/Documents/API/ZettleAPI-Extracts.md, section 1.2.3.3. Unused ProductDTO fields

    return $ProductDTO;
}

#========================================================================================
# function     buildProductPriceDTO
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->buildProductPriceDTO($Product);
#----------------------------------------------------------------------------------------
# description  Build product DTO required for Zettle updates based on the given product/subproduct.
#               Generated DTO will only include the necessary information to export product price.
#               For a master product it will create a fake Zettle variation since prices are handled at variation level.
#----------------------------------------------------------------------------------------
# input        $Product | a (sub)product | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub buildProductPriceDTO {
    my $self = shift;
    my ($Product) = @_;
    TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');

    my $Shop = $self->_getSite();
    my $LanguageID = $Shop->get('LanguageID');
    my $Environment = $self->_getEnvironment();
    my $ZettleCurrencyCode = $Environment->getMerchantCurrency();
    
    my $ProductDTO = DE_EPAGES::Zettle::API::WebService::DTO::ProductDTO->new();
    
    my $Uuid = $self->_getProductUUID($Product);
    $ProductDTO->uuid($Uuid);

    my $ProductName = $Product->get('Name', $LanguageID);
    $ProductDTO->name($ProductName);

    my $aVariationDTO = $self->_buildProductPriceVariations($Product, $ZettleCurrencyCode);

    $ProductDTO->variations($aVariationDTO);

    return $ProductDTO;
}

sub _buildProductVariations {
    my $self = shift;
    my ($Product, $IncludePrice, $LanguageID, $CurrencyID) = @_;

    my @aZettleVariations = ();

    my $aSubProducts = $Product->get('SubProducts');
    if(scalar @{$aSubProducts} > 0){
        foreach my $SubProduct (@$aSubProducts) {
            my $hLocAttributes = $SubProduct->get(['Name', 'Description'], $LanguageID);
            my $hAttributes = $SubProduct->get(['ManufacturerSKU', 'UPCEAN']);

            my $VariationDTO = DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO->new();

            my $Uuid = $self->_getProductUUID($SubProduct);
            $VariationDTO->uuid($Uuid);

            # We fill the name field for the sake of completeness but when Zettle processes variation it assigns a new name based on variation options values (i.e: blue, M, Cotton)
            my $Name = $hLocAttributes->{'Name'};
            $VariationDTO->name($Name) if $Name;

            my $Description = $hLocAttributes->{'Description'};
            $VariationDTO->description($Description) if $Description;

            my $Sku = $hAttributes->{'ManufacturerSKU'};
            $VariationDTO->sku($Sku) if $Sku;

            my $Barcode = $hAttributes->{'UPCEAN'};
            $VariationDTO->barcode($Barcode) if $Barcode;

            # NOTE: although imageLookupKeys can be added to variations, they are not displayed in Zettle, and they advised not to upload variation images

            if($IncludePrice){
                my $Price = $self->_getProductPrice($SubProduct, $CurrencyID);
                # NOTE: Price in cents. See DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.5.1.4. Product prices
                $VariationDTO->price($Price*100) if defined $Price;
                #TODO: how to calculate cost price? Is it available in ePages?
                my $CostPrice;
                $VariationDTO->costPrice($CostPrice*100) if defined $CostPrice;
                $VariationDTO->currencyCode($CurrencyID);
            }

            my $Options = BuildVariantOptions($SubProduct);
            $VariationDTO->options($Options);

            push(@aZettleVariations, $VariationDTO);
        }
    }else{
        # Create a fake variation for product without variations (Zettle stores prices and store at variation level)
        my $hLocAttributes = $Product->get(['Name', 'Description'], $LanguageID);
        my $hAttributes = $Product->get(['ManufacturerSKU', 'UPCEAN']);

        my $VariationDTO = DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO->new();

        my $Uuid = $self->_getProductUUID($Product);
        $VariationDTO->uuid($Uuid);

        # We fill the name field for the sake of completeness but when Zettle processes variation it assigns a new name based on variation options values (i.e: blue, M, Cotton)
        my $Name = $hLocAttributes->{'Name'};
        $VariationDTO->name($Name) if $Name;

        my $Description = $hLocAttributes->{'Description'};
        $VariationDTO->description($Description) if $Description;

        my $Sku = $hAttributes->{'ManufacturerSKU'};
        $VariationDTO->sku($Sku) if $Sku;

        my $Barcode = $hAttributes->{'UPCEAN'};
        $VariationDTO->barcode($Barcode) if $Barcode;

        # NOTE: although imageLookupKeys can be added to variations, they are not displayed in Zettle, and they advised not to upload variation images

        if($IncludePrice){
            my $Price = $self->_getProductPrice($Product, $CurrencyID);
            # NOTE: Price in cents. See DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.5.1.4. Product prices
            $VariationDTO->price($Price*100) if defined $Price;
            #TODO: how to calculate cost price? Is it available in ePages?
            my $CostPrice;
            $VariationDTO->costPrice($CostPrice*100) if defined $CostPrice;
            $VariationDTO->currencyCode($CurrencyID);
        }

        # We don't have variations for master products, neither variation attributes so we fake them too
        my $Options = [
            { 'name' => 'FakeOption', 'value' => 'FakeValue' }
        ];
        $VariationDTO->options($Options);

        push(@aZettleVariations, $VariationDTO);
    }

    return \@aZettleVariations;
}

# Create a list VariationDTO that will include only UUID and price information for $Product
# throws NoProductUuidError
# throws NoProductPriceError
sub _buildProductPriceVariations {
    my $self = shift;
    my ($Product, $CurrencyID) = @_;

    my @aZettleVariations = ();

    my ($VariationDTO, $Uuid, $Price);
    
    my $aSubProducts = $Product->get('SubProducts');
    if(scalar @{$aSubProducts} > 0){
        foreach my $SubProduct (@$aSubProducts) {
            $VariationDTO = DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO->new();

            $Uuid = $self->_getProductUUID($SubProduct);
            $VariationDTO->uuid($Uuid);
            
            $Price = $self->_getProductPrice($SubProduct, $CurrencyID);
            next unless defined $Price;
            # NOTE: Price in cents. See DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.5.1.4. Product prices
            $VariationDTO->price($Price*100);
            $VariationDTO->currencyCode($CurrencyID);
            
            push(@aZettleVariations, $VariationDTO);
        }
    }else{
        $VariationDTO = DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO->new();

        $Uuid = $self->_getProductUUID($Product);
        Error('NoProductUuidError', {'ProductId' => $Product->id()}) unless defined $Uuid;
        $VariationDTO->uuid($Uuid);
        
        $Price = $self->_getProductPrice($Product, $CurrencyID);
        Error('NoProductPriceError', {'ProductId' => $Product->id()}) unless defined $Price;

        if(defined $Price){
            # NOTE: Price in cents. See DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.5.1.4. Product prices
            $VariationDTO->price($Price*100);
            $VariationDTO->currencyCode($CurrencyID);
            
            push(@aZettleVariations, $VariationDTO);
        }
    }

    return \@aZettleVariations;
}

sub _getProductPrice {
    my $self = shift;
    my ($Product, $CurrencyID) = @_;

    my $DisplayPrice = $Product->displayPrice({
        CurrencyID => $CurrencyID,
        TaxModel => $Product->getSite()->get('TaxModel'),
    });

    return $DisplayPrice;
}

#========================================================================================
# function     _getVatPercentage
#----------------------------------------------------------------------------------------
# syntax       $VatPercentage = $self->_getVatPercentage($Product, $Country);
#----------------------------------------------------------------------------------------
# description  Returns VAT percentage for $Product applied in $Country.
#----------------------------------------------------------------------------------------
# input        $Product  | a product | DE_EPAGES::Product::API::Product
# input        $Country  | country iso code alpha 2 | char(2)
#----------------------------------------------------------------------------------------
# return       $VatPercentage | VAT applied to $Product in $Country (min: 0, max: 100) | int
#========================================================================================
sub _getVatPercentage {
    my $self = shift;
    my ($Product, $Country) = @_;

    my $CountryID = GetPKeyCountryByCode2($Country); # GB -> 826;
    my $Shop = $Product->getSite();
    my $TaxMatrix = $Shop->get('TaxMatrix');
    my $TaxClass = $Product->get('TaxClass')
                    // $TaxMatrix->defaultTaxClass;
    my $TaxArea = $TaxMatrix->taxAreaByCountry($CountryID)
                    // $TaxMatrix->defaultTaxArea;
    my $TaxRate = $TaxMatrix->taxRate($TaxArea,  $TaxClass);
    # Minimum 0, maximum 100
    my $VatPercentage = $TaxRate*100;
    # GetLog()->debug("Zettle::API::WebService::DTO::DTOFactory::_getVatPercentage: VatPercentage=$VatPercentage [TaxMatrix=".$TaxMatrix->alias().", TaxClass=".$TaxClass->alias().", TaxArea=".$TaxArea->alias()."]");

    return $VatPercentage;
}

#========================================================================================
# function     _getProductUUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Uuid = $self->_getProductUUID($Product);
#----------------------------------------------------------------------------------------
# description  Retrieve Zettle UUID for $Product. If UUID is empty it will be generated and updated.
#----------------------------------------------------------------------------------------
# input        $Product | a (sub) product | DE_EPAGES::Product::API::Object::Product
#----------------------------------------------------------------------------------------
# return       $Uuid | Product UUID used for Zettle | string
#========================================================================================
sub _getProductUUID {
    my $self = shift;
    my ($Product) = @_;

    my $Uuid = $Product->get('ZettleUUID');

    # If product has not been assigned a UUID, generate it and update product attribute
    if(not defined $Uuid){
        $Uuid = GenerateUUID();
        $Product->set({'ZettleUUID' => $Uuid});
    }

    return $Uuid;
}

sub _getSite {
    return shift->{'Shop'};
}

sub _getEnvironment {
    return shift->{'Environment'};
}

1;