#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::DTO::EventInitiatorDTO
# state        public
#----------------------------------------------------------------------------------------
# description  class to handle information about the user who made an update in Zettle that initiated an event.
#               See <DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateDTO>.
#----------------------------------------------------------------------------------------
# notes        The information is structured in the following way:
#               <ul>
#                   <li> userType - type of user - string
#                   <li> timestamp - #TODO: event timestamp in ISO-8601 format??? - string
#                   <li> clientUuid - #TODO: uuid of user who made the update??? - string
#                   <li> uuid - #TODO: product owner uuid??? - string
#               </ul>
#----------------------------------------------------------------------------------------
# example      {
#                   'userType' => 'USER',
#                   'timestamp' => '2021-08-24T17:10:44.856+0000',
#                   'clientUuid' => '51e57b0d-1ea2-48a4-b0a0-91d90424e62c',
#                   'uuid' => '1a5431c6-824d-11eb-92c1-8c85ccaff8db'
#               }
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::DTO::EventInitiatorDTO;

use strict;

use JSON qw ( to_json );

use DE_EPAGES::Zettle::API::Tools qw ( FormatDateTime ParseDateTime );

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::WebService::DTO::EventInitiatorDTO->new();
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::WebService::DTO::EventInitiatorDTO
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

# input        $UserType | Type of user (i.e. USER) | string
sub userType {
    my $self = shift;
    my ($UserType) = @_;
    return $self->{'userType'} unless defined($UserType);
    $self->{'userType'} = $UserType;
}

# input        $Timestamp | event timestamp in ISO-8601 format | Datetime
sub timestamp {
    my $self = shift;
    my ($Timestamp) = @_;
    return $self->{'timestamp'} unless defined($Timestamp);
    $self->{'timestamp'} = $Timestamp;
}

# input        $ClientUuid | uuid of user who made the update | string
sub clientUuid {
    my $self = shift;
    my ($ClientUuid) = @_;
    return $self->{'clientUuid'} unless defined($ClientUuid);
    $self->{'clientUuid'} = $ClientUuid;
}

# input        $OwnerUuid | product owner uuid (UUIDv1) | string
sub ownerUuid {
    my $self = shift;
    my ($OwnerUuid) = @_;
    return $self->{'uuid'} unless defined($OwnerUuid);
    $self->{'uuid'} = $OwnerUuid;
}

#========================================================================================
# function     fromHash
# state        public
#----------------------------------------------------------------------------------------
# syntax       $EventInitiatorDTO = DE_EPAGES::Zettle::API::WebService::DTO::EventInitiatorDTO->fromHash($hInitiator);
#----------------------------------------------------------------------------------------
# description  parse balance data to produce a EventInitiatorDTO object
#----------------------------------------------------------------------------------------
# input        $hInitiator | With the following keys:
#                            <ul>
#                                <li> created - product creation timestamp in ISO-8601 format - string
#                                <li> organizationUuid - organization uuid -  string
#                                <li> locationUuid - location uuid where product is stored - string
#                                <li> productUuid - super product uuid - string
#                                <li> variantUuid - variation uuid - string
#                                <li> balance - stock level (string representation of integer value) - string
#                            </ul>
#----------------------------------------------------------------------------------------
# return       $EventInitiatorDTO | a EventInitiatorDTO object | DE_EPAGES::Zettle::API::WebService::DTO::EventInitiatorDTO
#========================================================================================
sub fromHash {
    my $class = shift;
    my ($hInitiator) = @_;

    my $EventInitiatorDTO = $class->new();
    $EventInitiatorDTO->userType($hInitiator->{'userType'});
    my $DateTime = ParseDateTime($hInitiator->{'timestamp'});
    $EventInitiatorDTO->timestamp($DateTime);
    $EventInitiatorDTO->clientUuid($hInitiator->{'clientUuid'});
    $EventInitiatorDTO->ownerUuid($hInitiator->{'uuid'});

    return $EventInitiatorDTO;
}

# This method is invoked internally when <toJson> method is called.
sub TO_JSON {
    my $self = shift;
    my $hDatetimeAttributes = {};
    my $Timestamp = $self->timestamp();
    $hDatetimeAttributes->{'timestamp'} = FormatDateTime($Timestamp) if defined $Timestamp;
    my $Json = { %{ $self }, %{ $hDatetimeAttributes } };
    return $Json;
}

#========================================================================================
# function     toJson
# state        public
#----------------------------------------------------------------------------------------
# syntax       $EventInitiatorJSON = $self->toJson();
#----------------------------------------------------------------------------------------
# description  generate a JSON representation of current object
#----------------------------------------------------------------------------------------
# return       $EventInitiatorJSON | a JSON string | string
#========================================================================================
sub toJson {
    my $self = shift;

    return to_json($self, { 'pretty' => 1, 'allow_blessed' => 1, 'convert_blessed' => 1 });
}

1;
