#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::DTO::LocationInventoryDTO
# state        public
#----------------------------------------------------------------------------------------
# description  class that wraps location inventory response
#               See Responses > Model > locationInventory in https://inventory.izettle.com/swagger#/inventory/getInventoryByTrackedProducts
#               See Responses > Model > locationInventory in https://inventory.izettle.com/swagger#/inventory/getInventoryForLocation
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::DTO::LocationInventoryDTO;

use strict;

use DE_EPAGES::Zettle::API::WebService::DTO::VariationBalanceDTO;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::WebService::DTO::LocationInventoryDTO->new();
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::WebService::DTO::LocationInventoryDTO
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

#========================================================================================
# function     locationUuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationUuid = $self->locationUuid();
# syntax       $self->locationUuid($UUID);
#----------------------------------------------------------------------------------------
# description  Get/Set location UUID.
#----------------------------------------------------------------------------------------
# input        $Uuid | Location UUIDv1 | string
#----------------------------------------------------------------------------------------
# return       $LocationUuid | Current location UUID if $Uuid parameter is not provided | string
#========================================================================================
sub locationUuid {
    my $self = shift;
    my ($Uuid) = @_;
    return $self->{'locationUuid'} unless defined($Uuid);
    $self->{'locationUuid'} = $Uuid;
}

#========================================================================================
# function     products
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aProducts = $self->products();
# syntax       $self->products($aProducts);
#----------------------------------------------------------------------------------------
# description  Get/Set product UUIDs.
#----------------------------------------------------------------------------------------
# input        $aProducts | List of product UUIDs | ref.array.string
#----------------------------------------------------------------------------------------
# return       $aProducts | Current product UUIDs if $aProducts parameter is not provided | ref.array.string
#========================================================================================
sub products {
    my $self = shift;
    my ($aProducts) = @_;
    return $self->{'trackedProducts'} unless defined($aProducts);
    $self->{'trackedProducts'} = $aProducts;
}

#========================================================================================
# function     variationBalances
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aVariationBalanceDTO = $self->variationBalances();
# syntax       $self->variationBalances($aVariationBalanceDTO);
#----------------------------------------------------------------------------------------
# description  Get/Set variation balances.
#               See DE_EPAGES::Zettle::API::WebService::DTO::VariationBalanceDTO.
#----------------------------------------------------------------------------------------
# input        $aVariationBalanceDTO | List of VariationBalanceDTO objects | ref.array.object
#----------------------------------------------------------------------------------------
# return       $aVariationBalanceDTO | Current variation balances if $aVariationBalanceDTO parameter is not provided | ref.array.object
#========================================================================================
sub variationBalances {
    my $self = shift;
    my ($aVariationBalanceDTO) = @_;
    return $self->{'variants'} unless defined($aVariationBalanceDTO);
    $self->{'variants'} = $aVariationBalanceDTO;
}

#========================================================================================
# function     latest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LatestUpdate = $self->latest();
# syntax       $self->latest($LatestUpdate);
#----------------------------------------------------------------------------------------
# description  The latest update among all variations.
#               It's a datetime value expressed as a string (i.e. 2021-08-15T11:06:52.769+0000).
#               See Responses > Model > locationInventory > latest in https://inventory.izettle.com/swagger#/inventory/getInventoryForLocation
#----------------------------------------------------------------------------------------
# input        $LatestUpdate | Latest update | string
#----------------------------------------------------------------------------------------
# return       $LatestUpdate | Latest update if $LatestUpdate parameter is not provided | string
#========================================================================================
#TODO: find out the real meaning of this field
sub latest {
    my $self = shift;
    my ($LatestUpdate) = @_;
    return $self->{'latest'} unless defined($LatestUpdate);
    $self->{'latest'} = $LatestUpdate;
}

#========================================================================================
# function     fromHash
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationInventoryDTO = DE_EPAGES::Zettle::API::WebService::DTO::LocationInventoryDTO->fromHash($hLocationInventory);
#----------------------------------------------------------------------------------------
# description  Parse location inventory info and return a LocationInventoryDTO object.
#               See Responses > Model > locationInventory in https://inventory.izettle.com/swagger#/inventory/getInventoryForLocation
#----------------------------------------------------------------------------------------
# input        $hLocationInventory | Location balance as part of ProductInventory API response with following keys:
#               <ul>
#                   <li>locationUuid - location UUID - string
#                   <li>trackedProducts - list of product UUIDs - ref.array.string
#                   <li>variants - list of variation balances with the following keys:
#                       <ul>
#                           <li>locationUuid - location UUID - string
#                           <li>locationType - type of location [ SUPPLIER, STORE, SOLD, BIN ] - string
#                           <li>productUuid - super product UUID - string
#                           <li>variantUuid - variation UUID - string
#                           <li>balance - stock level - string
#                       </ul> | ref.array.hash
#                   <li>latest - lastest update among all variations in this response - string
#               </ul> | ref.hash
# example      {
#                   "locationUuid": "8f59b9ee-828b-11eb-b512-f9a20a1c5cc2",
#                   "trackedProducts": ["61180610-1f42-1948-2520-b976395f1ee2"],
#                   "variants": [{
#                       "locationUuid": "8f59b9ee-828b-11eb-b512-f9a20a1c5cc2",
#                       "locationType": "STORE",
#                       "productUuid": "61180610-1f42-1948-2520-b976395f1ee2",
#                       "variantUuid": "61180610-6c26-1362-2332-b976395f1e82",
#                       "balance": "13"
#                   }, {
#                       "locationUuid": "8f59b9ee-828b-11eb-b512-f9a20a1c5cc2",
#                       "locationType": "STORE",
#                       "productUuid": "61180610-1f42-1948-2520-b976395f1ee2",
#                       "variantUuid": "61180610-3ddb-19de-2bd6-b976395f1e70",
#                       "balance": "22"
#                   }],
#                   "latest": "2021-08-15T11:06:52.769+0000"
#               }
#----------------------------------------------------------------------------------------
# return       $LocationInventoryDTO | a LocationInventoryDTO object | DE_EPAGES::Zettle::API::WebService::DTO::LocationInventoryDTO
#========================================================================================
sub fromHash {
    my $class = shift;
    my ($hLocationInventory) = @_;

    my $LocationInventoryDTO = $class->new();
    $LocationInventoryDTO->locationUuid($hLocationInventory->{'locationUuid'});
    $LocationInventoryDTO->products($hLocationInventory->{'trackedProducts'});
    my @aVariationBalanceDTO;
    foreach my $hVariationBalance (@{$hLocationInventory->{'variants'}}){
        my $VariationBalanceDTO = DE_EPAGES::Zettle::API::WebService::DTO::VariationBalanceDTO->fromHash($hVariationBalance);
        push(@aVariationBalanceDTO, $VariationBalanceDTO);
    }
    $LocationInventoryDTO->variationBalances(\@aVariationBalanceDTO);
    $LocationInventoryDTO->latest($hLocationInventory->{'latest'});

    return $LocationInventoryDTO;
}

1;
