#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::DTO::ProductInventoryDTO
# state        public
#----------------------------------------------------------------------------------------
# description  class that wraps product balance info
#               See Responses > Model > productBalance in https://inventory.izettle.com/swagger#/inventory/getInventoryForProduct
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::DTO::ProductInventoryDTO;

use strict;

use DE_EPAGES::Zettle::API::WebService::DTO::VariationBalanceDTO;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::WebService::DTO::ProductInventoryDTO->new();
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::WebService::DTO::ProductInventoryDTO
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

#========================================================================================
# function     locationUuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationUuid = $self->locationUuid();
# syntax       $self->locationUuid($UUID);
#----------------------------------------------------------------------------------------
# description  Get/Set location UUID.
#----------------------------------------------------------------------------------------
# input        $Uuid | Location UUIDv1 | string
#----------------------------------------------------------------------------------------
# return       $LocationUuid | Current location UUID if $Uuid parameter is not provided | string
#========================================================================================
sub locationUuid {
    my $self = shift;
    my ($Uuid) = @_;
    return $self->{'locationUuid'} unless defined($Uuid);
    $self->{'locationUuid'} = $Uuid;
}

#========================================================================================
# function     variationBalances
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aVariationBalanceDTO = $self->variationBalances();
# syntax       $self->variationBalances($aVariationBalanceDTO);
#----------------------------------------------------------------------------------------
# description  Get/Set variation balances.
#               See DE_EPAGES::Zettle::API::WebService::DTO::VariationBalanceDTO.
#----------------------------------------------------------------------------------------
# input        $aVariationBalanceDTO | List of VariationBalanceDTO objects | ref.array.object
#----------------------------------------------------------------------------------------
# return       $aVariationBalanceDTO | Current variation balances if $aVariationBalanceDTO parameter is not provided | ref.array.object
#========================================================================================
sub variationBalances {
    my $self = shift;
    my ($aVariationBalanceDTO) = @_;
    return $self->{'variants'} unless defined($aVariationBalanceDTO);
    $self->{'variants'} = $aVariationBalanceDTO;
}

#========================================================================================
# function     fromHash
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ProductInventoryDTO = DE_EPAGES::Zettle::API::WebService::DTO::ProductInventoryDTO->fromHash($hProductBalance);
#----------------------------------------------------------------------------------------
# description  Parse product balance info and return a ProductInventoryDTO object.
#----------------------------------------------------------------------------------------
# input        $hProductBalance | Product balance with following keys:
#               <ul>
#                   <li>locationUuid - location UUID - string
#                   <li>variants - list of variation balances with the following keys:
#                       <ul>
#                           <li>locationUuid - location UUID - string
#                           <li>locationType - type of location [ SUPPLIER, STORE, SOLD, BIN ] - string
#                           <li>productUuid - super product UUID - string
#                           <li>variantUuid - variation UUID - string
#                           <li>balance - stock level - string
#                       </ul> | ref.array.hash
#               </ul> | ref.hash
#----------------------------------------------------------------------------------------
# return       $ProductInventoryDTO | a ProductInventoryDTO object | DE_EPAGES::Zettle::API::WebService::DTO::ProductInventoryDTO
#========================================================================================
sub fromHash {
    my $class = shift;
    my ($hProductBalance) = @_;

    my $ProductInventoryDTO = $class->new();
    $ProductInventoryDTO->locationUuid($hProductBalance->{'locationUuid'});
    my @aVariationBalanceDTO;
    foreach my $hVariationBalance (@{$hProductBalance->{'variants'}}){
        my $VariationBalanceDTO = DE_EPAGES::Zettle::API::WebService::DTO::VariationBalanceDTO->fromHash($hVariationBalance);
        push(@aVariationBalanceDTO, $VariationBalanceDTO);
    }
    $ProductInventoryDTO->variationBalances(\@aVariationBalanceDTO);

    return $ProductInventoryDTO;
}

1;
