#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::DTO::Requirements
# state        public
#----------------------------------------------------------------------------------------
# description  class to prepare the API outbound data for requests
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::DTO::Requirements;
use base qw( Exporter );

use strict;
our @EXPORT_OK = qw(
    AdjustProductName
    AdjustProductDescription
    IsProductExportable
);

use DE_EPAGES::Zettle::API::Constants qw ( API_MAX_NAME_LENGTH API_MAX_DESCRIPTION_LENGTH );

#========================================================================================
# function     AdjustProductName
# state        public
#----------------------------------------------------------------------------------------
# syntax       AdjustProductName($Name);
#----------------------------------------------------------------------------------------
# description  Ensure name length requirements (max. API_MAX_NAME_LENGTH) truncating string if necessary
#----------------------------------------------------------------------------------------
# input        $Name | name of a product | string
#----------------------------------------------------------------------------------------
# return        $AdjustedName | input string if length is fine or truncated string otherwise | string
#========================================================================================
sub AdjustProductName {
    my ($Name) = @_;

    return (length $Name > API_MAX_NAME_LENGTH)
            ? substr($Name, 0, API_MAX_NAME_LENGTH)
            : $Name;
}

#========================================================================================
# function     AdjustProductDescription
# state        public
#----------------------------------------------------------------------------------------
# syntax       AdjustProductDescription($Description);
#----------------------------------------------------------------------------------------
# description  Ensure description length requirements (max. API_MAX_DESCRIPTION_LENGTH) truncating string if necessary
#----------------------------------------------------------------------------------------
# input        $Description | description of a product | string
#----------------------------------------------------------------------------------------
# return        $AdjustedDescription | input string if length is fine or truncated string otherwise | string
#========================================================================================
sub AdjustProductDescription {
    my ($Description) = @_;

    return (length $Description > API_MAX_DESCRIPTION_LENGTH)
            ? substr($Description, 0, API_MAX_DESCRIPTION_LENGTH)
            : $Description;
}

#========================================================================================
# function     IsProductExportable
# state        public
#----------------------------------------------------------------------------------------
# syntax       IsProductExportable($Product);
#----------------------------------------------------------------------------------------
# description  Test if given $Product meets criteria in order to be exported to Zettle:
#               <ul>
#                 <li>Product has at least one subproduct.
#                 <li>Product is visible.
#               <ul>
#----------------------------------------------------------------------------------------
# input        $Product | a product object | DE_EPAGES::Product::API::Object::Product
#----------------------------------------------------------------------------------------
# return        $IsExportable | TRUE if product meets criteria, false otherwise | boolean
#========================================================================================
sub IsProductExportable {
    my ($Product) = @_;

    #TODO: remove "IsVisible" criteria from IsProductExportable (for current tests only visible products are exported to Zettle).
    return 0 unless $Product->get('IsVisible');

    return 1;
}

1;