#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateBalanceDTO
# state        public
#----------------------------------------------------------------------------------------
# description  class to handle balance information as part of StockUpdateDTO object.
#               See <DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateDTO>.
#----------------------------------------------------------------------------------------
# notes        The information is structured in the following way:
#               <ul>
#                   <li> created - product creation timestamp in ISO-8601 format - string
#                   <li> organizationUuid - organization uuid -  string
#                   <li> locationUuid - location uuid where product is stored - string
#                   <li> productUuid - super product uuid - string
#                   <li> variantUuid - variation uuid - string
#                   <li> balance - stock level (string representation of integer value) - string
#               </ul>
#----------------------------------------------------------------------------------------
# example      {
#                   "created": "2021-08-23T10:54:03.124+0000",
#                   "organizationUuid": "1a52e2a8-824d-11eb-acc8-aafa2648b59f",
#                   "locationUuid": "8f59b9ee-828b-11eb-b512-f9a20a1c5cc2",
#                   "productUuid": "611ea330-d0fd-114f-2671-b976395ffb2b",
#                   "variantUuid": "611ea330-684f-1b83-2d32-b976395ffbd0",
#                   "balance": "2"
#               }
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateBalanceDTO;

use strict;

use JSON qw ( to_json );

use DE_EPAGES::Zettle::API::Tools qw ( FormatDateTime ParseDateTime );

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateBalanceDTO->new();
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateBalanceDTO
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

# input        $Created | timestamp when variation was created | Datetime
sub created {
    my $self = shift;
    my ($Created) = @_;
    return $self->{'created'} unless defined($Created);
    $self->{'created'} = $Created;
}

# input        $OrganizationUuid | Organization identifier (UUIDv1) | string
sub organizationUuid {
    my $self = shift;
    my ($OrganizationUuid) = @_;
    return $self->{'organizationUuid'} unless defined($OrganizationUuid);
    $self->{'organizationUuid'} = $OrganizationUuid;
}

# input        $LocationUuid | Location identifier (UUIDv1) | string
sub locationUuid {
    my $self = shift;
    my ($LocationUuid) = @_;
    return $self->{'locationUuid'} unless defined($LocationUuid);
    $self->{'locationUuid'} = $LocationUuid;
}

# input        $ProductUuid | Super product identifier (UUIDv1) | string
sub productUuid {
    my $self = shift;
    my ($ProductUuid) = @_;
    return $self->{'productUuid'} unless defined($ProductUuid);
    $self->{'productUuid'} = $ProductUuid;
}

# input        $VariationUuid | Variation identifier (UUIDv1) | string
sub variationUuid {
    my $self = shift;
    my ($VariationUuid) = @_;
    return $self->{'variantUuid'} unless defined($VariationUuid);
    $self->{'variantUuid'} = $VariationUuid;
}

# input        $Balance | Variation stock level | integer
sub balance {
    my $self = shift;
    my ($Balance) = @_;
    return $self->{'balance'} unless defined($Balance);
    $self->{'balance'} = $Balance;
}

#========================================================================================
# function     fromHash
# state        public
#----------------------------------------------------------------------------------------
# syntax       $StockUpdateBalanceDTO = DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateBalanceDTO->fromHash($hBalance);
#----------------------------------------------------------------------------------------
# description  parse balance data to produce a StockUpdateBalanceDTO object
#----------------------------------------------------------------------------------------
# input        $hBalance | With the following keys:
#                           <ul>
#                               <li> created - product creation timestamp in ISO-8601 format - string
#                               <li> organizationUuid - organization uuid -  string
#                               <li> locationUuid - location uuid where product is stored - string
#                               <li> productUuid - super product uuid - string
#                               <li> variantUuid - variation uuid - string
#                               <li> balance - stock level (string representation of integer value) - string
#                           </ul>
#----------------------------------------------------------------------------------------
# return       $StockUpdateBalanceDTO | a StockUpdateBalanceDTO object | DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateBalanceDTO
#========================================================================================
sub fromHash {
    my $class = shift;
    my ($hBalance) = @_;

    my $StockUpdateBalanceDTO = $class->new();
    my $Created = $hBalance->{'created'};
    if(defined $Created){
        my $CreatedDt = ParseDateTime($Created);
        $StockUpdateBalanceDTO->created($CreatedDt);
    }
    $StockUpdateBalanceDTO->organizationUuid($hBalance->{'organizationUuid'});
    $StockUpdateBalanceDTO->locationUuid($hBalance->{'locationUuid'});
    $StockUpdateBalanceDTO->productUuid($hBalance->{'productUuid'});
    $StockUpdateBalanceDTO->variationUuid($hBalance->{'variantUuid'});
    $StockUpdateBalanceDTO->balance($hBalance->{'balance'});

    return $StockUpdateBalanceDTO;
}

sub TO_JSON {
    my $self = shift;
    my $hDatetimeAttributes = {};
    $hDatetimeAttributes->{'created'} = FormatDateTime($self->created) if defined $self->created;
    my $Json = { %{ $self }, %{ $hDatetimeAttributes } };
    return $Json;
}

sub toJson {
    my $self = shift;

    return to_json($self, { 'pretty' => 1, 'allow_blessed' => 1, 'convert_blessed' => 1 });
}

1;
