#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::EndpointHandler
# state        private
#----------------------------------------------------------------------------------------
# description  This module provides all the methods to retrieve Zettle API endpoint URLs
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::EndpointHandler;
use base qw ( Exporter );

use strict;

our @EXPORT_OK = qw (
    GetAuthApiUrl
    GetDisconnectApiUrl
    GetTokenApiUrl
    GetUserInfoApiUrl

    GetAccountInfoApiUrl

    GetBulkUploadImageApiUrl
    GetUploadImageApiUrl

    GetAllProductsApiUrl
    GetBulkExportApiUrl
    GetBulkExportStatusApiUrl
    GetDeleteProductApiUrl
    GetDeleteProductListApiUrl
    GetProductCountApiUrl
    
    GetBulkUpdateProductTrackingApiUrl
    GetDisableProductTrackingApiUrl
    GetEnableProductTrackingApiUrl
    GetLocationTypeInventoryApiUrl
    GetLocationInventoryApiUrl
    GetProductInventoryApiUrl
    GetUpdateProductBalanceApiUrl 
    GetBulkUpdateProductBalanceApiUrl
    GetAllLocationsApiUrl
);

use DE_EPAGES::Object::API::Factory qw( LoadRootObject );

# Method URLs relative to OAuth API endpoint
my $OAuthApiMethodUrl = {
    'auth'       => '/authorize',
    'disconnect' => '/application-connections/self',
    'token'      => '/token',
    'userInfo'   => '/users/me',
};

# Method URLs relative to Secure API endpoint
my $SecureApiMethodUrl = {
    'accountInfo' => '/api/resources/organizations/self',
};

# Method URLs relative to Image API endpoint
my $ImageApiMethodUrl = {
    'bulkUploadImages' => '/v2/images/organizations/self/products/bulk',
    'uploadImage'      => '/v2/images/organizations/self/products',
};

# Method URLs relative to Product API endpoint
my $ProductApiMethodUrl = {
    'allProducts'       => '/organizations/self/products/v2',
    'bulkExport'        => '/organizations/self/import/v2',
    'bulkExportStatus'  => '/organizations/self/import/status/#ExportUUID',
    'deleteProduct'     => '/organizations/self/products/#ProductUUID',
    'deleteProductList' => '/organizations/self/products',
    'productCount'      => '/organizations/self/products/v2/count',
};

# Method URLs relative to Inventory API endpoint
my $InventoryApiMethodUrl = {
    'bulkUpdateProductTracking' => '/organizations/self/v2/inventory/bulk',
    'disableProductTracking'    => '/organizations/self/inventory/products/#ProductUUID',
    'enableProductTracking'     => '/organizations/self/inventory',
    'locationTypeInventory'     => '/organizations/self/inventory/locations',
    'locationInventory'         => '/organizations/self/inventory/locations/#LocationUUID',
    'productInventory'          => '/organizations/self/inventory/locations/#LocationUUID/products/#ProductUUID',
    'updateProductBalance'      => '/organizations/self/inventory',
    'bulkUpdateProductBalance'  => '/organizations/self/v2/inventory/bulk',
    'allLocations'              => '/organizations/self/locations',
};

#================================ OAuth endpoint related methods =================================

sub GetAuthApiUrl {
    return _GetOAuthEndpoint() . $OAuthApiMethodUrl->{'auth'};
}

sub GetDisconnectApiUrl {
    return _GetOAuthEndpoint() . $OAuthApiMethodUrl->{'disconnect'};
}

sub GetTokenApiUrl {
    return _GetOAuthEndpoint() . $OAuthApiMethodUrl->{'token'};
}

sub GetUserInfoApiUrl {
    return _GetOAuthEndpoint() . $OAuthApiMethodUrl->{'userInfo'};
}

#================================ Secure endpoint related methods =================================

sub GetAccountInfoApiUrl {
    return _GetSecureEndpoint() . $SecureApiMethodUrl->{'accountInfo'};
}

#================================ Image endpoint related methods =================================

sub GetBulkUploadImageApiUrl {
    return _GetImageEndpoint() . $ImageApiMethodUrl->{'bulkUploadImages'};
}

sub GetUploadImageApiUrl {
    return _GetImageEndpoint() . $ImageApiMethodUrl->{'uploadImage'};
}

#================================ Product endpoint related methods =================================

sub GetAllProductsApiUrl {
    return _GetProductEndpoint() . $ProductApiMethodUrl->{'allProducts'};
}

sub GetBulkExportApiUrl {
    return _GetProductEndpoint() . $ProductApiMethodUrl->{'bulkExport'};
}

#========================================================================================
# function     GetBulkExportStatusApiUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetBulkExportStatusApiUrl($ExportUUID);
#----------------------------------------------------------------------------------------
# description  Retrieve URL used to query bulk export status
#               See: https://products.izettle.com/swagger#/import/getStatusByUuid
#               See: https://products.izettle.com/swagger#/import/getLatestImportStatus
#----------------------------------------------------------------------------------------
# input        $ExportUUID | optional. Export UUID (default: query last bulk export status) | string
#----------------------------------------------------------------------------------------
# return       $StatusURL | Export status URL | string
#========================================================================================
sub GetBulkExportStatusApiUrl {
    my ($ExportUUID) = @_;
    $ExportUUID //= '';

    my $ParameterizedMethodUrl = $ProductApiMethodUrl->{'bulkExportStatus'};
    $ParameterizedMethodUrl =~ s/#ExportUUID/$ExportUUID/;

    return _GetProductEndpoint() . $ParameterizedMethodUrl;
}

#========================================================================================
# function     GetDeleteProductApiUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetDeleteProductApiUrl($ProductUUID);
#----------------------------------------------------------------------------------------
# description  Retrieve URL used to delete a single product
#----------------------------------------------------------------------------------------
# input        $ProductUUID | Product UUID | string
#----------------------------------------------------------------------------------------
# return       $DeleteURL | URL required for DELETE request | string
#========================================================================================
sub GetDeleteProductApiUrl {
    my ($ProductUUID) = @_;

    my $ParameterizedMethodUrl = $ProductApiMethodUrl->{'deleteProduct'};
    $ParameterizedMethodUrl =~ s/#ProductUUID/$ProductUUID/;

    return _GetProductEndpoint() . $ParameterizedMethodUrl;
}

sub GetDeleteProductListApiUrl {
    my ($ProductUUID) = @_;
    return _GetProductEndpoint() . $ProductApiMethodUrl->{'deleteProductList'};
}

sub GetProductCountApiUrl {
    return _GetProductEndpoint() . $ProductApiMethodUrl->{'productCount'};
}

#================================ Inventory endpoint related methods =================================

sub GetBulkUpdateProductTrackingApiUrl {
    return _GetInventoryEndpoint() . $InventoryApiMethodUrl->{'bulkUpdateProductTracking'};
}

#========================================================================================
# function     GetDisableProductTrackingApiUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetDisableProductTrackingApiUrl($ProductUUID);
#----------------------------------------------------------------------------------------
# description  Retrieve URL used to disable inventory tracking for given product.
#----------------------------------------------------------------------------------------
# input        $ProductUUID | Product UUID | string
#----------------------------------------------------------------------------------------
# return       $DeleteURL | URL required for DELETE request | string
#========================================================================================
sub GetDisableProductTrackingApiUrl {
    my ($ProductUUID) = @_;

    my $ParameterizedMethodUrl = $InventoryApiMethodUrl->{'disableProductTracking'};
    $ParameterizedMethodUrl =~ s/#ProductUUID/$ProductUUID/;

    return _GetInventoryEndpoint() . $ParameterizedMethodUrl;
}

sub GetEnableProductTrackingApiUrl {
    return _GetInventoryEndpoint() . $InventoryApiMethodUrl->{'enableProductTracking'};
}

sub GetLocationTypeInventoryApiUrl {
    return _GetInventoryEndpoint() . $InventoryApiMethodUrl->{'locationTypeInventory'};
}

#========================================================================================
# function     GetLocationInventoryApiUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetLocationInventoryApiUrl($LocationUUID);
#----------------------------------------------------------------------------------------
# description  Generate URL used to retrieve inventory for given location.
#----------------------------------------------------------------------------------------
# input        $LocationUUID | Product UUID | string
#----------------------------------------------------------------------------------------
# return       $GetURL | URL required for GET request | string
#========================================================================================
sub GetLocationInventoryApiUrl {
    my ($LocationUUID) = @_;

    my $ParameterizedMethodUrl = $InventoryApiMethodUrl->{'locationInventory'};
    $ParameterizedMethodUrl =~ s/#LocationUUID/$LocationUUID/;

    return _GetInventoryEndpoint() . $ParameterizedMethodUrl;
}

#========================================================================================
# function     GetProductInventoryApiUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetProductInventoryApiUrl($LocationUUID, $ProductUUID);
#----------------------------------------------------------------------------------------
# description  Generate URL used to retrieve inventory for given product in specified location.
#----------------------------------------------------------------------------------------
# input        $LocationUUID | Product UUID | string
# input        $ProductUUID | Product UUID | string
#----------------------------------------------------------------------------------------
# return       $GetURL | URL required for GET request | string
#========================================================================================
sub GetProductInventoryApiUrl {
    my ($LocationUUID, $ProductUUID) = @_;

    my $ParameterizedMethodUrl = $InventoryApiMethodUrl->{'productInventory'};
    $ParameterizedMethodUrl =~ s/#LocationUUID/$LocationUUID/;
    $ParameterizedMethodUrl =~ s/#ProductUUID/$ProductUUID/;

    return _GetInventoryEndpoint() . $ParameterizedMethodUrl;
}

sub GetUpdateProductBalanceApiUrl {
    return _GetInventoryEndpoint() . $InventoryApiMethodUrl->{'updateProductBalance'};
}

sub GetBulkUpdateProductBalanceApiUrl {
    return _GetInventoryEndpoint() . $InventoryApiMethodUrl->{'bulkUpdateProductBalance'};
}

sub GetAllLocationsApiUrl { 
    return _GetInventoryEndpoint() . $InventoryApiMethodUrl->{'allLocations'};
}
 
#=================================== Private methods ====================================

# Retrieve OAuth endpoint
# See https://github.com/iZettle/api-documentation/blob/master/authorization.adoc
sub _GetOAuthEndpoint {
    return _LoadSystemAttribute('Zettle_OAuthEndpoint');
}

# Retrieve Image endpoint
# See https://github.com/iZettle/api-documentation/blob/master/image.adoc
sub _GetImageEndpoint {
    return _LoadSystemAttribute('Zettle_ImageEndpoint');
}

# Retrieve Product endpoint
# See https://github.com/iZettle/api-documentation/blob/master/product-library.adoc
sub _GetProductEndpoint {
    return _LoadSystemAttribute('Zettle_ProductEndpoint');
}

# Retrieve Inventory endpoint
# See https://github.com/iZettle/api-documentation/blob/master/inventory.adoc
sub _GetInventoryEndpoint {
    return _LoadSystemAttribute('Zettle_InventoryEndpoint');
}

# Retrieve Secure endpoint
# See DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.4.2. Retrieve a Merchant's Account Information
sub _GetSecureEndpoint {
    return 'https://secure.izettle.com';
}

# Return the value of the given system attribute
sub _LoadSystemAttribute {
    my ($AttributeName) = @_;

    my $System = LoadRootObject();
    my $AttributeValue = $System->get($AttributeName);

    return $AttributeValue;
}

1;
