#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::ImageService
# base         DE_EPAGES::Zettle::API::WebService::RestService
# state        private
#----------------------------------------------------------------------------------------
# description  Service wrapper around Zettle API image-related methods
#               See https://github.com/iZettle/api-documentation/blob/master/image.adoc
#               See https://image.izettle.com/swagger
#----------------------------------------------------------------------------------------
# notes        Image service allows users to upload product images to be stored and used within Zettle applications.
#               Image formats supported are: [BMP,GIF,JPEG,PNG,TIFF].
#               The images can uploaded via an URL which must be < 5MB.
#               Zettle image service is eventually consistent and can reject requests if overloaded.
#               Images can also be uploaded using a byte array with limitations of > 50*50px and < 5MB (not implemented in current ImageService version).
#               Image URLs to be imported by Zettle must use the https protocol and have to be properly encoded (no non-ASCII characters in the path).
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::ImageService;
use base DE_EPAGES::Zettle::API::WebService::RestService;

use strict;

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( LogError );
use DE_EPAGES::Core::API::File qw( fsplit );
use DE_EPAGES::Zettle::API::WebService::DTO::ImageUploadResponseDTO;
use DE_EPAGES::Zettle::API::WebService::DTO::BulkImageUploadResponseDTO;
use DE_EPAGES::Zettle::API::WebService::EndpointHandler qw ( GetUploadImageApiUrl GetBulkUploadImageApiUrl );

# Hash used to map from image extension to API image format
my $EXTENSION_TO_IMAGE_FORMAT = {
    '.bmp'  => 'BMP',
    '.gif'  => 'GIF',
    '.jpg'  => 'JPEG',
    '.jpeg' => 'JPEG',
    '.png'  => 'PNG',
    '.tiff' => 'TIFF'
};

#ERRORDEFINITION
sub ErrorDefinition () {
    return {
        'UNSUPPORTED_IMAGE_FORMAT' =>
          '#Extension is not a supported image format (#ImageUrl) - Supported formats: '.join(',', keys %$EXTENSION_TO_IMAGE_FORMAT),
    };
}
#ERRORDEFINITION END

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::WebService::ImageService->new($Environment);
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# input        $Environment | a Zettle environment object | DE_EPAGES::Zettle::API::Object:ZettleEnvironment
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::WebService::ImageService
#========================================================================================
sub new {
    my $class = shift;
    my ($Environment) = @_;

    my $self = $class->SUPER::new($Environment);
    return $self;
}

#========================================================================================
# function     uploadImage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ImageUploadResponseDTO = $self->uploadImage($ImageUrl);
#----------------------------------------------------------------------------------------
# description  Creates a job for uploading the image, the lookup key is returned directly.
#               The images are eventually available for lookup.
#               Image formats supported are: [BMP,GIF,JPEG,PNG,TIFF].
#               The images must be > 50*50px and < 5MB.
#               See: https://image.izettle.com/swagger#/images/uploadProductImage
#----------------------------------------------------------------------------------------
# input        $ImageUrl | Public webpath of a product large image | string
#----------------------------------------------------------------------------------------
# return       $ImageUploadResponseDTO | a ImageUploadResponseDTO object | DE_EPAGES::Zettle::API::WebService::DTO::ImageUploadResponseDTO
#========================================================================================
sub uploadImage {
    my $self = shift;

    my ($ImageUrl) = @_;

    my $ImageUploadUrl = GetUploadImageApiUrl();
    my $hHeadersParams = {
        'Content-type' => 'application/json'
    };
    my $hContent = $self->_buildUploadImageContents($ImageUrl);

    my $HttpResponse = $self->sendPostRequest($ImageUploadUrl, $hHeadersParams, $hContent);
    my $hResponse = $self->decodeResponseContent($HttpResponse);

    my $ImageUploadResponseDTO = DE_EPAGES::Zettle::API::WebService::DTO::ImageUploadResponseDTO->parseResponse($hResponse);

    return $ImageUploadResponseDTO;
}

#========================================================================================
# function     bulkUploadImages
# state        public
#----------------------------------------------------------------------------------------
# syntax       $BulkImageUploadResponseDTO = $self->bulkUploadImages($aImageUrls);
#----------------------------------------------------------------------------------------
# description  Uploads one or more images from a list of URLs.
#               Image formats supported are: [BMP,GIF,JPEG,PNG,TIFF].
#               The images must be > 50*50px and < 5MB.
#               See: https://image.izettle.com/swagger#/images/uploadProductImageBulk
#----------------------------------------------------------------------------------------
# input        $aImageUrls | List of image URLs | ref.array.string
#----------------------------------------------------------------------------------------
# return       $BulkImageUploadResponseDTO | a BulkImageUploadResponseDTO object | DE_EPAGES::Zettle::API::WebService::DTO::BulkImageUploadResponseDTO
#========================================================================================
sub bulkUploadImages {
    my $self = shift;

    my ($aImageUrls) = @_;

    my $BulkImageUploadUrl = GetBulkUploadImageApiUrl();
    my $hHeadersParams = {
        'Content-type' => 'application/json'
    };
    my $hContent = $self->_buildBulkUploadImagesContent($aImageUrls);

    my $HttpResponse = $self->sendPostRequest($BulkImageUploadUrl, $hHeadersParams, $hContent);
    # NOTE: see DE_EPAGES/Zettle/Documents/API/ZettleAPI-Extracts.md, 4.2.1.2. Sample BulkProductImageUpload response
    my $hResponse = $self->decodeResponseContent($HttpResponse);

    my $BulkImageUploadResponseDTO = DE_EPAGES::Zettle::API::WebService::DTO::BulkImageUploadResponseDTO->fromHash($hResponse);

    return $BulkImageUploadResponseDTO;
}


#========================================================================================


#========================================================================================
# function     _getImageFormat
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ImageFormat = $self->_getImageFormat($ImageUrl);
#----------------------------------------------------------------------------------------
# description  Retrieve image format from file extension.
#----------------------------------------------------------------------------------------
# input        $ImageUrl | image url | string
#----------------------------------------------------------------------------------------
# return       $ImageFormat | image format, any of: BMP, GIF, JPEG, PNG, TIFF | string
#========================================================================================
sub _getImageFormat {
    my $self = shift;

    my ($ImageUrl) = @_;

    my @Elems = split(/\//, $ImageUrl);
    my $FileName = $Elems[-1];

    my ( undef, undef, $Extension ) = fsplit($FileName);

    unless ( exists $EXTENSION_TO_IMAGE_FORMAT->{ lc($Extension) } ) {
        Error(
            'UNSUPPORTED_IMAGE_FORMAT',
            {
                Extension => $Extension,
                ImageUrl  => $ImageUrl
            }
        );
    }
    
    return $EXTENSION_TO_IMAGE_FORMAT->{ lc($Extension) };
}

sub _buildUploadImageContents {
    my $self = shift;
    my ($ImageUrl) = @_;

    my $ImageFormat = $self->_getImageFormat($ImageUrl);
    return {
        'imageFormat' => $ImageFormat,
        # 'imageData' => undef, # This field can be used to export raw image data instead of URL
        'imageUrl' => $ImageUrl,
    }
}

sub _buildBulkUploadImagesContent {
    my $self = shift;
    my ($aImageUrls) = @_;

    my @aImageUploadRequestItems;
    foreach my $ImageUrl (@{$aImageUrls}){
        eval {
            my $ImageUploadRequestItem = $self->_buildUploadImageContents($ImageUrl);
            push(@aImageUploadRequestItems, $ImageUploadRequestItem);
        };
        # Log error for individual image (i.e. UNSUPPORTED_FORMAT) and process the rest of the images
        if(ExistsError()){
            my $Error = GetError();
            LogError($Error);
        }
    }

    return {
        'imageUploads' => \@aImageUploadRequestItems,
    };
}

1;
