#========================================================================================
# package      DE_EPAGES::Zettle::API::ZettleService
# state        public
#----------------------------------------------------------------------------------------
# description  object for calling functions on the ZettleService which serves as a Zettle facade
#========================================================================================
package DE_EPAGES::Zettle::API::ZettleService;

use strict;

use Data::Compare;
use JSON qw ( decode_json );

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( GetLog LogError );
use DE_EPAGES::Core::API::PerlTools qw ( GetCurrentDateTime );
use DE_EPAGES::Core::API::TestParameter  qw( TestObject );
use DE_EPAGES::Object::API::Factory  qw ( ExistsObjectByGUID LoadClassByAlias LoadObjectByGUID );
use DE_EPAGES::Zettle::API::Constants qw ( %INITIAL_EXPORT_STATUS );
use DE_EPAGES::Zettle::API::ImageHelper qw ( GetGuidImageMap GetImageUploadResults );
use DE_EPAGES::Zettle::API::TokenHelper;
use DE_EPAGES::Zettle::API::Tools qw ( GetExportableProducts LoadProductByZettleUUID );
use DE_EPAGES::Zettle::API::WebService::AuthService;
use DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateDTO;
use DE_EPAGES::Zettle::API::WebService::DTO::WebhookEventDTO;
use DE_EPAGES::Zettle::API::WebService::ImageService;
use DE_EPAGES::Zettle::API::WebService::InventoryService;
use DE_EPAGES::Zettle::API::WebService::LocationService;
use DE_EPAGES::Zettle::API::WebService::MerchantService;
use DE_EPAGES::Zettle::API::WebService::PusherService;
use DE_EPAGES::Zettle::API::WebService::ProductService;
use DE_EPAGES::Zettle::API::WebService::VariationOptionHelper qw ( BuildVariantOptions );


#ERRORDEFINITION
sub ErrorDefinition { return {}
}
#ERRORDEFINITION END

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::ZettleService->new($Environment)
#----------------------------------------------------------------------------------------
# description  creates a new ZettleService object
#----------------------------------------------------------------------------------------
# input        $Environment | the Zettle environment object | DE_EPAGES::Zettle::API::Object:ZettleEnvironment
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::ZettleService
#========================================================================================
sub new {
    my $class = shift;
    my ($Environment)  = @_;
    TestObject('ZettleEnvironment', $Environment, 'DE_EPAGES::Zettle::API::Object::ZettleEnvironment');
    
    my $self  = bless {
        'Environment' => $Environment
    }, $class;

    return $self;
}

#========================================================================================
# function     getEnvironment
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Environment = $self->getEnvironment()
#----------------------------------------------------------------------------------------
# description  Returns the Zettle Environment object.
#----------------------------------------------------------------------------------------
# return       $Environment | Zettle Environment | Object
#========================================================================================
sub getEnvironment {
    return shift->{'Environment'};
}

#========================================================================================
# function     connect
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->connect($AuthCode)
#----------------------------------------------------------------------------------------
# description  Use given $AuthCode to retrieve tokens from Zettle OAuth API and create/update zettle environment storing token data.
#               See also <function DE_EPAGES::Zettle::API::WebService::AuthService::getTokensOnAuthCode>.
#========================================================================================
sub connect {
    my $self = shift;
    my ($AuthCode) = @_;

    my $Environment = $self->getEnvironment();

    # Use OAuth service to exchange AuthCode for Access and Refresh tokens
    my $AuthService = DE_EPAGES::Zettle::API::WebService::AuthService->new($Environment);
    my $hTokens = $AuthService->getTokensOnAuthCode($AuthCode);
    
    my $TokenHelper = $self->_createTokenHelper($Environment);
    $TokenHelper->updateTokenData($hTokens);

    my $MerchantService = DE_EPAGES::Zettle::API::WebService::MerchantService->new($Environment);
    my $AccountInfo = $MerchantService->getAccountInfo();
    
    my $ContactEmail = $AccountInfo->{'contactEmail'};
    my $ReceiptEmail = $AccountInfo->{'receiptEmail'};
    my %MerchantInfo = (
        'uuid'     => $AccountInfo->{'uuid'},
        'email'    => $ContactEmail // $ReceiptEmail,
        'country'  => $AccountInfo->{'country'},
        'language' => $AccountInfo->{'language'},
        'currency' => $AccountInfo->{'currency'},
    );

    eval {
        $Environment->setMerchantInfo(%MerchantInfo);
    };
    # Abort setup if merchant info could not be updated:
    # - disconnect from Zettle (revoke tokens)
    # - delete ZettleEnvironment object
    if( ExistsError() ){
        my $Error = GetError();
        $self->disconnect();
        $Environment->delete();
        Error($Error);
    };
}

#========================================================================================
# function     disconnect
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->disconnect($MarkForDeletion)
#----------------------------------------------------------------------------------------
# input        $MarkForDeletion | Optional. Determines whether ZettleEnvironment will be marked to be deleted by batch job (default: TRUE) | boolean
#----------------------------------------------------------------------------------------
# description  Disconnect from Zettle API and remove token data from zettle environment
#========================================================================================
sub disconnect {
    my $self = shift;
    my ($MarkForDeletion) = @_;

    $MarkForDeletion //= 1;

    my $Environment = $self->getEnvironment();
    my $AuthService = DE_EPAGES::Zettle::API::WebService::AuthService->new($Environment);
    eval {
        $AuthService->disconnect();
    };
    # Ignore errors like token expired or token already deleted
    if ( ExistsError() ) {
        my $Error = GetError();
        LogError($Error);
    }

    my $TokenHelper = $self->_getTokenHelper();
    $TokenHelper->deleteTokenData();

    # Configure ZettleEnvironment so that batch job proceeds with deletion afterwards if MarkForDeletion is true
    $Environment->markForDeletion($MarkForDeletion);
}

#========================================================================================
# function     isTokenAcquired
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->isTokenAcquired()
#----------------------------------------------------------------------------------------
# description  Check if tokens were retrieved from Zettle
#               See <function DE_EPAGES::Zettle::API::TokenHelper::isTokenAcquired>
#----------------------------------------------------------------------------------------
# return        TRUE if tokens present in ZettleEnvironment, false otherwise
#========================================================================================
sub isTokenAcquired {
    my $self = shift;    
    
    my $TokenHelper = $self->_getTokenHelper();
    return $TokenHelper->isTokenAcquired();
}

#========================================================================================
# function     initialExport
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->initialExport()
#----------------------------------------------------------------------------------------
# description  Run an initial product export which will send to Zettle all products that match criteria, and update InitialExportStatus in environment.
#               See <function DE_EPAGES::Zettle::API::WebService::ProductService::exportProducts>.
#               See <function DE_EPAGES::Zettle::API::WebService::DTO::Requirements::IsProductExportable>.
#----------------------------------------------------------------------------------------
# input       $ReplaceZettleLibrary | Determines whether all products will be removed from Zettle before export | boolean
# input       $IncludePrices | Determines if prices will be included in product export | boolean
#----------------------------------------------------------------------------------------
# return      Number of items to be processed by Zettle.
#========================================================================================
sub initialExport {
    my $self = shift;
    my ($ReplaceZettleLibrary, $IncludePrices) = @_;

    my $Environment = $self->getEnvironment();
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);

    if ($ReplaceZettleLibrary){
        $self->_deleteLibrary();
    }else{
        # [STEP] ZettleUUIDs: use product info retrieved from Zettle to fill in product ZettleUUID attribute for previously exported products/variations
        my $aProductResponseDTO = $ProductService->getAllProducts();
        $self->_updateShopProductsUUID($aProductResponseDTO);
    }

    # [STEP] Retrieve all exportable products
    my $Shop = $Environment->getSite();
    my $aProducts = GetExportableProducts($Shop);

    # [STEP] ProductImages: export all product images and retrieve imageUploadResults to be used in product export.
    # NOTE: By default product 'ImageLarge' image is exported. A different image may be exported by adding $ImageFormat argument in _exportImages() call.
    # All product images can be exported by omitting $ImageFormat argument.
    my $hImageUploadResults = $self->_exportImages($aProducts, 'ImageLarge');

    # [STEP] ProductData: export fields of interest (which will include image lookupKeys from $hImageUploadResults)
    my $hResponse = $ProductService->exportProducts($aProducts, $IncludePrices, $hImageUploadResults);
    my $InitialExportUUID = $hResponse->{uuid};
    $Environment->setupInitialExport($InitialExportUUID, $IncludePrices);
    
    Error('InitialExportError', $hResponse) if $hResponse->{state} eq 'FINISHED_FAILED';
    
    my $State = $hResponse->{'state'};
    my $Status = $self->_getStatusFromState($State);
    $Environment->initialExportStatus($Status);

    # [STEP] Retrieve location UUIDs from Zettle
    $self->_updateAllLocationUUID($Environment);
    
    my $InventoryService = DE_EPAGES::Zettle::API::WebService::InventoryService->new($Environment);

    # [STEP] Enable InventoryTracking for exported products to ensure newly created products are tracked
    my $hProductBalanceBulk = $self->_enableInventoryTracking($InventoryService, $aProducts);
    
    # [STEP] Export stock levels
    if(scalar @{$aProducts}){
        $self->_exportStockLevels($InventoryService, $aProducts, $hProductBalanceBulk);
    }

    return $hResponse->{items};
}

#========================================================================================
# function     isInitialExportComplete
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->isInitialExportComplete()
#----------------------------------------------------------------------------------------
# description  Check if initial product export has been completed
#----------------------------------------------------------------------------------------
# return        TRUE if completed, false otherwise
#========================================================================================
sub isInitialExportComplete {
    my $self = shift;
    my $InitialExportStatus = $self->getInitialExportStatus();
    return ($InitialExportStatus == $INITIAL_EXPORT_STATUS{COMPLETED}) ? 1 : 0;
}

#========================================================================================
# function     getInitialExportStatus
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->getInitialExportStatus()
#----------------------------------------------------------------------------------------
# description  Retrieve updated initial export status querying Zettle if necessary (if export was RUNNING), updating environment export status is status changed.
#----------------------------------------------------------------------------------------
# return       $ExportStatus | See %INITIAL_EXPORT_STATUS values in <DE_EPAGES::Zettle::API::Constants> | integer
#========================================================================================
sub getInitialExportStatus {
    my $self = shift;
    my $Environment = $self->getEnvironment();
    my $InitialExportStatus = $Environment->initialExportStatus();

    # If last known status was 'running' we query Zettle for updated status
    if($InitialExportStatus == $INITIAL_EXPORT_STATUS{RUNNING}){
        my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
        my $ExportUUID = $Environment->initialExportUUID();
        my $State;
        eval {
            $State = $ProductService->getInitialExportState($ExportUUID);
        };
        if(ExistsError()){
            my $Error = GetError();
            LogError($Error);
            $State = $Error->vars()->{'state'};
        }
        $InitialExportStatus = $self->_getStatusFromState($State);
        # Update export status in environment if status changed
        $Environment->initialExportStatus($InitialExportStatus) unless $InitialExportStatus == $INITIAL_EXPORT_STATUS{RUNNING};
    }

    return $InitialExportStatus;
}

#========================================================================================
# function     exportProductUpdates
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportProductUpdates($IncludePrices);
#----------------------------------------------------------------------------------------
# description  Export product basic information to Zettle for products modified after previous export.
# notes        Previous product export datetime is stored in ZettleEnvironment 'LastProductExport' field.
#========================================================================================
sub exportProductUpdates {
    my $self = shift;
    my ($IncludePrices) = @_;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();
    my $LastProductExport = $Environment->lastProductExport();
    my $aProducts = $self->_getModifiedBasicDataProducts($Shop, $LastProductExport);

    #TODO: determine images modified since $LastProductExport (create and use ImageLastUpdate new profile to monitor ImageLarge)
    # InitialExport must export all product images whereas regular export should only export modified images.
    my $hImageUploadResults = $self->_exportImages($aProducts, 'ImageLarge');

    # Export product basic information
    my $hResponse;
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
    $hResponse = $ProductService->exportProducts($aProducts, $IncludePrices, $hImageUploadResults) if scalar @{$aProducts};
    
    $Environment->lastProductExport(GetCurrentDateTime()) if $hResponse->{state} ne 'FINISHED_FAILED';

    return $hResponse // {};
}

#========================================================================================
# function     exportPriceUpdates
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportPriceUpdates();
#----------------------------------------------------------------------------------------
# description  Export product prices to Zettle for products whose price was modified after previous price export.
# notes        Previous price export datetime is stored in ZettleEnvironment 'LastPriceExport' field.
#========================================================================================
sub exportPriceUpdates {
    my $self = shift;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();
    my $LastPriceExport = $Environment->lastPriceExport();
    my $aProducts;
    if(defined $LastPriceExport){
        $aProducts = $self->_getModifiedPriceProducts($Shop, $LastPriceExport);
    }else{
        $aProducts = GetExportableProducts($Shop);
    }

    #TODO: implement and test ProductService::exportPrices
    my $hResponse;
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
    $hResponse = $ProductService->exportPrices($aProducts) if scalar @{$aProducts};
    # my $IncludePrices = 1;
    # my $hResponse = $ProductService->exportProducts($aProducts, $IncludePrices, {}) if scalar @{$aProducts};

    $Environment->lastPriceExport(GetCurrentDateTime()) if $hResponse->{state} ne 'FINISHED_FAILED';

    return $hResponse // {};
}

#========================================================================================
# function     setupStockWebhook
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->setupStockWebhook($Destination, $ContactEmail);
#----------------------------------------------------------------------------------------
# description  Subscribe to inventory events to receive stock level updates via Zettle webhooks.
#               See also: <function DE_EPAGES::Zettle::API::ZettleService::incomingStockUpdate>.
# notes        After subscription is created a POST request is sent to $Destination when subscribed event triggers.
#               If Zettle request fails (is not correctly processed by ePages) an email is sent to merchant email.
#========================================================================================
sub setupStockWebhook {
    my $self = shift;
    my ($Destination, $ContactEmail) = @_;

    #TODO: REMOVE BEFORE PUSHING TO REPO
    $Destination =~ s/zettle-.*\.epnow-01\.upango\.es/epnow-01\.upango\.es/;

    my $Environment = $self->getEnvironment();

    # Inventory-related events
    my @aEventNames = qw (
        InventoryBalanceChanged
        InventoryTrackingStarted
        InventoryTrackingStopped
    );
    
    my $MethodName = (caller(0))[3];
    GetLog->debug("$MethodName: registering stock webhooks - ".join(',', @aEventNames));
    
    my $PusherService = DE_EPAGES::Zettle::API::WebService::PusherService->new($Environment);
    my $SubscriptionResponseDTO = $PusherService->createSubscription($Destination, $ContactEmail, \@aEventNames);

    my $Uuid = $SubscriptionResponseDTO->uuid();
    my $SigningKey = $SubscriptionResponseDTO->signingKey();
    $Environment->setupInventorySubscription($Uuid, $SigningKey);
}

#========================================================================================
# function     incomingStockUpdate
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->incomingStockUpdate($hEventContent);
#----------------------------------------------------------------------------------------
# description  Process $hEventContent which contains inventory-related events.
#               Incoming messages for the following events:
#               <ul>
#                   <li>InventoryBalanceChanged - includes stock level changes for a list of products
#                   <li>InventoryTrackingStarted - informs tracking was enabled for a product
#                   <li>InventoryTrackingStopped - informs tracking was disabled for a product
#               </ul>
#               Events will be parsed and the shop will be updated accordingly:
#               update stock levels or (un)set ZettleTrackingEnabled product attributes.
#----------------------------------------------------------------------------------------
# input        $hEventContent | contents of an incoming event message with the following keys:
#                                <ul>
#                                    <li>organizationUuid - string
#                                    <li>messageUuid - UUID v1 based on timestamp and messageId - string
#                                    <li>eventName - https://github.com/iZettle/api-documentation/blob/master/pusher-api/api-reference.md#supported-events - string
#                                    <li>messageId - unique UUID of the message - string
#                                    <li>payload - specific payload depending on the type of event.
#                                        see https://github.com/iZettle/api-documentation/blob/master/pusher-api/user-guides/subscriptions.md#payloads - hash
#                                    <li>timestamp - event timestamp in ISO-8601 format - string
#                                </ul>
#========================================================================================
sub incomingStockUpdate {
    my $self = shift;
    my ($hEventContent) = @_;
    my $MethodName = (caller(0))[3];

    my $Payload = $hEventContent->{'payload'};
    # Decode event payload (which is also encoded)
    $hEventContent->{'payload'} = decode_json($Payload) if defined $Payload;

    my $Environment = $self->getEnvironment();

    my $WebhookEventDTO = DE_EPAGES::Zettle::API::WebService::DTO::WebhookEventDTO->fromHash($hEventContent);

    #TODO: skip duplicate events comparing messageId field with stored values

    my $OrganizationUuid = $WebhookEventDTO->organizationUuid();
    my $EventName = $WebhookEventDTO->eventName();
    my $MerchantUuid = $Environment->getMerchantUUID();
    my $MessageId = $WebhookEventDTO->messageId();
    Error("Invalid organizationUuid received in $EventName event: $OrganizationUuid (MessageId: $MessageId)") unless $MerchantUuid eq $OrganizationUuid;

    my $hPayload = $WebhookEventDTO->payload;

    if($EventName eq 'InventoryBalanceChanged'){
        my $StockUpdateDTO = DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateDTO->fromHash($hPayload);

        # Skip incoming events triggered by updates sent to Zettle from the shop
        my $externalUuid = $StockUpdateDTO->externalUuid();
        return if defined $externalUuid && ($externalUuid eq $MerchantUuid);

        my $aBalanceBeforeDTO = $StockUpdateDTO->balanceBefore();
        my $aBalanceAfterDTO = $StockUpdateDTO->balanceAfter();

        # Loop balance changes and update stock levels
        for(my $idx=0; $idx < scalar @{$aBalanceBeforeDTO}; $idx++){
            eval{
                my $BalanceBeforeDTO = @{$aBalanceBeforeDTO}[$idx];
                my $BalanceBefore = $BalanceBeforeDTO->balance();
                my $VariationUuid = $BalanceBeforeDTO->variationUuid();

                my $BalanceAfterDTO = @{$aBalanceAfterDTO}[$idx];
                my $BalanceAfter = $BalanceAfterDTO->balance();

                my $BalanceDifference = $BalanceAfter-$BalanceBefore;

                my $Product = LoadProductByZettleUUID($VariationUuid);

                my $CurrentStockLevel = $Product->get('StockLevel');
                # Only update if incoming stock level is different from current stock level (this prevents unnecessary updates)
                if($BalanceAfter != $CurrentStockLevel){
                    $Product->set({
                        'StockLevel' => $CurrentStockLevel + $BalanceDifference,
                        'IgnoreZettleUpdate' => 1,  # Add a virtual attribute used in hook OnBeforeUpdateProduct to prevent sending Zettle updates back to Zettle
                    });
                }
            };
            if(ExistsError()){
                my $Error = GetError();
                $Error->addVars({
                    'IncomingEvent' => $EventName,
                    'MessageId' => $MessageId
                });
                LogError($Error);
            }else{
                $Environment->lastStockSyncZettle(GetCurrentDateTime());
            }
        }
    }elsif($EventName eq 'InventoryTrackingStarted'){
        my ($TrackedProductId, $TrackedProductZettleUuid);
        eval{
            $TrackedProductZettleUuid = $hPayload->{'productUuid'};
            my $TrackedProduct = LoadProductByZettleUUID($TrackedProductZettleUuid);
            if(defined $TrackedProduct){
                $TrackedProductId = $TrackedProduct->id();
                $TrackedProduct->set({'ZettleTrackingEnabled' => 1});
            }
        };
        if(ExistsError()){
            my $Error = GetError();
            $Error->addVars({
                'IncomingEvent' => $EventName,
                'MessageId' => $MessageId,
                'ZettleProductUuid' => $TrackedProductZettleUuid,
                'ProductID' => $TrackedProductId,
            });
            LogError($Error);
        }
    }elsif($EventName eq 'InventoryTrackingStopped'){
        my ($UntrackedProductId, $UntrackedProductZettleUuid);
        eval{
            $UntrackedProductZettleUuid = $hPayload->{'productUuid'};
            my $UntrackedProduct = LoadProductByZettleUUID($UntrackedProductZettleUuid);
            if(defined $UntrackedProduct){
                $UntrackedProductId = $UntrackedProduct->id();
                $UntrackedProduct->set({'ZettleTrackingEnabled' => 0});
            }
        };
        if(ExistsError()){
            my $Error = GetError();
            $Error->addVars({
                'IncomingEvent' => $EventName,
                'MessageId' => $MessageId,
                'ZettleProductUuid' => $UntrackedProductZettleUuid,
                'ProductID' => $UntrackedProductId,
            });
            LogError($Error);
        }
    }else{
        GetLog->debug("$MethodName - ignoring unexpected event of type ".$EventName);
    }

    return;
}


#========================================================================================


#========================================================================================
# function     _exportImages
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hImageUploadResults = $self->_exportImages($aProducts, $ImageFormat)
#----------------------------------------------------------------------------------------
# description  Export to Zettle images for given products in the specified format.
#----------------------------------------------------------------------------------------
# input        $aProducts | list of products objects | ref.array.object
# input        $ImageFormat | optional. name of the image format to be processed,
#                              any of: ImageLarge, ImageMedium, ImageSmall, ImageThumbnail
#                              If argument not present, all product images will be exported | string
#----------------------------------------------------------------------------------------
# return       $hImageUploadResults | a map of ImageUploadResult objects indexed by product GUID.
#                                      See <DE_EPAGES::Zettle::API::Object::ImageUploadResult> | ref.hash.array.object
#========================================================================================
sub _exportImages {
    my $self = shift;
    my ($aProducts, $ImageFormat) = @_;
    
    my $Environment = $self->getEnvironment();
    my $ImageService = DE_EPAGES::Zettle::API::WebService::ImageService->new($Environment);
    my $hImageUrls = GetGuidImageMap($aProducts, $ImageFormat);

    # Merge the list of image URLs of each product
    my @aImageUrls;
    foreach my $aImageUrlSubList (values %{$hImageUrls}) {
        push(@aImageUrls, @{$aImageUrlSubList});
    }

    my $BulkImageUploadResponseDTO = $ImageService->bulkUploadImages(\@aImageUrls);

    my $hImageUploadResults = GetImageUploadResults($BulkImageUploadResponseDTO, $hImageUrls);

    return $hImageUploadResults;
}

#========================================================================================
# function     _updateShopProductsUUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_updateShopProductsUUID($aProductResponseDTO)
#----------------------------------------------------------------------------------------
# input        $aProductResponseDTO | Array of ProductResponseDTO (see DE_EPAGES::Zettle::API::WebService::DTO::ProductResponseDTO) | ref.array.object
#----------------------------------------------------------------------------------------
# description  Use product info retrievd from Zettle to fill in product ZettleUUID attribute for products previously exported to Zettle.
#               The goal is to avoid generating product duplicates in Zettle given that UUID information is the unique identifier used in Zettle API requests.
#               This let's us reuse previously generated UUIDs instead of generating new ones, which would end up with duplicates in Zettle.
#----------------------------------------------------------------------------------------
# return       $hUpdatedProducts | hash of updated products where key is product/variation ID and value is updated UUID | ref.hash
#========================================================================================
sub _updateShopProductsUUID {
    my $self = shift;
    my ($aProductResponseDTO) = @_;

    # Hash of product UUID updates
    my $hUpdatedProducts = {};
    foreach my $ProductResponseDTO (@{$aProductResponseDTO}){
        my $ZettleProductUuid = $ProductResponseDTO->uuid();
        # If product was exported from the shop, externalReference will contain product GUID
        my $ZettleExternalReference = $ProductResponseDTO->externalReference();
        if(defined $ZettleExternalReference && ExistsObjectByGUID($ZettleExternalReference)){
            my $Product = LoadObjectByGUID($ZettleExternalReference);
            TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');
            $Product->set({'ZettleUUID' => $ZettleProductUuid});
            $hUpdatedProducts->{$Product->id} = $ZettleProductUuid;
            
            # Also update UUID for product variations
            my $ahRawVariations = $ProductResponseDTO->variations();
            foreach my $hRawVariation (@{$ahRawVariations}){
                my $ZettleVariationUuid = $hRawVariation->{'uuid'};
                my $ZettleVariationOptions = $hRawVariation->{'options'};
                
                # NOTE: Zettle variations don't have externalReference field so the best we can do is match by variation attribute options.
                my $SubProductMatch = $self->_matchSubProductByOptions($Product, $ZettleVariationOptions);
                if(defined $SubProductMatch){
                    $SubProductMatch->set({'ZettleUUID' => $ZettleVariationUuid});
                    $hUpdatedProducts->{$SubProductMatch->id} = $ZettleVariationUuid;
                }
            }
        }
    }

    return $hUpdatedProducts;
}

#========================================================================================
# function     _matchSubProductByOptions
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SubProduct = $self->_matchSubProductByOptions($Product, $ZettleVariationOptions)
#----------------------------------------------------------------------------------------
# input        $Product | a super product | DE_EPAGES::Product::API::Object::Product
# input        $ZettleVariationOptions | attribute values for a zettle variation | ref.array.hash
#               [
#                   { 'name' => 'Colour', 'value' => 'Blue' },
#                   { 'name' => 'Size', 'value' => 'M' },
#                   { 'name' => 'Material', 'value' => 'Cotton' },
#               ]
#----------------------------------------------------------------------------------------
# description  Search SubProducts of given $Product and return the one whose variation attribute options match $ZettleVariationOptions.
#----------------------------------------------------------------------------------------
# return       $SubProduct | a SubProduct that matches given options or undef | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub _matchSubProductByOptions {
    my $self = shift;
    my ($Product, $ZettleVariationOptions) = @_;

    foreach my $SubProduct (@{$Product->get('SubProducts')}){
        my $SubProductOptions = BuildVariantOptions($SubProduct);
        # We need to specify full Compare package to avoid perl critic complaining about PerlTools not 'use'd
        return $SubProduct if Data::Compare::Compare($SubProductOptions, $ZettleVariationOptions);
    }

    return;
}

#========================================================================================
# function     _deleteLibrary
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_deleteLibrary()
#----------------------------------------------------------------------------------------
# description  Delete all products from Zettle library.
#========================================================================================
sub _deleteLibrary {
    my $self = shift;

    my $Environment = $self->getEnvironment();
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
    
    # Retrieve existing products
    my $aProductResponseDTO = $ProductService->getAllProducts();
    my @aProductUUIDs;
    foreach my $ProductResponseDTO (@{$aProductResponseDTO}){
        my $ProductUUID = $ProductResponseDTO->uuid;
        push(@aProductUUIDs, $ProductUUID);
    }
    # Delete products retrieved
    $ProductService->deleteProducts(\@aProductUUIDs) if scalar(@aProductUUIDs);
}

#========================================================================================
# function     _enableInventoryTracking
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_enableInventoryTracking($InventoryService, $aProducts)
#----------------------------------------------------------------------------------------
# description  Enable InventoryTracking for given products.
#----------------------------------------------------------------------------------------
# input        $InventoryService | Inventory service | DE_EPAGES::Zettle::API::WebService::InventoryService
# input        $aProducts | List of products | ref.array.object
#----------------------------------------------------------------------------------------
# return       $hProductBalanceBulk | List of updated variation balances for STORE location, with the following keys:
#                                      <ul>
#                                          <li>locationUuid - Identifies the location - string
#                                          <li>variants - List of balances - array.hash
#                                              <ul>
#                                                  <li>productUuid - Identifies the product - string
#                                                  <li>variantUuid - Identifies the variant - string
#                                                  <li>balance - Balance of variation in location - string
#                                              </ul>
#                                      </ul> | ref.hash.*
#========================================================================================
sub _enableInventoryTracking {
    my $self = shift;
    my ($InventoryService, $aProducts) = @_;

    my @aProductUuid = map { $_->get('ZettleUUID') } @{$aProducts};
    my $TrackingStatus = 'START_TRACKING';

    my $hProductBalanceBulk = $InventoryService->bulkUpdateProductTracking(\@aProductUuid, $TrackingStatus);

    # We are subscribed to InventoryTrackingStarted event and it will generate an incoming callback for each of the products tracked
    # Even though, we force here TrackingEnabled in all products given that asynchronous messages might take a while to be received
    # This would cause stock levels not to be exported for some products
    $_->set({'ZettleTrackingEnabled' => 1}) foreach @{$aProducts};

    return $hProductBalanceBulk;
}

#========================================================================================
# function     _updateAllLocationUUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_updateAllLocationUUID($Environment)
#----------------------------------------------------------------------------------------
# description  Retrieve all location UUIDs from Zettle and store them in ZettleEnvironment.
#----------------------------------------------------------------------------------------
# input        $Environment | a Zettle environment object | DE_EPAGES::Zettle::API::Object:ZettleEnvironment
#----------------------------------------------------------------------------------------
# return       $aLocationDTO | List of locations.
#                               See DE_EPAGES::Zettle::API::WebService::DTO::LocationDTO | ref.array.object
#========================================================================================
sub _updateAllLocationUUID {
    my $self = shift;
    my ($Environment) = @_;

    my $LocationService = DE_EPAGES::Zettle::API::WebService::LocationService->new($Environment);
    my $aLocationDTO = $LocationService->getAllLocations();
    foreach my $LocationDTO (@{$aLocationDTO}){
        my $LocationUuid = $LocationDTO->uuid();
        if($LocationDTO->name() eq 'BIN'){
            $Environment->locationBinUuid($LocationUuid);
        }elsif($LocationDTO->name() eq 'SOLD'){
            $Environment->locationSoldUuid($LocationUuid);
        }elsif($LocationDTO->name() eq 'STORE'){
            $Environment->locationStoreUuid($LocationUuid);
        }elsif($LocationDTO->name() eq 'SUPPLIER'){
            $Environment->locationSupplierUuid($LocationUuid);
        }
    }

    return $aLocationDTO;
}

#========================================================================================
# function     _exportStockLevels
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_exportStockLevels($InventoryService, $aProducts, $hProductBalanceBulk)
#----------------------------------------------------------------------------------------
# description  Export stock levels for given products considering $hProductBalanceBulk (current stock levels in Zettle).
#----------------------------------------------------------------------------------------
# input        $InventoryService | Inventory service | DE_EPAGES::Zettle::API::WebService::InventoryService
# input        $aProducts | List of products | ref.array.object
# input        $hProductBalanceBulk | Current Zettle stock levels. List of variation balances with the following keys:
#                                      <ul>
#                                          <li>locationUuid - Identifies the location - string
#                                          <li>variants - List of balances - array.hash
#                                              <ul>
#                                                  <li>productUuid - Identifies the product - string
#                                                  <li>variantUuid - Identifies the variant - string
#                                                  <li>balance - Balance of variation in location - string
#                                              </ul>
#                                      </ul> | ref.hash.*
#----------------------------------------------------------------------------------------
# return       $hExportStockBulk | List of updated variation balances.
#                                   Same structure as $hProductBalanceBulk parameter | ref.hash.*
#========================================================================================
sub _exportStockLevels {
    my $self = shift;
    my ($InventoryService, $aProducts, $hProductBalanceBulk) = @_;

    my $hBalanceUpdate = {};
    # Use $hProductBalanceBulk as a reference to calculate the stock difference to send to Zettle for each variation
    # If variation is not present in hash we'll then send current variation stock level
    my $ahVariationBalances = $hProductBalanceBulk->{'variants'};
    
    # If variant balances is empty we need to make sure we use the right balances retrieving them from Zettle
    $ahVariationBalances = $self->_getZettleSTOREVariationBalances($InventoryService) unless defined $ahVariationBalances && scalar @{$ahVariationBalances};

    # Only export stock levels for tracked products or we'll get an error
    my @aTrackedProducts = grep { $_->get('ZettleTrackingEnabled') == 1 } @{$aProducts};
    foreach my $Product (@aTrackedProducts){
        my $ProductUuid = $Product->get('ZettleUUID');
        my $aSubProducts = $Product->get('SubProducts');
        my $hVariationUpdates = {};
        if(scalar @{$aSubProducts}){
            # Calculate stock update for each subproduct: difference between shop stock and zettle stock
            foreach my $SubProduct (@{$aSubProducts}){
                my $VariantUuid = $SubProduct->get('ZettleUUID');
                my $StockDelta = $self->_getProductStockDelta($SubProduct, $ahVariationBalances);
                # Stock changes sent to Zettle must be greater than zero
                $hVariationUpdates->{$VariantUuid} = $StockDelta if $StockDelta > 0;
            }
        }else{
            # If no SubProducts, search for a fake zettle variant with same ZettleUuid as $Product
            my $VariantUuid = $Product->get('ZettleUUID');
            my $StockDelta = $self->_getProductStockDelta($Product, $ahVariationBalances);
            $hVariationUpdates->{$VariantUuid} = $StockDelta if $StockDelta > 0;
        }
        $hBalanceUpdate->{$ProductUuid} = $hVariationUpdates;
    }

    my $hExportStockBulk = $InventoryService->bulkUpdateProductBalance($hBalanceUpdate);
    
    my $Environment = $self->getEnvironment();
    $Environment->lastStockSync(GetCurrentDateTime());

    return $hExportStockBulk;
}

#========================================================================================
# function     _getZettleSTOREVariationBalances
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ahVariationBalances = $self->_getZettleSTOREVariationBalances($InventoryService);
#----------------------------------------------------------------------------------------
# description  Use InventoryService to retrieve all variation balances from Zettle STORE location.
#----------------------------------------------------------------------------------------
# input        $InventoryService | Inventory service | DE_EPAGES::Zettle::API::WebService::InventoryService
#----------------------------------------------------------------------------------------
# return       $ahVariationBalances | List of variation balances with the following keys:
#                                      <ul>
#                                          <li>productUuid - Identifies the product - string
#                                          <li>variantUuid - Identifies the variant - string
#                                          <li>balance - Balance of variation in location - string
#                                      </ul> | ref.array.hash
#========================================================================================
sub _getZettleSTOREVariationBalances {
    my $self = shift;
    my ($InventoryService) = @_;

    my @ahVariationBalances = ();
    # Retrieve balances from Zettle STORE location
    my $LocationInventoryDTO = $InventoryService->getLocationInventory();
    foreach my $VariationBalanceDTO (@{$LocationInventoryDTO->variationBalances()}){
        my $VariationBalance = {
            'productUuid' => $VariationBalanceDTO->productUuid(),
            'variantUuid' => $VariationBalanceDTO->variationUuid(),
            'balance' => $VariationBalanceDTO->balance(),
        };
        push(@ahVariationBalances, $VariationBalance);
    }

    return \@ahVariationBalances;
}

#========================================================================================
# function     _getProductStockDelta
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_getProductStockDelta($Product, $ahVariationBalances)
#----------------------------------------------------------------------------------------
# description  Calculate difference between shop stock level and zettle stock level.
#               Search $Product ZettleUUID in $ahVariationBalances in order to calculate difference.
#               If ZettleUUID is not present in hash return $Product StockLevel.
#               If it's present, return ShopStock - ZettleStock.
#----------------------------------------------------------------------------------------
# input        $Product | a Product object | DE_EPAGES::Product::API::Object::Product
# input        $ahVariationBalances | List of variation balances with the following keys:
#                                      <ul>
#                                          <li>productUuid - Identifies the product - string
#                                          <li>variantUuid - Identifies the variant - string
#                                          <li>balance - Balance of variation in location - string
#                                      </ul> | ref.array.hash
#----------------------------------------------------------------------------------------
# return       $StockDelta | Difference between Zettle stock and shop stock | integer
#========================================================================================
sub _getProductStockDelta {
    my $self = shift;
    my ($Product, $ahVariationBalances) = @_;

    my $ZettleUuid = $Product->get('ZettleUUID');
    my $ShopStock = $Product->get('StockLevel') // 0;
    
    # Search $ahVariationBalances for a variant that matches $ZettleUuid
    my $ZettleStock;
    foreach my $hVariantBalance (@{$ahVariationBalances}){
        if($hVariantBalance->{'variantUuid'} eq $ZettleUuid){
            $ZettleStock = $hVariantBalance->{'balance'};
            last;
        }
    }

    my $StockDelta;
    if(defined $ZettleStock){
        $StockDelta = $ShopStock - $ZettleStock;
    }else{
        $StockDelta = $ShopStock;
    }

    return $StockDelta;
}

sub _getModifiedBasicDataProducts {
    my $self = shift;
    my ($Shop, $LastBasicDataProductExport) = @_;

    my $aModifiedProducts = $self->_getModifiedProductsByProfile($Shop, $LastBasicDataProductExport, 'Content');

    return $aModifiedProducts;
}

sub _getModifiedPriceProducts {
    my $self = shift;
    my ($Shop, $LastPriceExport) = @_;

    my $aModifiedProducts = $self->_getModifiedProductsByProfile($Shop, $LastPriceExport, 'ListPrice');

    return $aModifiedProducts;
}

sub _getModifiedProductsByProfile {
    my $self = shift;
    my ($Shop, $LastUpdate, $ProfileName) = @_;

    my $Profile = LoadClassByAlias('Product')->lastUpdateProfile($ProfileName);

    # Load shop objects updated after given date relative to $Profile
    my $Iterator = $Shop->getUpdatedObjects($Profile, $LastUpdate);
        
    my @aModifiedProducts;
    while( my $Item = <$Iterator> ) {
        my $Object = $Item->{'Object'};
        push(@aModifiedProducts, $Object) if ref $Object eq 'DE_EPAGES::Product::API::Object::Product';
    }

    return \@aModifiedProducts;
}

#========================================================================================
sub _getTokenHelper {
    my $self = shift;
    my ($Environment) = $self->getEnvironment();

    $self->{'TokenHelper'} //= $self->_createTokenHelper($Environment);
    return $self->{'TokenHelper'};
}

#========================================================================================
sub _createTokenHelper {
    my $self = shift;
    my ($Environment) = @_;
    $self->{'TokenHelper'} //= DE_EPAGES::Zettle::API::TokenHelper->new($Environment);
    return $self->{'TokenHelper'};
}

#========================================================================================
sub _getStatusFromState {
    my $self = shift;
    my ($State) = @_;

    my $ExportStatus;
    if($State eq 'IMPORTING'){
        $ExportStatus = $INITIAL_EXPORT_STATUS{RUNNING};
    }elsif($State eq 'FINISHED_SUCCESS'){
        $ExportStatus = $INITIAL_EXPORT_STATUS{COMPLETED};
    }elsif($State eq 'FINISHED_FAILED'){
        $ExportStatus = $INITIAL_EXPORT_STATUS{FAILED};
    }

    return $ExportStatus;
}

1;
