#========================================================================================
# package      DE_EPAGES::Zettle::Hooks::Product
# state        public
# description  export stock level updates to Zettle
#========================================================================================
package DE_EPAGES::Zettle::Hooks::Product;

use strict;

use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( LogError );
use DE_EPAGES::Zettle::API::WebService::InventoryService;
use DE_EPAGES::Zettle::API::Constants qw ( ZETTLE_ENVIRONMENT );


#========================================================================================
# function     OnAfterProductClone
#----------------------------------------------------------------------------------------
# syntax       OnAfterProductClone( $hParams );
#----------------------------------------------------------------------------------------
# description  Remove ZettleUUID after duplication of product. Triggered Hook AfterProductClone.
#----------------------------------------------------------------------------------------
# input        $hParams |
#               <ul>
#                   <li>Object - product object - object</li>
#                   <li>Clone - clone product object - object</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnAfterProductClone {
    my ($hParams) = @_;

    my $Clone = $hParams->{'Clone'};

    $Clone->set({'ZettleUUID' => undef});

    # also remove ZettleUUID for subproducts
    for (@{$Clone->get('SubProducts')}) {
        $_->set({'ZettleUUID' => undef});
    }
}

#========================================================================================
# function     OnAfterUpdateProduct
# state        public
#----------------------------------------------------------------------------------------
# syntax       OnAfterUpdateProduct( $hValues );
#----------------------------------------------------------------------------------------
# description  Export stock level updates to Zettle
#               (Hook OBJ_AfterUpdateProduct)
#----------------------------------------------------------------------------------------
# input        $hParams | objects
#               <ul>
#                   <li>Object - object - DE_EPAGES::Product::API::Object::Product
#                   <li>Values - new product values - ref.hash
#               </ul> | ref.hash.*
#========================================================================================
sub OnBeforeUpdateProduct {
    my ($hParams) = @_;

    
    # only react if stocklevel changed.
    return unless defined $hParams->{'Values'};
    return unless defined $hParams->{'Values'}->{'StockLevel'};

    # ignore incoming webhook updates (when we process Zettle stock updates we shall not send them back to Zettle)
    if(defined $hParams->{'Values'}->{'IgnoreZettleUpdate'} && $hParams->{'Values'}->{'IgnoreZettleUpdate'} == 1){
        delete $hParams->{'Values'}->{'IgnoreZettleUpdate'};
        return;
    }

    my $Product = $hParams->{'Object'};
    my $hNewValues = $hParams->{'Values'};
    my $NewStockLevel = $hNewValues->{'StockLevel'};

    #TODO: should we test IsUnity
    # if ($Shop->get('IsUnityRegistered')) {

    my $Shop = $Product->getSite();
    return unless $Shop->existsChild(ZETTLE_ENVIRONMENT);

    my $Environment = $Shop->get(ZETTLE_ENVIRONMENT);
    my $InventoryService = DE_EPAGES::Zettle::API::WebService::InventoryService->new($Environment);
    my $OldStockLevel = $Product->get('StockLevel');
    
    my $SuperProduct = $Product->superProduct;
    my $ProductUuid = (defined $SuperProduct) ? $SuperProduct->get('ZettleUUID') : $Product->get('ZettleUUID');
    my $VariantUuid = $Product->get('ZettleUUID');
    my $BalanceChange = $NewStockLevel-$OldStockLevel;

    eval{
        $InventoryService->updateProductBalance($ProductUuid, $VariantUuid, $BalanceChange);
    };
    if(ExistsError()){
        my $Error = GetError();
        LogError($Error);
    }

    return;
}

1;


