#========================================================================================
# package      DE_EPAGES::Zettle::UI::Shop
# state        private
#----------------------------------------------------------------------------------------
# description  Handles Zettle redirect and landing view (ViewAction logic and Zettle connection)
#========================================================================================
package DE_EPAGES::Zettle::UI::Shop;
use base DE_EPAGES::Presentation::UI::Object;

use strict;
use HTTP::Status qw ( HTTP_OK HTTP_BAD_REQUEST );
use JSON qw ( to_json );

use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( GetLog LogError );
use DE_EPAGES::Object::API::Factory qw( LoadClassByAlias LoadRootObject );
use DE_EPAGES::Shop::API::Url qw ( BuildShopUrl );
use DE_EPAGES::Zettle::API::Constants qw ( INTERNAL_CARTRIDGE_NAME ZETTLE_HOME_URL ZETTLE_ENVIRONMENT );
use DE_EPAGES::Zettle::API::Logger;
use DE_EPAGES::Zettle::API::Object::ZettleEnvironment;
use DE_EPAGES::Zettle::API::ZettleService;
use DE_EPAGES::Zettle::API::ZettleUrlBuilder;

use constant LANDING_VIEWACTION => 'UnityMBO-ViewZettleLanding';
use constant GENERAL_VIEWACTION => 'UnityMBO-ViewZettleGeneral';
use constant SETTINGS_VIEWACTION => 'UnityMBO-ViewZettleSettings';

#ERRORDEFINITION
sub ErrorDefinition { return {
    'NoAppClientID'            => "AppClientID not set in store #Store",
    'NoAppClientSecret'        => "AppClientSecret not set in store #Store",
    }
}
#ERRORDEFINITION END

sub UnityViewRedirect {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    # Check all necessary predefined configuration is setup (backlink URL, app credentials ...)
    _testPBOSetup();
        
    my $NextViewAction; # Next view action for servlet redirect (LANDING_VIEWACTION, GENERAL_VIEWACTION, SETTINGS_VIEWACTION)
    my $ViewObject; # Object for servlet redirect (Shop for LANDING_VIEWACTION, ZettleEnvironment for GENERAL_VIEWACTION and SETTINGS_VIEWACTION)

    my $Environment = $self->_getEnvironment($Shop);
    if(defined $Environment){
        my $ZettleService = DE_EPAGES::Zettle::API::ZettleService->new($Environment);
        my $IsTokenAcquired = $ZettleService->isTokenAcquired();

        if($IsTokenAcquired){
            my $IsInitialExportComplete = $ZettleService->isInitialExportComplete();
            # If initial export is complete show General tab, otherwise show Settings tab
            $NextViewAction = ($IsInitialExportComplete) ? GENERAL_VIEWACTION : SETTINGS_VIEWACTION;
            $ViewObject = $Environment;
        }else{
            $NextViewAction = LANDING_VIEWACTION;
            $ViewObject = $Shop;
        }
    }
    # If not exists Environment show Landing page
    else{
        $NextViewAction = LANDING_VIEWACTION;
        $ViewObject = $Shop;
    }

    $self->_redirect($Servlet, $NextViewAction, $ViewObject);

    return;
}

sub UnityViewLanding {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    my $UrlBuilder = DE_EPAGES::Zettle::API::ZettleUrlBuilder->new();
    my $ConnectUrl = $UrlBuilder->buildConnectUrl($Servlet, $Shop);
    $Servlet->vars('ConnectUrl', $ConnectUrl);
    
    my $System = LoadRootObject();
    my $ZettleLandingUrl = $System->get('Zettle_LandingURL');
    $ZettleLandingUrl //= ZETTLE_HOME_URL;
    $Servlet->vars('ZettleLandingUrl', $ZettleLandingUrl);

    $self->SUPER::View($Servlet);

    return;
}

sub ZettleOAuthCallback {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $Form = $Servlet->form();
    my $hQueryParameters = $Form->queryValues();
    my $payload;
    my $Code;
    my $Logger = DE_EPAGES::Zettle::API::Logger->new();
    
    # if auth code is received via backlink we create Zettle environment and exchange auth token for access and refresh tokens
    my $AuthCode = $Servlet->params('code');
    if(isValid($AuthCode)){
        $Logger->logCommunication('Auth code received: ', $hQueryParameters);
        my $Environment = $self->_createOrGetEnvironment($Shop);
        
        eval {
            my $ZettleService = DE_EPAGES::Zettle::API::ZettleService->new($Environment);
            $ZettleService->connect($AuthCode);
            
            $self->_setupWebhooks($Servlet, $Environment);
        };
        if (ExistsError()) {
            my $Error = GetError();
            LogError($Error);
            my $ErrorInfo = {
                GUID => $Error->guid
            };
            if($Error->code() eq 'NativeDie' || ! defined $Error->vars()->{'Response'}){
                $ErrorInfo->{'Code'} = "InternalError";
                $ErrorInfo->{'Message'} = $Error->message;
            }else{
                my $hErrorDetails = _getAuthError($Error);
                $ErrorInfo->{'Code'} = $hErrorDetails->{'code'};
                $ErrorInfo->{'Message'} = $hErrorDetails->{'message'};
            }
            $payload = "window.opener.ep.fn.showError('" . to_json($ErrorInfo) . "');";
            $Code = HTTP_BAD_REQUEST;
        }
        else{
            # If Zettle is connected redirect to Settings tab
            $payload = "window.opener.ep.fn.callbackSuccess('".SETTINGS_VIEWACTION."', '".$Environment->id."');";
            $Code = HTTP_OK;
        }
    }else{
        $Logger->logCommunication('Unexpected parameters to OAuth Callback: ', $hQueryParameters);
        $payload = "window.opener.ep.fn.showError('unknown_error');";
        $Code = HTTP_BAD_REQUEST;
    }


    my $jsContent = <<"EOF";
<script type="text/javascript">
    $payload
    window.close();
</script>
EOF
    return $self->_writeHTTPResponse($Servlet, $Code, {'ContentType' => 'text/html'}, $jsContent);
}

#========================================================================================

# Test that required zettle configuration is properly setup in PBO
sub _testPBOSetup {
    my $System = LoadRootObject();
    my $StoreName = $System->get('StoreName');

    my $ClientID = $System->get('Zettle_AppClientID');
    Error('NoAppClientID', { 'Store' => $StoreName }) unless isValid($ClientID);

    my $ClientSecret = $System->get('Zettle_AppClientSecret');
    Error('NoAppClientSecret', { 'Store' => $StoreName }) unless isValid($ClientSecret);
}

#========================================================================================
# function     _redirect
# state        private
#----------------------------------------------------------------------------------------
# syntax       $url = $self->_redirect($Servlet, $ViewAction, $Object);
#----------------------------------------------------------------------------------------
# description  Redirects to shop URL associated to the given $ViewAction and $Object
#----------------------------------------------------------------------------------------
# input        $Servlet | a Servlet object | DE_EPAGES::Presentation::Servlet::BasePermission
# input        $ViewAction | the name of a ViewAction | string
# input        $Object | any object associated with ViewAction | object
#========================================================================================
sub _redirect {
    my $self = shift;
    my ($Servlet, $ViewAction, $Object) = @_;
    my $MethodName = (caller(0))[3];
    
    my $URL = BuildShopUrl($Object,
        {'ViewAction' => $ViewAction},
        {'Type' => 'admin', 'UseSSL' => 1, 'AddToken' => 1},
        $Servlet);

    GetLog()->debug("$MethodName: redirecting to URL=$URL");
    $Servlet->redirect($URL);

    return;
}

#========================================================================================
# function     _createOrGetEnvironment
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_createOrGetEnvironment();
#----------------------------------------------------------------------------------------
# description  Create Zettle environment object or retrieve it from shop if it already exists.
#----------------------------------------------------------------------------------------
# input        $Shop | a Shop object | DE_EPAGES::Shop::API::Object::Shop
#----------------------------------------------------------------------------------------
# return       $Environment | environment object | DE_EPAGES::Zettle::API::Object::ZettleEnvironment
#========================================================================================
sub _createOrGetEnvironment {
    my $self = shift;
    my ($Shop) = @_;

    my $Environment = $self->_getEnvironment($Shop);
    
    $Environment = LoadClassByAlias(ZETTLE_ENVIRONMENT)->insertObject({
        Parent => $Shop,
        Alias => INTERNAL_CARTRIDGE_NAME,
        Position => 210
    }) unless defined $Environment;

    return $Environment;
}

#========================================================================================
# function     _getEnvironment
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Environment = $self->_getEnvironment();
#----------------------------------------------------------------------------------------
# description  Returns the ZettleEnvironment object if it exists.
#----------------------------------------------------------------------------------------
# input        $Shop | a Shop object | DE_EPAGES::Shop::API::Object::Shop
#----------------------------------------------------------------------------------------
# return       $Environment | Zettle environment | DE_EPAGES::Zettle::API::Object::ZettleEnvironment or undef
#========================================================================================
sub _getEnvironment {
    my $self = shift;
    my ($Shop) = @_;
    return $Shop->existsChild(ZETTLE_ENVIRONMENT) ? $Shop->child(ZETTLE_ENVIRONMENT) : undef;
}

sub _writeHTTPResponse {
    my $self = shift;
    my ($Servlet, $Code, $hParams, $Content) = @_;

    $Servlet->response->code($Code);
    foreach my $key (keys %{$hParams}) {
        $Servlet->params($key, $hParams->{$key});
    }
    return $Servlet->writeContentOutput($Content);
}

#========================================================================================
# function     _getAuthError
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hErrorDetails = _getAuthError($Error);
#----------------------------------------------------------------------------------------
# description  Retrieve error details from 'Response' variable in $Error.
#----------------------------------------------------------------------------------------
# input        $Error | an error object | DE_EPAGES::Core::API::Error
#----------------------------------------------------------------------------------------
# return       $hErrorDetails | error details from Response, with the following keys:
#                                <ul>
#                                    <li>code - zettle error code - string
#                                    <li>message - zettle error description - string
#                                </ul> | ref.hash
#========================================================================================
sub _getAuthError {
    my ($Error) = @_;

    my $Response = $Error->vars()->{'Response'};
    # $Response may contain an error in JSON format like the following:
    # {"error":"invalid_grant","error_description":"invalid authorization code"}
    my $DecodedResponse = JSON->new->utf8->decode($Response);

    my $hErrorDetails = {
        'code' => $DecodedResponse->{'error'},
        'message' => $DecodedResponse->{'error_description'}
    };

    return $hErrorDetails;
}

#========================================================================================
# function     _setupWebhooks
# state        private
#----------------------------------------------------------------------------------------
# syntax       _setupWebhooks($Servlet, $Environment);
#----------------------------------------------------------------------------------------
# description  Configure external webhooks to receive asynchronous updates from Zettle.
#               Multiple events are available but we only subscribe to stock related events:
#               <ul>
#                   <li> InventoryTrackingStarted
#                   <li> InventoryTrackingStopped
#                   <li> InventoryBalanceChanged
#               </ul>
#               For supported events see: https://github.com/iZettle/api-documentation/blob/master/pusher-api/api-reference.md#supported-events
#----------------------------------------------------------------------------------------
# input        $Servlet | a Servlet object | DE_EPAGES::Presentation::Servlet::BasePermission
# return       $Environment | environment object | DE_EPAGES::Zettle::API::Object::ZettleEnvironment
#========================================================================================
sub _setupWebhooks {
    my $self = shift;
    my ($Servlet, $Environment) = @_;

    my $UrlBuilder = DE_EPAGES::Zettle::API::ZettleUrlBuilder->new();
    my $Destination = $UrlBuilder->buildWebhookStockUrl($Servlet, $Environment);
    my $ContactEmail = $Servlet->user()->get('EMail'); # Use merchant email as contact email to notify webhook errors

    my $Service = DE_EPAGES::Zettle::API::ZettleService->new($Environment);
    $Service->setupStockWebhook($Destination, $ContactEmail);
}

1;
